 /***************************************************************************
 *          Apache Cayenne interface for RDBMS Linter source code           *
 *        This is unpublished proprietary source code of RELEX, Inc.        *
 *     The copyright notice does not evidence any actual or intended        *
 *                  publication of such source code.                        *
 *                                                                          *
 *                 Copyright (c) 2003-2014 RELEX Inc.                       *
 *                                                                          *
 *                   This software was developed by                         *
 *                                                                          *
 *                            RELEX, Inc.                                   *
 *                         Voronezh, Russia                                 *
 *                                                                          *
 *                        All rights reserved.                              *
 ****************************************************************************
 AUTHOR:
 ----------------------------------------------------------------------------
 PURPOSE:
 ----------------------------------------------------------------------------
 REVISION HISTORY:
*/
package org.apache.cayenne.dba.linter;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.apache.cayenne.CayenneRuntimeException;
import org.apache.cayenne.access.DataNode;
import org.apache.cayenne.access.QueryLogger;
import org.apache.cayenne.dba.JdbcPkGenerator;
import org.apache.cayenne.map.DbEntity;

/**
 * A sequence-based PK generator used by {@link linterAdapter}.
 */
public class LinterPkGenerator extends JdbcPkGenerator {

    private static final String _SEQUENCE_PREFIX = "S_";

    /**
     * @deprecated since 2.0 - other generators do not expose the default prefix.
     */
    public static final String SEQUENCE_PREFIX = "S_";

    protected int pkFromDatabase(DataNode node, DbEntity ent) throws Exception {

        String pkGeneratingSequenceName = sequenceName(ent);

if (node.getDataSource() == null)
                        throw new CayenneRuntimeException(
                                "node.getdatasource not defined");

        Connection con = node.getDataSource().getConnection();

if (con == null)
                        throw new CayenneRuntimeException(
                                "con not defined");

        try {
            Statement st = con.createStatement();
            try {
                String sql = "SELECT " + pkGeneratingSequenceName + ".nextval";
                QueryLogger.logQuery(sql, Collections.EMPTY_LIST);
                ResultSet rs = st.executeQuery(sql);
                try {
                    // Object pk = null;
                    if (!rs.next()) {
                        throw new CayenneRuntimeException(
                                "Error generating pk for DbEntity " + ent.getName());
                    }
                    return rs.getInt(1);
                }
                finally {
                    rs.close();
                }
            }
            finally {
                st.close();
            }
        }
        finally {
            con.close();
        }
    }

    public void createAutoPk(DataNode node, List dbEntities) throws Exception {
        Collection sequences = getExistingSequences(node);
        Iterator it = dbEntities.iterator();

        while (it.hasNext()) {
            DbEntity entity = (DbEntity) it.next();
            if (!sequences.contains(sequenceName(entity))) {
                this.runUpdate(node, createSequenceString(entity));
            }
        }
    }

    /**
     * Creates a list of CREATE SEQUENCE statements for the list of DbEntities.
     */
    public List createAutoPkStatements(List dbEntities) {
        List list = new ArrayList(dbEntities.size());
        Iterator it = dbEntities.iterator();

        while (it.hasNext()) {
            DbEntity ent = (DbEntity) it.next();
            list.add(createSequenceString(ent));
        }

        return list;
    }

    /**
     * Drops PK sequences for all specified DbEntities.
     */
    public void dropAutoPk(DataNode node, List dbEntities) throws Exception {
        Collection sequences = getExistingSequences(node);

        Iterator it = dbEntities.iterator();
        while (it.hasNext()) {
            DbEntity ent = (DbEntity) it.next();
            if (sequences.contains(sequenceName(ent))) {
                runUpdate(node, dropSequenceString(ent));
            }
        }
    }

    /**
     * Creates a list of DROP SEQUENCE statements for the list of DbEntities.
     */
    public List dropAutoPkStatements(List dbEntities) {

        List list = new ArrayList(dbEntities.size());
        Iterator it = dbEntities.iterator();

        while (it.hasNext()) {
            DbEntity entity = (DbEntity) it.next();
            list.add(dropSequenceString(entity));
        }

        return list;
    }

    /**
     * Fetches a list of existing sequences that might match Cayenne generated ones.
     */
    protected List getExistingSequences(DataNode node) throws SQLException {

        // check existing sequences
        Connection con = node.getDataSource().getConnection();

        try {
            Statement sel = con.createStatement();
            try {
                StringBuffer buffer = new StringBuffer();
                buffer.append("SELECT $$$NAME FROM LINTER_SYSTEM_USER.$$$SEQ ").append(
                        "WHERE $$$OWNER=LINTER_USER_ID and $$$NAME LIKE '").append(_SEQUENCE_PREFIX).append("%'");

                String sql = buffer.toString();
                QueryLogger.logQuery(sql, Collections.EMPTY_LIST);
                ResultSet rs = sel.executeQuery(sql);
                try {
                    List sequenceList = new ArrayList();
                    while (rs.next()) {
                        sequenceList.add(rs.getString(1));
                    }
                    return sequenceList;
                }
                finally {
                    rs.close();
                }
            }
            finally {
                sel.close();
            }
        }
        finally {
            con.close();
        }
    }

    /**
     * Returns default sequence name for DbEntity.
     */
    protected String sequenceName(DbEntity entity) {

        String entName = entity.getName();
        String seqName = _SEQUENCE_PREFIX + entName;

        if (entity.getSchema() != null && entity.getSchema().length() > 0) {
            seqName = entity.getSchema() + "." + seqName;
        }
        return seqName;
    }

    /**
     * Returns a SQL string needed to drop any database objects associated with automatic
     * primary key generation process for a specific DbEntity.
     */
    protected String dropSequenceString(DbEntity ent) {
        StringBuffer buf = new StringBuffer();
        buf.append("DROP SEQUENCE ").append(sequenceName(ent));
        return buf.toString();
    }

    /**
     * Returns CREATE SEQUENCE statement for entity.
     */
    protected String createSequenceString(DbEntity entity) {
        StringBuffer buf = new StringBuffer();
        buf
                .append("CREATE OR REPLACE SEQUENCE ")
                .append(sequenceName(entity))
                .append(" START WITH 200")
                .append(" INCREMENT BY ")
                .append(getPkCacheSize())
                .append(" NO MAXVALUE ")
                .append(" NO CYCLE ");
        return buf.toString();
    }
}
