 /***************************************************************************
 *                 RELEX, Inc. Linter DBMS source code                      *
 *              This is public source code of RELEX, Inc.                   *
 *                                                                          *
 *      You have a royalty-free right to use, modify, reproduce and         *
 *     distribute the Sample Files (and/or any modified version) in         *
 *        any way you find useful, provided that you agree that             *
 *       RELEX has no warranty obligations or liability for any             *
 *           Sample Application Files which are modified.                   *
 *                                                                          *
 *                 Copyright (c) 1998-2011 RELEX Inc.                       *
 *                                                                          *
 *                   This software was developed by                         *
 *                                                                          *
 *                            RELEX, Inc.                                   *
 *                         Voronezh, Russia                                 *
 *                                                                          *
 *                        All rights reserved.                              *
 ****************************************************************************
 AUTHOR:  Galizin O.V.
 ----------------------------------------------------------------------------
 PURPOSE: Linter control library
 ----------------------------------------------------------------------------
 REVISION HISTORY:
 15.07.08 - Initially created
 01.08.08 - Ermakov - QNX4 build
 22.06.09 - Ermakov - Correct Pid search, Correct variable argument parsing
 23.06.09 - Galizin - Read pipe char by char for prevent missing mesasges
 25.06.09 - Ermakov - Added setcp.sql for execution
 11.01.12 - Glukhov N.A. - Android port
 ----------------------------------------------------------------------------
*/

#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#if !defined(QNX) && !defined(__ANDROID__)
#include <sys/shm.h>
#endif
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <dirent.h>
#include <fcntl.h>
#include <signal.h>
#include "linctrl.h"

#ifdef __LINCTRL_TRACE
#include "log.h"
#else
#define LOG_INIT(LevelString, FileName)
#define TRS_DEFAULT_LEVELS
#define LOG_SHUT
#define LOG_LEVEL 0
#define LOG_DEFAULT_LEVELS(Lev)
#define L_TRACE 1
#define TRS_TRACE while (0) EchoLogLINCTL(
#define LOG_ENABLED(Lev)  ( 0 )
#define LOG_ENABLE
#define LOG_DISABLE
#define TRS_END )
int EchoLogLINCTL( char* format, ...);
#define TRS_FLUSH

#define TRS_PROTO
#define TRS_VOID_PROTO
#define TRS_CALL
#define TRS_VOID_CALL
#define TRS_ENTER
#define TRS_RETURN(str)
#define PID_ENABLE
#define PID_DISABLE
#define TRACE_ENABLE
#define TRACE_DISABLE
#define LOG_FD -1

#define L_ERROR 2
#define TRS_ERROR while (0) EchoLogLINCTL(
#define L_MSG 4
#define TRS_MSG while (0) EchoLogLINCTL(
#endif /* __LINCTRL_TRACE */

static int LinDatabaseStartV(TRS_PROTO char** Argv, char* WorkDir, 
                              int* StdInFd, int* StdOutFd, char** Envs);

static int LinDatabaseStartV(TRS_PROTO char** Args, char* WorkDir, int* StdInFd, int* StdOutFd, char** Envs)
{
  int Pid;
  int i = 0;
  int   Pipe[2]; /* this is a test exec pipe */
  static int   PrevRead = -1;
  int   RetVal;
  
  TRS_ENTER;
  if ( LOG_ENABLED(L_MSG) )
  {
    char** TmpArg = Args;
    while( *TmpArg )
    {
      TRS_MSG "Arg = %s", *TmpArg TRS_END;
      TmpArg++;
    }
  }
  
  /* If the pipe will be closed - process is start sycccessfully 
   On error child process will send code throught it */
  if ( pipe(Pipe) < 0 )
  {
    TRS_ERROR "Proc. pipe() filed. Status %d, %s", errno, strerror(errno) TRS_END;
    TRS_RETURN("FAILED");
    return -1;
  }

  /* for auto close on exec child process */
  if ( fcntl(Pipe[1], F_SETFD, FD_CLOEXEC) < 0 )
  {
    TRS_ERROR "fcntl(F_SETFD, FD_CLOEXEC) filed. Status %d, %s", errno, strerror(errno) TRS_END;
    TRS_RETURN("FAILED");
    return -1;
  }
  
  signal(SIGPIPE, SIG_IGN);

  Pid = fork();
  if ( Pid == 0 )  /* Child */
  {
    close(Pipe[0]);  /* close pipe for reading */

    if ( StdInFd )                        /* 15.02.06 .. */
    {
      if ( StdInFd[0] != 0 )
      {
        dup2(StdInFd[0], 0);
        close(StdInFd[0]);
      }
      close(StdInFd[1]);
    }
    
    if ( StdOutFd )
    {
      if ( StdOutFd[1] != 1)
      {
        dup2(StdOutFd[1], 1);
        close(StdOutFd[1]);
      }
      close(StdOutFd[0]);
    }
    
    if ( Envs )
    {
      while ( *Envs )
      {
        putenv(*Envs);
        Envs++;
      }
    }
    
    if ( WorkDir )
    {
      if ( chdir(WorkDir) )
      {
        int Err = errno;
        
        write(Pipe[1], &Err, sizeof(int));
        _exit(-1);
      }
    }
    
    if ( execvp(Args[0], Args) == -1 )
    {
      int Err = errno;
      write(Pipe[1], &Err, sizeof(int));
    }
    _exit(-1);
  } /* End Child process */

  if ( Pid == -1 )
  {
    TRS_ERROR "fork() failed. Status %d, %s", errno, strerror(errno) TRS_END;
    TRS_RETURN("FAILED");
    close(Pipe[0]);
    close(Pipe[1]);
    return -1;
  }
  
  close(Pipe[1]); /* Close writing pipe */

/*  fcntl(Pipe[0], F_SETFL, O_NONBLOCK);*/

  {
    int Code;
    
    while ( (Code = read(Pipe[0], &RetVal, sizeof(int)) ) != 0 )
    {
      if ( Code == sizeof(int) ) /* Child return error code */
      {
        errno = RetVal;
        close(Pipe[0]);
        TRS_MSG " Error running process %s. status %d, %s", Args[0], RetVal, strerror(RetVal) TRS_END;
        TRS_RETURN("FAILED");
        return -1;
      }
      if ( Code == -1  && errno == EINTR )
        continue;

      close(Pipe[0]);
      TRS_MSG " Error read from process pipe. status %d, %s", errno, strerror(errno) TRS_END;
      TRS_RETURN("FAILED");
      return -1;
    }
    close(Pipe[0]);
  }
  
  TRS_RETURN("SUCCESS");
  return Pid;  
}

static char* FindOutput(TRS_PROTO int Fd, char* Needle, int AdditionalLen) /* 23.06.09 */
{
  static char Buf[1024];
  int NeedleLen;
  int Len;
  int ReadPos = 0;
  
  TRS_ENTER;
  TRS_MSG "Find %s in %d Fd", Needle, Fd TRS_END;
  NeedleLen = strlen(Needle);
  memset(Buf, 0, NeedleLen);
  
  while( (Len = read(Fd, Buf + ReadPos, 1) ) > 0 )          /* 23.06.09 .. */
  {
    if ( Buf[ReadPos] == Needle[ReadPos] )
    {
      ReadPos++;
      if ( ReadPos >= NeedleLen )
      {
        /* Read additional data */
        read(Fd, Buf + ReadPos, AdditionalLen);
        TRS_RETURN(Buf);
        return Buf;
      }
    }
    else
    {
      ReadPos = 0;
    }                                                       /* .. 23.06.09 */
  }
  
  TRS_RETURN("Not Found");
  return NULL;
}

L_LONG LINDBCtrlSetDbPath(t_LINDBCtrl* LINDBCtrl, char* DbPath)
{
  int Len;
  char Buf[1025];
  
  TRS_MSG "Call LINDBCtrlSetDbPath(%s)", DbPath==NULL?"":DbPath TRS_END;
  
  if ( (LINDBCtrl->Flg & LINDB_INIT_MASK) != LINDB_INIT_FLG )
  {
    LINDBCtrl->RetStatus = E_LINCTRL_NOINIT;
    LINDBCtrl->SysStatus = 0;
    TRS_MSG "LINDBCtrlSetDbPath is not init" TRS_END;
    return -1;
  }
  
  if ( DbPath == NULL )
  {
    DbPath = getcwd(Buf, sizeof(Buf)-1);
    TRS_MSG "Use current dir %s ad db directory", DbPath TRS_END;
  }
  
  if ( LINDBCtrl->DbPath )
    free(LINDBCtrl->DbPath);
  
  Len = strlen(DbPath)+1;
  LINDBCtrl->DbPath = (char*)calloc(Len, 1);
  if ( LINDBCtrl->DbPath == NULL )
  {
    LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
    LINDBCtrl->SysStatus = ENOMEM;
    TRS_MSG "End NOMEM" TRS_END;
    return -1;
  }

  memcpy(LINDBCtrl->DbPath, DbPath, Len);
  TRS_MSG "End LINDBCtrlSetDbPath" TRS_END;
  return 0;
}

L_LONG LINDBCtrlInit(t_LINDBCtrl* LINDBCtrl, char* LinterPath)
{
  TRS_MSG "Call LINDBCtrlInit(%s)", LinterPath==NULL?"":LinterPath TRS_END;

  if ( (LINDBCtrl->Flg & LINDB_INIT_MASK) == LINDB_INIT_FLG )
  {
    LINDBCtrl->RetStatus = E_LINCTRL_ALREADYINIT;
    LINDBCtrl->SysStatus = 0;
    TRS_MSG "LINDBCtrlInit LINDBCtrl already init" TRS_END;
    return -1;
  }
  
  if ( LINDBCtrl->LinterPath != NULL )
  {
    free(LINDBCtrl->LinterPath);
    LINDBCtrl->LinterPath = NULL;
  }
  
  if ( LinterPath == NULL )
  {
    char *PathEnv = getenv("PATH");
    char TmpPath[1024];
    char *CurSrc;
    char *CurDst;
    
    if ( PathEnv == NULL )
    {
      LINDBCtrl->RetStatus = E_LINCTRL_GETPATH;
      LINDBCtrl->SysStatus = ENOENT;
      TRS_MSG "LINDBCtrlSetLinterPath no PATH env" TRS_END;
      return -1;
    }
    
    CurSrc   = PathEnv;
    CurDst   = TmpPath;
    while ( *CurSrc )
    {
      if ( *CurSrc == ':' || *(CurSrc+1) == 0 )
      {
        struct stat Stat;
        
        memcpy( CurDst, "/linter", 8);
        if ( stat(TmpPath, &Stat) != -1 )
          break;
        CurDst = TmpPath;
      }
      else
      {
        *CurDst = *CurSrc;
        CurDst++;
      }
      CurSrc++;
    }
    
    if ( *CurSrc == 0 )
    {
      LINDBCtrl->RetStatus = E_LINCTRL_NOTFOUND;
      LINDBCtrl->SysStatus = ENOENT;
      TRS_MSG "LINDBCtrlSetLinterPath no linter in PATH env" TRS_END;
      return -1;
    }
    
    CurDst = 0; /* Cut "linter" word */
    while ( *CurDst != 0 )
    {
      if ( strcpy(CurDst, "/bin") == 0 )
      {
        *CurDst = 0;
        break;
      }
      CurDst++;
    }
    LinterPath = TmpPath;
  }
  {  
    int Len = strlen(LinterPath) + 1;
    struct stat Stat;
    int RetVal;
    
    if ( (RetVal=stat(LinterPath, &Stat)) == -1 || !S_ISDIR(Stat.st_mode) )
    {
      LINDBCtrl->RetStatus = E_LINCTRL_NOTDIR;
      if ( RetVal)
        LINDBCtrl->SysStatus = errno;
      else
        LINDBCtrl->SysStatus = 0;
      TRS_MSG "LINDBCtrlSetLinterPath %s is not dir ", LinterPath TRS_END;
      return -1;
    }
    
    if ( LinterPath[0] != '/' ) /* Path is not absolute */
    {
      char Buf[1025];
      int PwdLen;
      
      Buf[sizeof(Buf)-1] = 0;
      getcwd(Buf, sizeof(Buf)-1);
      PwdLen = strlen(Buf);
      
      LINDBCtrl->LinterPath = (char*)calloc(Len+PwdLen+4, 1);
      if ( LINDBCtrl->LinterPath == NULL )
      {
        LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
        LINDBCtrl->SysStatus = ENOMEM;
        TRS_MSG "LINDBCtrlInit no memory" TRS_END;
        return -1;
      }
      
      strcpy(LINDBCtrl->LinterPath, Buf);
      strcat(LINDBCtrl->LinterPath,"/");
      strcat(LINDBCtrl->LinterPath,LinterPath);
    }
    else
    {
      LINDBCtrl->LinterPath = (char*)calloc(Len, 1);
      if ( LINDBCtrl->LinterPath == NULL )
      {
        LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
        LINDBCtrl->SysStatus = ENOMEM;
        TRS_MSG "LINDBCtrlInit no memory" TRS_END;
        return -1;
      }
  
      memcpy(LINDBCtrl->LinterPath, LinterPath, Len);
    }
  }

  LINDBCtrl->Flg &= ~LINDB_INIT_MASK;
  LINDBCtrl->Flg |= LINDB_INIT_FLG;
  LINDBCtrlSetDbPath(LINDBCtrl, NULL);
  
  TRS_MSG "End LINDBCtrlSetLinterPath" TRS_END;
  return 0;
}


static L_LONG RemoveDatabaseFiles(t_LINDBCtrl* LINDBCtrl)
{
  DIR*  Dir;
  struct dirent* Dirent;
  
  char* DbTmpPath;
  int   DbPathLen;

  TRS_MSG "LinDatabaseControl CREATE DATABASE" TRS_END;
  DbPathLen = strlen(LINDBCtrl->DbPath);
  DbTmpPath = calloc(DbPathLen+32,1 );
  if ( DbTmpPath == NULL )
  {
    LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
    LINDBCtrl->SysStatus = ENOMEM;
    TRS_MSG "LinDatabaseControl no MEM" TRS_END;
    return -1;
  }
  memcpy(DbTmpPath, LINDBCtrl->DbPath, DbPathLen);

  Dir = opendir(LINDBCtrl->DbPath);
  if ( Dir == NULL )
  {
    LINDBCtrl->RetStatus = E_LINCTRL_OPENDIR;
    LINDBCtrl->SysStatus = errno;
    free(DbTmpPath);
    TRS_MSG "LinDatabaseControl opendir. Status %d", errno TRS_END;
    return -1;
  }
  
  while ( (Dirent = readdir(Dir)) != NULL )
  {
    char* EndPtr;

    TRS_MSG "There is file %s in database dir", Dirent->d_name TRS_END;
    if ( (strtol( Dirent->d_name, &EndPtr, 10) != 0 &&
         *EndPtr == '.' && 
         strtol(EndPtr+1, NULL, 10) != 0) || 
         memcmp(Dirent->d_name, "00000000.",9) == 0 || 
         memcmp(Dirent->d_name,"phrase.idx", 10) == 0)
    {
      DbTmpPath[DbPathLen] = '/';
      strcpy(DbTmpPath+DbPathLen+1, Dirent->d_name);
      TRS_MSG "Remove %s", DbTmpPath TRS_END;
      if ( remove(DbTmpPath) != 0 )
      { /* May bee need test for regular file */
        LINDBCtrl->RetStatus = E_LINCTRL_REMOVE;
        LINDBCtrl->SysStatus = errno;
        closedir(Dir);
        free(DbTmpPath);
        TRS_MSG "LinDatabaseControl remove %s. Status %d", Dirent->d_name,errno TRS_END;
        return -1;
      }
    }
  }
  closedir(Dir);
  free (DbTmpPath);
  return 0;
}

L_LONG LinDatabaseControl(t_LINDBCtrl* LINDBCtrl, t_LinDatabaseCommand Command,
                         L_LONG Flags, ...)
{
  va_list VaList;
  char* Arg4;
  int Arg4Len = 0;
  char* Arg5;
  
  TRS_MSG "Call LinDatabaseControl (Command=%d, Flags=0x%X", Command, Flags TRS_END;

  if ( (LINDBCtrl->Flg & LINDB_INIT_MASK) != LINDB_INIT_FLG )
  {
    LINDBCtrl->RetStatus = E_LINCTRL_NOINIT;
    LINDBCtrl->SysStatus = 0;
    return -1;
  }


  LINDBCtrl->RetStatus = 0;
    
  va_start(VaList, Flags);
  Arg4 = va_arg(VaList, char*);
  Arg5 = va_arg(VaList, char*);
  va_end(VaList);
  
  switch ( Command )
  {
    case LIN_DATABASE_CREATE:
    {
      char* CreateCommand;
      char* GendbPath;
      int   GendbPathLen;
      
      if ( Flags & LIN_DATABASE_CREATE_PURGE) /* Remove all database dirs before creating database */
      {
        L_LONG RetVal = RemoveDatabaseFiles(LINDBCtrl);
        if ( RetVal )
          return -1;
      } /* LIN_DATABASE_CREATE_PURGE end */
      
      CreateCommand = (char*)calloc(1024, 1 );
      if ( CreateCommand == NULL )
      {
        LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
        LINDBCtrl->SysStatus = ENOMEM;
        TRS_MSG "LIN_DATABASE_CREATE no MEM" TRS_END;
        return -1;
      }
      GendbPathLen = strlen(LINDBCtrl->LinterPath);
      GendbPath  = (char*)calloc(GendbPathLen+32, 1);
      if ( GendbPath == NULL )
      {
        LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
        LINDBCtrl->SysStatus = ENOMEM;
        free(CreateCommand);
        TRS_MSG "LIN_DATABASE_CREATE no MEM" TRS_END;
        return -1;
      }
      
      strcpy(GendbPath, LINDBCtrl->LinterPath);
      strcat(GendbPath, "/bin/gendb");
      
      strcpy(CreateCommand, "CREATE DATABASE ");
      if ( Flags & LIN_DATABASE_CREATE_IGNORE )
        strcat(CreateCommand, "NO QUERY ");
      if ( Flags & LIN_DATABASE_CREATE_GENDB_ADD)
      {
        if ( Arg4 )                                              /* 22.06.09 */
          Arg4Len = strlen(Arg4);                                /* 22.06.09 */
  
        if ( Arg4Len + strlen(CreateCommand) > 1024 - 2  )
        { /* May bee need test for regular file */
          LINDBCtrl->RetStatus = E_LINCTRL_INVARG;
          LINDBCtrl->SysStatus = EINVAL;
          free(GendbPath);
          free(CreateCommand);
          TRS_MSG "LinDatabaseControl GENDB to long add arg" TRS_END;
          return -1;
        }
        strcat(CreateCommand, Arg4);
      }
      
      {
        char *ArgV[] = {NULL, NULL};
        int   StdIn[2] = {-1,-1};
        int   StdOut[2] = {-1,-1};
        char  Buf[256];
        int   Len;
        int   RetVal;
        
        ArgV[0] = GendbPath;
        if ( pipe(StdIn) || pipe(StdOut) )
        {
          LINDBCtrl->RetStatus = E_LINCTRL_PIPE;
          LINDBCtrl->SysStatus = errno;
          TRS_MSG "LinDatabaseControl gendb. pipe failed. status %d", errno TRS_END;
          goto FreeAll;
        }
        
        if ( (RetVal = LinDatabaseStartV(TRS_CALL ArgV, LINDBCtrl->DbPath, 
                     StdIn, StdOut, NULL)) == -1 )
        {
          LINDBCtrl->RetStatus = E_LINCTRL_RUN;
          LINDBCtrl->SysStatus = errno;
          TRS_MSG "LinDatabaseControl GENDB RUN failed %d", RetVal TRS_END;
          goto FreeAll;
        }
        
        close(StdIn[0]);
        close(StdOut[1]);
        
        while ( write(StdIn[1], CreateCommand, strlen(CreateCommand)) == -1 )
        {
          TRS_MSG "Writing to stdin pipe failed. Status %d", errno TRS_END;
          if ( errno == EINTR )
            continue;
          LINDBCtrl->RetStatus = E_LINCTRL_WRITE;
          LINDBCtrl->SysStatus = errno;
        }
        while ( write(StdIn[1],";\nexit;\n", 7) == -1 )
        {
          TRS_MSG "Writing to stdin pipe failed. Status %d", errno TRS_END;
          if ( errno == EINTR )
            continue;
          LINDBCtrl->RetStatus = E_LINCTRL_WRITE;
          LINDBCtrl->SysStatus = errno;
        }
        close(StdIn[1]);
        {
          char* RetVal;
          if ( (RetVal = FindOutput(TRS_CALL StdOut[0], "GENDB-E", 0)) != NULL ) /* 23.06.09 */
          {
            TRS_MSG "LinDatabaseControl GENDB failed:%s", RetVal TRS_END;
            LINDBCtrl->RetStatus = E_LINCTRL_GENDBFAILED;
            LINDBCtrl->SysStatus = 0;
          }
        }
        close(StdOut[0]);
        free(CreateCommand);
        free(GendbPath);
        
        {
          int Status = 0;
          waitpid(RetVal, &Status, WNOHANG);
          LINDBCtrl->SysStatus = WEXITSTATUS(Status);
        }
        return LINDBCtrl->RetStatus == 0 ? 0:-1;
FreeAll:        
        close(StdOut[0]);
        close(StdOut[1]);
        close(StdIn[0]);
        close(StdIn[1]);
        free(CreateCommand);
        free(GendbPath);
        while ( waitpid(-1, NULL, WNOHANG) > 0 )
          ; /* BLANK */
        return -1;
      }
      break;
    }  
    
    case LIN_DATABASE_RUN: /* Run linter on the database */
    {
      char *Args[64] = {NULL};
      char *Envs[8] = {NULL};
      int  NumArg=0;
      int  NumEnv=0;
      
      if ( LINDBCtrl->LinterPath == NULL )
      {
        LINDBCtrl->RetStatus = -1;
        LINDBCtrl->SysStatus = 0;
        TRS_MSG "Undefined path to linter distributive" TRS_END;
        return -1;
      }
      
      if ( LINDBCtrl->LinterMbx != NULL )
      {
        Envs[NumEnv] = calloc(strlen(LINDBCtrl->LinterMbx) + 16, 1);
        if ( Envs[NumEnv] == NULL )
        {
          LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
          LINDBCtrl->SysStatus = 0;
          return -1;
        }
        strcpy(Envs[NumEnv], "LINTER_MBX=");
        strcat(Envs[NumEnv], LINDBCtrl->LinterMbx);
        NumEnv++;
      }
      
      Args[NumArg] = calloc(strlen(LINDBCtrl->LinterPath)+16,1);
      if ( Args[NumArg] == NULL )
      {
        LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
        LINDBCtrl->SysStatus = 0;
        goto FreeMem;
      }
      strcpy(Args[NumArg], LINDBCtrl->LinterPath);
      strcat(Args[NumArg], "/bin/linter");
      NumArg++;
      
      if ( Flags & LIN_DATABASE_RUN_RDONLY )
      {
        Args[NumArg] = calloc(16,1);
        if ( Args[NumArg] == NULL )
        {
          LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
          LINDBCtrl->SysStatus = 0;
          goto FreeMem;
        }
        strcpy(Args[NumArg], "/RO");
        NumArg++;
      }

      if ( Flags & LIN_DATABASE_RUN_TCP )
      {
        Args[NumArg] = calloc(16,1);
        if ( Args[NumArg] == NULL )
        {
          LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
          LINDBCtrl->SysStatus = 0;
          goto FreeMem;
        }
        strcpy(Args[NumArg], "/TCP");
        NumArg++;
      }

      if ( Flags & LIN_DATABASE_RUN_JDBC )
      {
        Args[NumArg] = calloc(16,1);
        if ( Args[NumArg] == NULL )
        {
          LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
          LINDBCtrl->SysStatus = 0;
          goto FreeMem;
        }
        strcpy(Args[NumArg], "/JDBCS");
        NumArg++;
      }

      if ( LINDBCtrl->TmpPath )
      {
        Args[NumArg] = calloc(strlen(LINDBCtrl->TmpPath ) + 16,1);
        if ( Args[NumArg] == NULL )
        {
          LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
          LINDBCtrl->SysStatus = 0;
          goto FreeMem;
        }
        strcpy(Args[NumArg], "/TMPDIR=");
        strcat(Args[NumArg], LINDBCtrl->TmpPath );
        NumArg++;
      }
      
      if ( Flags & LIN_DATABASE_RUN_ARG )
      {
        char* CurrentArgChar = Arg4;
        char* PrevArgChar = Arg4;
        
        while ( *CurrentArgChar && NumArg < 64 )
        {
          if ( !isspace(*CurrentArgChar) && 
               (isspace(CurrentArgChar[1])|| CurrentArgChar[1] == 0) )
          {
            Args[NumArg] = calloc(CurrentArgChar-PrevArgChar+2,1);
            if ( Args[NumArg] == NULL )
            {
              LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
              LINDBCtrl->SysStatus = 0;
              goto FreeMem;
            }
            memcpy(Args[NumArg], PrevArgChar, CurrentArgChar-PrevArgChar+1);
            Args[NumArg][CurrentArgChar-PrevArgChar+1] = 0;
            NumArg++;
          }
          if ( isspace(*CurrentArgChar) && !isspace(CurrentArgChar[1]) )
          {
            PrevArgChar = CurrentArgChar+1;
          }
          CurrentArgChar++;
        }
      }
      
      {
        int RetVal;
        int StdOutPipe[2] = {-1};
        char* Pid;
        
        if ( pipe(StdOutPipe) != 0 )
        {
          LINDBCtrl->RetStatus = E_LINCTRL_PIPE;
          LINDBCtrl->SysStatus = errno;
          TRS_MSG "pipe failed %d", errno TRS_END;
          goto FreeMem;
        }
        
        if ( (RetVal = LinDatabaseStartV(TRS_CALL Args, LINDBCtrl->DbPath, NULL, 
           StdOutPipe, Envs)) == -1 )
        {
          LINDBCtrl->RetStatus = E_LINCTRL_RUN;
          LINDBCtrl->SysStatus = errno;
          TRS_MSG "Running linter failed" TRS_END;
          goto FreeMem;
        }
        
        close(StdOutPipe[1]);
        Pid = FindOutput(TRS_CALL StdOutPipe[0], ", Pid - ", 10); /* 23.06.09 */
        if (Pid)                                             /* 22.06.09 ... */
          RetVal = strtol(Pid+8, NULL, 10);
        else
          RetVal = 0;                                        /* ... 22.06.09 */
        if ( Pid == NULL ||
             FindOutput(TRS_CALL StdOutPipe[0],
                          "*** RDBMS Linter is running", 0) == NULL ) /* 23.06.09 */
        
        {
          int Status = 0;
          close(StdOutPipe[0]);
          waitpid(RetVal, &Status, WNOHANG);
          
          LINDBCtrl->RetStatus = E_LINCTRL_LINFAILED;
          LINDBCtrl->SysStatus = WEXITSTATUS(Status);
          goto FreeMem;
        }
        
        close(StdOutPipe[0]);
        LINDBCtrl->LinterPid = RetVal;                           /* 22.06.09 */
        LINDBCtrl->Flg &= ~LINDB_ATTACHED;
      }
FreeMem:
      while(NumEnv>=0)
      {
        free(Envs[NumEnv]);
        NumEnv--;
      }
      while(NumArg>=0)
      {
        free(Args[NumArg]);
        NumArg--;
      }
      return (LINDBCtrl->RetStatus?-1:0);
    }

    case LIN_DATABASE_ATTACH: /* Get pid of already started linter */
    {
      L_LONG DigMbx = 20561;
      int RetVal;
      int ShmId;
      L_LONG* Shm;

      TRS_MSG "LIN_DATABASE_ATTACH" TRS_END;

#if defined(QNX) || defined(__ANDROID__)
        LINDBCtrl->RetStatus = E_LINCTRL_INVARG;
        LINDBCtrl->SysStatus = 0;
        TRS_MSG "Not suppoted" TRS_END;
        return -1;
#else
      if ( LINDBCtrl->LinterPath == NULL )
      {
        LINDBCtrl->RetStatus = -1;
        LINDBCtrl->SysStatus = 0;
        TRS_MSG "No init" TRS_END;
        return -1;
      }
      
      if ( LINDBCtrl->LinterMbx != NULL )
        DigMbx = atoi(LINDBCtrl->LinterMbx);
      
      TRS_MSG "LinterMBX = %d", DigMbx TRS_END;
      
      ShmId = shmget(DigMbx, 0, 0);
      if ( ShmId == -1 )
      {
        LINDBCtrl->RetStatus = E_LINCTRL_SHMGET;
        LINDBCtrl->SysStatus = errno;
        TRS_MSG "shmget faild. Status %d", errno TRS_END;
        return -1;
      }
      
      Shm = (L_LONG*)shmat(ShmId, NULL, SHM_RDONLY);
      if ( Shm == (L_LONG*)-1 )
      {
        LINDBCtrl->RetStatus = E_LINCTRL_SHMGET;
        LINDBCtrl->SysStatus = errno;
        TRS_MSG "shmat failed. Status %d", errno);
        return -1;
      }
      
      LINDBCtrl->LinterPid = *Shm;
      
      shmdt(Shm);
      LINDBCtrl->Flg |= LINDB_ATTACHED;
      return 0;
#endif
    }
    
    case LIN_DATABASE_SHUT:
    {
      TRS_MSG "LIN_DATABASE_SHUT" TRS_END;
      if ( LINDBCtrl->LinterPath == NULL )
      {
        LINDBCtrl->RetStatus = -1;
        LINDBCtrl->SysStatus = 0;
        TRS_MSG "No init" TRS_END;
        return -1;
      }
      
      if ( LINDBCtrl->LinterPid == 0 )
      {
        LINDBCtrl->RetStatus = E_LINCTRL_LINPID;
        LINDBCtrl->SysStatus = 0;
        TRS_MSG "No run linter" TRS_END;
        return -1;
      }
      
      if ( !(Flags & LIN_DATABASE_SHUT_IGNOREROLE) && 
            (LINDBCtrl->Flg & LINDB_ATTACHED) )
      {
        LINDBCtrl->RetStatus = E_LINCTRL_ROLE;
        LINDBCtrl->SysStatus = 0;
        TRS_MSG "Attached to RDBMS and try shut without LIN_DATABASE_SHUT_IGNOREROLE" TRS_END;
        return -1;
      }
      
      if ( Flags & LIN_DATABASE_SHUT_BY_SHUT_RUN )
      {
        char *Args[8] = {0};
        char *Envs[3] = {0};
        int NumArg = 0;
        int NumEnv = 0;
        int RetVal;
        int StdOutPipe[2];
        
        TRS_MSG "By run shut command" TRS_END;

        if ( strstr(Arg4, "/") == NULL )
        {
          LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
          LINDBCtrl->SysStatus = 0;
          return -1;
        }
        
        if ( LINDBCtrl->LinterMbx != NULL )
        {
          Envs[NumEnv] = calloc(strlen(LINDBCtrl->LinterMbx) + 16, 1);
          {
            LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
            LINDBCtrl->SysStatus = 0;
            return -1;
          }
          strcpy(Envs[NumEnv], "LINTER_MBX=");
          strcat(Envs[NumEnv], LINDBCtrl->LinterMbx);
          NumEnv++;
        }

/*        if ( LINDBCtrl->TmpPath )
        {
          Envs[NumEnv] = calloc(strlen(LINDBCtrl->TmpPath) + 16, 1);
          if ( Envs[NumEnv] == NULL )
          {
            LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
            LINDBCtrl->SysStatus = 0;
            goto FreeMem;
          }
          strcpy(Envs[NumEnv], "LINTER_TMP=");
          strcat(Envs[NumEnv], LINDBCtrl->TmpPath);
          NumEnv++;
        }*/
        
        Args[NumArg] = calloc(strlen(LINDBCtrl->LinterPath)+16,1);
        if ( Args[NumArg] == NULL )
        {
          free (Envs[0]);
          LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
          LINDBCtrl->SysStatus = 0;
        }
        strcpy(Args[NumArg], LINDBCtrl->LinterPath);
        strcat(Args[NumArg], "/bin/shut");
        NumArg++;
        
        if ( Flags & LIN_DATABASE_SHUT_FORCE )
        {
          Args[NumArg] = "-R";
          NumArg++;
        }
        
        Args[NumArg] = "-U";
        NumArg++;
        
        
        Args[NumArg] = Arg4;
        NumArg++;
        
        
        pipe(StdOutPipe);
        if ( (RetVal = LinDatabaseStartV(TRS_CALL Args, NULL, NULL,  StdOutPipe, Envs)) == -1 )
        {
          LINDBCtrl->RetStatus = E_LINCTRL_RUN;
          LINDBCtrl->SysStatus = 0;
          TRS_MSG "Running shut failed" TRS_END;
          goto FreeShutCom;
        }
        close(StdOutPipe[1]);
        
        FindOutput(TRS_CALL StdOutPipe[0], "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", 0);/* 23.06.09 */
        {
          int Status = 0;
          
          if ( waitpid(RetVal, &Status, 0) != RetVal || 
               Status != 0 ) 
          {
            LINDBCtrl->RetStatus = E_LINCTRL_SHUTFAILED;
            LINDBCtrl->SysStatus = WEXITSTATUS(Status);
            TRS_MSG "Status shut failed. %d", Status TRS_END;
          }
        }
FreeShutCom:        
        close(StdOutPipe[1]);
        free ( Envs[0] );
        free ( Args[0] );
        close(StdOutPipe[0]);
        return LINDBCtrl->RetStatus == 0? 0:-1;
      }
      else /* !run shut but send TERM signal */
      {  
        int Sig;
        
        TRS_MSG "Shut by send signal" TRS_END;
        if ( Flags & LIN_DATABASE_SHUT_FORCE )
        {
          Sig = SIGKILL;
        }
        else
        {
          Sig = SIGTERM;
        }

        if ( kill (LINDBCtrl->LinterPid, Sig ) != 0 )
        {
          LINDBCtrl->RetStatus = E_LINCTRL_KILLFAILED;
          LINDBCtrl->SysStatus = errno;
          TRS_MSG "Kill (sig = %d) failed. Status %d", Sig, errno TRS_END;
          return -1;
        }
        
        return 0;
      }
    }
    
    case LIN_DATABASE_ALIVE:
    {
      TRS_MSG "LIN_DATABASE_ALIVE" TRS_END;
      if ( LINDBCtrl->LinterPath == NULL )
      {
        LINDBCtrl->RetStatus = -1;
        LINDBCtrl->SysStatus = 0;
        TRS_MSG "No init" TRS_END;
        return -1;
      }
      
      if ( LINDBCtrl->LinterPid == 0 )
      {
        LINDBCtrl->RetStatus = E_LINCTRL_LINPID;
        LINDBCtrl->SysStatus = 0;
        TRS_MSG "No run linter" TRS_END;
        return -1;
      }
      
      if ( !(Flags & LIN_DATABASE_ALIVE_SIG) )
      {
        char *Args[2] = {0};
        char *Envs[3] = {0};
        int NumArg = 0;
        int NumEnv = 0;
        int RetVal;
        
        TRS_MSG "By run chklinter command" TRS_END;

        if ( LINDBCtrl->LinterMbx != NULL )
        {
          Envs[NumEnv] = calloc(strlen(LINDBCtrl->LinterMbx) + 16, 1);
          if ( Envs[NumEnv] == NULL )
          {
            LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
            LINDBCtrl->SysStatus = 0;
            return -1;
          }
          strcpy(Envs[NumEnv], "LINTER_MBX=");
          strcat(Envs[NumEnv], LINDBCtrl->LinterMbx);
          NumEnv++;
        }

/*        if ( LINDBCtrl->TmpPath != NULL )
        {
          Envs[NumEnv] = calloc(strlen(LINDBCtrl->TmpPath) + 16, 1);
          {
            LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
            LINDBCtrl->SysStatus = 0;
            return -1;
          }
          strcpy(Envs[NumEnv], "LINTER_TMP=");
          strcat(Envs[NumEnv], LINDBCtrl->LinterMbx);
          NumEnv++;
        }*/
        
        Args[NumArg] = calloc(strlen(LINDBCtrl->LinterPath)+16, 1);
        if ( Args[NumArg] == NULL )
        {
          free (Envs[0]);
          LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
          LINDBCtrl->SysStatus = 0;
        }
        strcpy(Args[NumArg], LINDBCtrl->LinterPath);
        strcat(Args[NumArg], "/bin/chklinter");
        NumArg++;
        
        if ( (RetVal = LinDatabaseStartV(TRS_CALL Args, NULL, NULL,  NULL, Envs)) == -1 )
        {
          LINDBCtrl->RetStatus = E_LINCTRL_RUN;
          LINDBCtrl->SysStatus = 0;
          TRS_MSG "Running chklinter failed" TRS_END;
          goto FreeChk;
        }

        {
          int Status = 0;
          
          if ( waitpid(RetVal, &Status, 0) != RetVal || 
               Status == 0 ) 
          {
            LINDBCtrl->RetStatus = E_LINCTRL_CHKFAILED;
            LINDBCtrl->SysStatus = WEXITSTATUS(Status);
          }
          TRS_MSG "Status chklinter %d", Status TRS_END;
        }
FreeChk:        
        free ( Envs[0] );
        free ( Envs[1] );
        free ( Args[0] );
        return LINDBCtrl->RetStatus == 0? 0:-1;
      }
      else /* check by signal */
      {  
        TRS_MSG "By send signal" TRS_END;
        if ( kill (LINDBCtrl->LinterPid, 0 ) != 0 )
        {
          LINDBCtrl->RetStatus = E_LINCTRL_KILLFAILED;
          LINDBCtrl->SysStatus = errno;
          TRS_MSG "Kill (0) failed. Status %d", errno TRS_END;
          return -1;
        }
        return 0;
      }
    }
    case LIN_DATABASE_DICT:
    {
      char* SqlList[] = { 
#if _VER_MAX>= 600
        "cstables.sql",
        "charsets.sql",
        "mbcps.sql",
        "setcp.sql",                                        /* 25.06.09 */
/*      "descmb.sql",*/                                     /* 22.06.09 */
#endif
        "systab.sql",
        "distr.sql",
        "security.sql",
        "extsec.sql",
        "catalog.sql",
/*      "ora_cat.sql", */                                   /* 22.06.09 */
        "inkernel.sql",
        "cerrors.sql",
        "search.sql",
        "default.sql",
        NULL
        };
        char **CurrentSql;
        char *Args[16] = {0};
        char *Envs[2] = {0};
        int NumArg = 0;
        int RetVal;
        char* FullSqlFile;
        
        TRS_MSG "Create dict command" TRS_END;

        if ( LINDBCtrl->LinterMbx != NULL )
        {
          Envs[0] = calloc(strlen(LINDBCtrl->LinterMbx) + 16, 1);
          if ( Envs[0] == NULL )
          {
            LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
            LINDBCtrl->SysStatus = 0;
            return -1;
          }
          strcpy(Envs[0], "LINTER_MBX=");
          strcat(Envs[0], LINDBCtrl->LinterMbx);
        }
        
        Args[NumArg] = calloc(strlen(LINDBCtrl->LinterPath)+16, 1);
        if ( Args[NumArg] == NULL )
        {
          free (Envs[0]);
          LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
          LINDBCtrl->SysStatus = 0;
        }
        strcpy(Args[NumArg], LINDBCtrl->LinterPath);
        strcat(Args[NumArg], "/bin/inl");
        NumArg++;
        
        Args[NumArg] = "-u";
        NumArg++;
        
        Args[NumArg] = Arg4;
        NumArg++;

        Args[NumArg] = "-f";
        NumArg++;

        if ( Flags & LIN_DATABASE_DICT_NAME )
        {
          SqlList[0] = Arg5;
          SqlList[1] = NULL;
        }

        CurrentSql = SqlList;
        while(*CurrentSql)
        {
          int Len = strlen(*CurrentSql);
          
          TRS_MSG "Handle %s sql file", *CurrentSql TRS_END;
          
          if ( strstr(*CurrentSql, "/") != NULL )
          {
            FullSqlFile = calloc(Len + 1, 1);
            if ( FullSqlFile == NULL )
            {
              LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
              LINDBCtrl->SysStatus = 0;
              TRS_MSG "MEM" TRS_END;
              goto FreeInl;
            }
            memcpy(FullSqlFile, *CurrentSql, Len);
          }
          else
          {
            FullSqlFile = calloc( Len + strlen(LINDBCtrl->LinterPath) + 16, 1);
            if ( FullSqlFile == NULL )
            {
              LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
              LINDBCtrl->SysStatus = 0;
              TRS_MSG "MEM" TRS_END;
              goto FreeInl;
            }
            strcpy(FullSqlFile, LINDBCtrl->LinterPath);
            strcat(FullSqlFile, "/dict/");
            strcat(FullSqlFile, *CurrentSql);
          }
          
          Args[NumArg] = FullSqlFile;
          
          RetVal = LinDatabaseStartV(TRS_CALL Args, NULL, NULL,  NULL, Envs);
          free(FullSqlFile);
          if ( RetVal == -1 )
          {
            LINDBCtrl->RetStatus = E_LINCTRL_RUN;
            LINDBCtrl->SysStatus = errno;
            TRS_MSG "Running inl failed" TRS_END;
            goto FreeInl;
          }

          {
            int Status = 0;
          
            if ( waitpid(RetVal, &Status, 0) != RetVal || 
                 Status != 0 ) 
            {
              TRS_MSG "Status inl %d", Status TRS_END;
              LINDBCtrl->RetStatus = E_LINCTRL_INL;
              LINDBCtrl->SysStatus = WEXITSTATUS(Status);
              goto FreeInl;
            }
          }

          CurrentSql++;
        } /* While */
        
        if ((Flags & LIN_DATABASE_DICT_NAME) == 0)
        {
          NumArg = 0;
          strcpy(Args[NumArg], LINDBCtrl->LinterPath);
          strcat(Args[NumArg], "/bin/loarel");
          NumArg++;
          
          Args[NumArg] = "-s";
          NumArg++;

          Args[NumArg] = "0";
          NumArg++;
          
          Args[NumArg] = "-u";
          NumArg++;
          
          Args[NumArg] = Arg4;
          NumArg++;
          
          Args[NumArg] = "-t";
          NumArg++;
          
          Args[NumArg] = "ERRORS";
          NumArg++;
          
          Args[NumArg] = "-f";
          NumArg++;

          FullSqlFile = calloc( strlen(LINDBCtrl->LinterPath) + 32, 1);
          if ( FullSqlFile == NULL )
          {
            LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
            LINDBCtrl->SysStatus = 0;
            TRS_MSG "MEM" TRS_END;
            goto FreeInl;
          }
          strcpy(FullSqlFile, LINDBCtrl->LinterPath);
          strcat(FullSqlFile, "/dict/errors.lod");
          Args[NumArg] = FullSqlFile;
          Args[NumArg+1] = NULL;

          RetVal = LinDatabaseStartV(TRS_CALL Args, NULL, NULL,  NULL, Envs);
          free(FullSqlFile);
          if ( RetVal == -1 )
          {
            LINDBCtrl->RetStatus = E_LINCTRL_RUN;
            LINDBCtrl->SysStatus = errno;
            TRS_MSG "Running loarel failed" TRS_END;
            goto FreeInl;
          }

          {
            int Status = 0;
          
            if ( waitpid(RetVal, &Status, 0) != RetVal || 
                 Status != 0 ) 
            {
              TRS_MSG "Status loarel %d", Status TRS_END;
              LINDBCtrl->RetStatus = E_LINCTRL_LOAREL;
              LINDBCtrl->SysStatus = WEXITSTATUS(Status);
              goto FreeInl;
            }
          }
        }
        
        
FreeInl:  
        free ( Args[0] );
        free ( Envs[0] );
        return LINDBCtrl->RetStatus == 0? 0:-1;
    }
  }
  return LINDBCtrl->RetStatus == 0? 0:-1;
}

L_LONG LINDBCtrlSetLinterMbx(t_LINDBCtrl* LINDBCtrl, char* LinterMbx)
{
  TRS_MSG "LINDBCtrlSetLinterMbx %s", LinterMbx);
  
  if ( (LINDBCtrl->Flg & LINDB_INIT_MASK) != LINDB_INIT_FLG )
  {
    TRS_MSG "LINDBCtrlSetLinterMbx LinterMbx==NULL" TRS_END;
    LINDBCtrl->RetStatus = E_LINCTRL_NOINIT;
    LINDBCtrl->SysStatus = 0;
    return -1;
  }
  
  if ( LinterMbx == NULL )
  {
    TRS_MSG "LINDBCtrlSetLinterMbx LinterMbx==NULL" TRS_END;
    LINDBCtrl->RetStatus = E_LINCTRL_INVARG;
    LINDBCtrl->SysStatus = 0;
    return -1;
  }
  
  if ( LINDBCtrl->LinterMbx != NULL )
    free(LINDBCtrl->LinterMbx);
  
  LINDBCtrl->LinterMbx = calloc(strlen(LinterMbx)+1, 1);
  if ( LINDBCtrl->LinterMbx == NULL )
  {
    LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
    LINDBCtrl->SysStatus = 0;
    return -1;
  }
  
  strcpy(LINDBCtrl->LinterMbx, LinterMbx);
  TRS_MSG "End LINDBCtrlSetLinterMbx" TRS_END;
  return 0;
}

L_LONG LINDBCtrlSetTmpPath(t_LINDBCtrl* LINDBCtrl, char* TmpPath)
{
  TRS_MSG "Call LINDBCtrlSetTmpPath %s", TmpPath);
  
  if ( (LINDBCtrl->Flg & LINDB_INIT_MASK) != LINDB_INIT_FLG )
  {
    TRS_MSG "LINDBCtrlSetTmpPath is not init LINDBCtrl" TRS_END;
    LINDBCtrl->RetStatus = E_LINCTRL_NOINIT;
    LINDBCtrl->SysStatus = 0;
    return -1;
  }
  
  if ( TmpPath == NULL )
  {
    TRS_MSG "LINDBCtrlSetTmpPath LinterMbx==NULL" TRS_END;
    LINDBCtrl->RetStatus = E_LINCTRL_INVARG;
    LINDBCtrl->SysStatus = 0;
    return -1;
  }

  if ( LINDBCtrl->TmpPath != NULL )
    free(LINDBCtrl->TmpPath);
  
  LINDBCtrl->TmpPath = calloc(strlen(TmpPath)+1, 1);
  if ( LINDBCtrl->TmpPath == NULL )
  {
    LINDBCtrl->RetStatus = E_LINCTRL_NOMEM;
    LINDBCtrl->SysStatus = 0;
    return -1;
  }
  
  strcpy(LINDBCtrl->TmpPath, TmpPath);
  TRS_MSG "End LINDBCtrlSetTmpPath" TRS_END;
  return 0;
}

L_LONG LINDBCtrlUninit(t_LINDBCtrl* LINDBCtrl)
{
  if ( LINDBCtrl->DbPath )
    free(LINDBCtrl->DbPath);
  if ( LINDBCtrl->LinterPath)
    free(LINDBCtrl->LinterPath);
  if ( LINDBCtrl->LinterMbx)
    free( LINDBCtrl->LinterMbx);
  if ( LINDBCtrl->TmpPath )
    free(LINDBCtrl->TmpPath );
  memset(LINDBCtrl, 0, sizeof(t_LINDBCtrl));
  return 0;
}
