<?php
//
// +----------------------------------------------------------------------+
// | PHP version 4.0                                                      |
// +----------------------------------------------------------------------+
// | Copyright (c) 1997, 1998, 1999, 2000 The PHP Group                   |
// +----------------------------------------------------------------------+
// | This source file is subject to version 2.02 of the PHP license,      |
// | that is bundled with this package in the file LICENSE, and is        |
// | available at through the world-wide-web at                           |
// | http://www.php.net/license/2_02.txt.                                 |
// | If you did not receive a copy of the PHP license and are unable to   |
// | obtain it through the world-wide-web, please send a note to          |
// | license@php.net so we can mail you a copy immediately.               |
// +----------------------------------------------------------------------+
// | Authors: Alex Kononyhin                                              |
// |          Artem Gluhov                                                |
// +----------------------------------------------------------------------+
// | 28.02.14 -- DV -- Return error after commit/rollback problems        |
// +----------------------------------------------------------------------+
//
// Database independent query interface definition for PHP's Linter SQL
// extension.
//

require_once 'DB/common.php';

class DB_linter extends DB_common {

  var $connection;
  var $phptype, $dbsyntax;
  var $real_query;
  var $last_query;
  var $linerr, $syserr;
  var $last_cursor;
  var $rownums = array();
  var $dsn = array();
  var $autocommit = true;
  var $param_info = array(); // holds info about binded params for each prepared query

function DB_linter() {
  //printf("debug linter::DB_linter\n");
  $this->phptype = 'linter';
  $this->dbsyntax = 'linter';
//  $this->phptype = 'odbc';
//  $this->dbsyntax = 'sql92';
  $this->features = array(
      'prepare' => true,
      'pconnect' => false,
      'transactions' => true,
      'numrows' => true,
      'limit' => 'alter',
  );

  $this->errorcode_map = array(
    903  => DB_ERROR_CONSTRAINT,
    905  => DB_ERROR_CONSTRAINT,
    906  => DB_ERROR_CONSTRAINT,
    1001 => DB_ERROR_CONNECT_FAILED,
    1016 => DB_ERROR_DIVZERO,
    1503 => DB_ERROR_ALREADY_EXISTS,
    1508 => DB_ERROR_ALREADY_EXISTS,
    1513 => DB_ERROR_CANNOT_DROP,
    2022 => DB_ERROR_INVALID_NUMBER,
    2023 => DB_ERROR_INVALID_NUMBER,
    2032 => DB_ERROR_INVALID_DATE,
    2050 => DB_ERROR_SYNTAX,
    2051 => DB_ERROR_SYNTAX,
    2100 => DB_ERROR_CANNOT_CREATE,
    2132 => DB_ERROR_CANNOT_CREATE,
    2133 => DB_ERROR_CANNOT_DROP,
    2150 => DB_ERROR_ALREADY_EXISTS,
    2202 => DB_ERROR_NOSUCHTABLE,
    2207 => DB_ERROR_NOSUCHFIELD,
    2210 => DB_ERROR_NOSUCHFIELD,
    2217 => DB_ERROR_VALUE_COUNT_ON_ROW,
    2218 => DB_ERROR_VALUE_COUNT_ON_ROW,
    2219 => DB_ERROR_INVALID_NUMBER,
    2220 => DB_ERROR_CONSTRAINT_NOT_NULL,
    2221 => DB_ERROR_INVALID,
    2267 => DB_ERROR_NOT_FOUND,
    2353 => DB_ERROR_NOSUCHTABLE
  );

  $this->connection = 0;
  $this->linerr = $this->syserr = 0;
  $this->last_cursor = 0;
}

function setError($cursor = 0) {
  //printf("debug linter::setError\n");
  $err_text = '';
  if ($cursor != 0) {
    $this->linerr = linter_last_error($cursor, LINTER_ERROR);
    $this->syserr = linter_last_error($cursor, SYSTEM_ERROR);
    $err_text = Linter_Error_Msg($cursor);
  } else {
    if ($this->connection != 0) {
      $this->linerr = linter_last_error($this->connection, LINTER_ERROR);
      $this->syserr = linter_last_error($this->connection, SYSTEM_ERROR);
      $err_text = Linter_Error_Msg($this->connection);
    } else {
      $this->linerr = linter_last_error(LINTER_ERROR);
      $this->syserr = linter_last_error(SYSTEM_ERROR);
      $err_text = "";
    }
  }

  if ($this->linerr >= 2000 && $this->linerr < 3000) // SQL syntax errors
  {
    $row = $this->syserr & 0xFFFF;
    $pos = $this->syserr >> 16;
    $err_text = $err_text . sprintf(". Syntax error at row %d, position %d", $row, $pos);
  }

  $errno = $this->errorCode($this->linerr);
  return $this->raiseError($errno, null, null, null,
                           $this->linerr . ' ** ' . $err_text);
}

function connect(&$dsn, $persistent = false) {
  //printf("debug linter::connect\n");

  if (!PEAR::loadExtension('linter')) {
    return $this->raiseError(DB_ERROR_EXTENSION_NOT_FOUND);
  }
  
  if (is_array($dsn)) {
    $dsninfo = &$dsn;
  } else {
    $dsninfo = DB::parseDSN($dsn);
  }
  if (!$dsninfo/* || !$dsninfo['phptype']*/) {
    return $this->raiseError(); // XXX ERRORMSG
  }
  $this->dsn = $dsninfo;

  $db = $dsninfo['database'] ? $dsninfo['database'] : '';
  $user = $dsninfo['username'];
  $pw = $dsninfo['password'];
  $mode = isset($dsninfo['mode']) ? $dsninfo['mode'] : TM_EXCLUSIVE;
  $conn = @linter_open_connect($user, $pw, $db, $mode);
  if ($conn < 0) {
    return $this->setError();
  }
  $this->connection = $conn;
  return DB_OK;
}

function disconnect() {
  //printf("debug linter::disconnect\n");
  @linter_close_connect($this->connection);
}

/**
 * Send a query to Linter SQL and return the results as a Linter SQL resource
 * identifier.
 *
 * @param $query the SQL query
 *
 * @access public
 *
 * @return int returns a valid Linter SQL cursor if opened.
     * A DB error is  returned on failure.
 */
function simpleQuery($query, $cursor = 0) {
  //printf("debug linter::simpleQuery, $query\n");

  $this->last_cursor = 0;
  $this->last_query = $query;

  if(empty($cursor)) {
    $cursor = @linter_open_cursor($this->connection);
    if ($cursor < 0) {
      return $this->setError();
    }
  }
  $this->last_cursor = $cursor;

  $this->linerr = 0;
  $query = $query . ";";
  $result = @linter_exec_direct($cursor, $query);
  if ($result != LPE_SUCCESS) {
    $linerr = linter_last_error($cursor, LINTER_ERROR);
    if ( $linerr != 2) //no data
      return $this->setError($cursor);
  }

  // Determine which queries that should return data, and which
  // should return an error code only.
  if (preg_match('/(SELECT)/i', $query)) {
    $this->rownums[$cursor] = 0;
    return $cursor;
  }

  if ($this->autocommit)
    return $this->commit();
  else
    return DB_OK;
}


/**
 * Fetch a row and insert the data into an existing array.
 *
 * @param $result Linter SQL result identifier
 * @param $arr (reference) array where data from the row is stored
 * @param $fetchmode how the array data should be indexed
 *
 * @access public
 *
 * @return int DB_OK on success, a DB error on failure
 */
function fetchInto($result, &$arr, $fetchmode = DB_FETCHMODE_DEFAULT, $rownum = null) {
  //printf("debug linter::fetchInto, rownum = '$rownum'\n");

  if ($rownum === null)
    $rownum = $this->rownums[$result];

  if (linter_fetch($result, FETCH_ABSNUM, $rownum + 1) < 0)
    if($this->errorNative($result) == 2)
      return NULL;
    else
      return $this->setError($result);

  $this->rownums[$result]++; 

  if ($fetchmode == DB_FETCHMODE_DEFAULT) {
    $fetchmode = $this->fetchmode;
  }
  if ($fetchmode & DB_FETCHMODE_ASSOC) {
    $arr = @linter_get_data_array($result);
  } else {
    $arr = @linter_get_data_row($result);
  }

  if ($arr < 0) {
    return $this->setError($result);
  }
  
  for ($i = 0; $i < sizeof($arr); $i++) {
    if ($this->is_null($result, $i+1))
      $arr[$i] = '';
  }

  if ($this->options['portability'] & DB_PORTABILITY_LOWERCASE && $arr)
    $arr = array_change_key_case($arr, CASE_LOWER);
  if ($this->options['portability'] & DB_PORTABILITY_RTRIM)
    $this->_rtrimArrayValues($arr);
  if ($this->options['portability'] & DB_PORTABILITY_NULL_TO_EMPTY)
    $this->_convertNullArrayValuesToEmpty($arr);

  return DB_OK;
}

  
function SeekRow($cursor, $rownum) {
  if (linter_fetch($result, FETCH_ABSNUM, $rownum + 1) < 0) {
    return $this->setError($result);
  }
  return DB_OK;
}


/**
 * Free the internal resources associated with $result.
 *
 * @param $result Linter SQL result identifier or DB statement identifier
 *
 * @access public
 *
 * @return bool TRUE on success, FALSE if $result is invalid
 */
function freeResult($result) {
  $ret = linter_close_cursor($result);
  if($ret < 0)
    return false;
  else
    return true;
}


/**
 * Get the native error code of the last error (if any) that
 * occured on the current connection.
 *
 * @access public
 *
 * @return int native Linter SQL error code
 */
function errorNative($cursor) {
  return linter_last_error($cursor, LINTER_ERROR);
}


/**
 * Prepares a query for multiple execution with execute().
 *
 * @return int identifier for this prepared query
 */
function prepare($query) {
  //printf("debug linter::prepare $query\n");

  $this->last_query = $query;

  $cursor = @linter_open_cursor($this->connection);
  if ($cursor < 0) {
    return $this->setError();
  }
  $this->last_cursor = $cursor;      
  $query = $query . ";";

  $tokens = preg_split('/((?<!\\\)[&?!])/', $query, -1,
                        PREG_SPLIT_DELIM_CAPTURE);
  $param_info = array();
  $token    = 0;
  $newquery = '';

  foreach ($tokens as $key => $val) {
    if (($val == '?') || ($val == '!') || ($val == '&')) {
      $newquery .= '?';
      if ($val == '?')
        $param_info[$token++] = 0;
      if ($val == '&')
        $param_info[$token++] = 1;
      if ($val == '!')
        $param_info[$token++] = 2;
    } else {
      $val = preg_replace('/\\\([&?!])/', "\\1", $val);
      $newquery .= $val;
    }
  }
  
  $ret = @linter_prepare($cursor, $newquery);
  if ($ret < 0) {
    return $this->setError($cursor);
  }
  $this->real_query = $query;
  $this->param_info[$cursor] = $param_info;
  return $cursor;
}


function freePrepared($cursor, $free_resource = true)
{
//  printf("debug linter::freePrepared\n");
  if ($free_resource)
  {
    if (Linter_Close_Cursor($cursor) < 0)
      return false;

     unset($this->param_info[$cursor]);
  }
  return true;
}


/**
 * Executes a prepared query.
 */
function &execute($cursor, $params = array()) {
  //printf("debug linter::execute cursorid = $cursor\n");
  $this->last_cursor = 0;
  $this->linerr = 0;

  if (is_array($params)) {
    for( $i = 0; $i < count($params); $i++)
    {
      if ($this->param_info[$cursor][$i] == 1)
        $params[$i] = file_get_contents($params[$i]);
      if ($this->param_info[$cursor][$i] == 2)
      {
        $tmp = $params[$i];
        if ((($tmp[0] == "'") && ($tmp[strlen($tmp)-1] == "'")) ||
            (($tmp[0] == '"') && ($tmp[strlen($tmp)-1] == '"'))
           )
          $tmp = substr($tmp, 1, strlen($tmp) - 2);
        $tmp = str_replace("''","'", $tmp);
        $params[$i] = $tmp;
      }
    }
  }
  else {
    if ($this->param_info[$cursor][0] == 1)
      $params = file_get_contents($params);
  }

  $result = linter_execute($cursor, array_values((array)$params));

  if ($result < 0) {
    $linerr = linter_last_error($cursor, LINTER_ERROR);
    if ($linerr != 2) { //no data
      $tmp = & $this->setError($cursor);
      return $tmp;
    }
  }

  $this->last_cursor = $cursor;
  if (preg_match('/(SELECT|SHOW)/i', $this->real_query)) {
    $this->pushErrorHandling(PEAR_ERROR_RETURN);
    $tmp =& new DB_result($this, $cursor);
    $this->popErrorHandling();
    
    $this->rownums[$cursor] = 0;
    return $tmp;
  } else {
    $tmp = 0;
    if ($this->autocommit)
      $tmp = $this->commit();
    else
      $tmp = DB_OK;
    return $tmp;
  }
}


function autoCommit($onoff = false) {
  //printf("debug linter::autocommit $onoff\n");
  $this->autocommit = $onoff;
}


function commit($cursor = 0) {
  //printf("debug linter::commit\n");
  if(empty($cursor))
    $cursor = $this->connection;
  if (@linter_exec_direct($cursor, 'commit;') != LPE_SUCCESS)   /* 28.02.14 */
      return $this->setError($cursor);                          /* 28.02.14 */
  return DB_OK;
}


function rollback($cursor = 0) {
  //printf("debug linter::rollback\n");
  if(empty($cursor))
    $cursor = $this->connection;
  if (@linter_exec_direct($cursor, 'rollback;') != LPE_SUCCESS) /* 28.02.14 */
      return $this->setError($cursor);                          /* 28.02.14 */
  return DB_OK;
}


/**
 * Get the number of columns in a result set.
 *
 * @param $result Linter SQL result identifier
 *
 * @access public
 *
 * @return int the number of columns per row in $result
 */
function numCols($result) {
  $cols = linter_get_cursor_opt($result, CO_COL_COUNT);
  if ($cols < 0) {
    return $this->setError($result);
  }
  return $cols;
}


function numRows($cursor) {
  $rows = linter_get_cursor_opt($cursor, CO_ROW_COUNT);
  if ($rows < 0) {
    return $this->setError($cursor);
  }
  return $rows;
}


function is_null($cursor, $column) {
  $col_obj = linter_get_col_prop($cursor, $column);
  if($col_obj < 0) {
    return $this->setError($cursor);
  } else {
    return($col_obj->is_null);
  }
}


function is_blob($cursor, $column) {
  $col_obj = linter_get_col_prop($cursor, $column);
  if($col_obj < 0) {
    return $this->setError($cursor);
  } else {
    if($col_obj->type == LDT_BLOB)
      return TRUE;
    else
      return FALSE;
  }
}


function ColName($cursor, $column) {
  $col_obj = linter_get_col_prop($cursor, $column);
  if($col_obj < 0) {
    return $this->setError($cursor);
  } else {
    return($col_obj->name);
  }
}


function get_cursor_opt($cursor, $option) {
  $value = linter_get_cursor_opt($cursor, $option);
  if($value < 0) {
    return $this->setError($cursor);
  } else {
    return $value;
  }
}


function set_cursor_opt($cursor, $option, $value) {
  $res = linter_set_cursor_opt($cursor, $option, $value);
  if($res < 0) {
    return false;
  } else {
    return true;
  }
}


function append_blob($cursor, $blob, $colnum = '', $b_convert = false, $blobtype=0) {
  if( $b_convert )
    $blob = convert_cyr_string($blob, "w", "a");
  $blobsize = strlen($blob);
  if(!empty($colnum)) {
    if(!in_array("linter_blob_add", get_extension_funcs("linter"))) {
      return $this->raiseError(DB_ERROR_UNSUPPORTED);
    }
    if(in_array("linter_blob_add_ex", get_extension_funcs("linter"))) {
      $ret = @linter_blob_add_ex($cursor, $blob, $blobsize, $colnum, $blobtype);
    } else {
      $ret = @linter_blob_add($cursor, $blob, $blobsize, $colnum);
    }
    if($ret < 0)
      return $this->setError($cursor);
  } else {
    if(in_array("linter_blob_append_ex", get_extension_funcs("linter"))) {
      $ret = @linter_blob_append_ex($cursor, $blob, $blobsize, $blobtype);
    } else {
      $ret = @linter_blob_append($cursor, $blob, $blobsize);
    }
    if($ret < 0)
      return $this->setError($cursor);
  }
  return(1);
}


function get_blob($cursor, $colnum = '', $b_convert = false) {
  if(!empty($colnum)) {
    if(!in_array("linter_blob_fetch", get_extension_funcs("linter")))
      return $this->raiseError(DB_ERROR_UNSUPPORTED);

   $size = @linter_blob_get_size($cursor, $colnum);
   if($size < 0)
     return $this->setError($cursor);
   $ret = @linter_blob_fetch($cursor, 1, $size, $colnum);
   if($ret < 0)
     return $this->setError($cursor);
  } else {
    $size = @linter_blob_get_size($cursor);
    if($size < 0)
      return $this->setError($cursor);
    $ret = @linter_blob_get_data($cursor, 1, $size);
    if($ret < 0)
      return $this->setError($cursor);
  }
  $blob = $ret;
  if( $b_convert )
    $blob = convert_cyr_string($blob, "a", "w");
  return($blob);
}


function clear_blob($cursor, $colnum = '') {
  if(!empty($colnum)) {
    if(!in_array("linter_blob_purge", get_extension_funcs("linter")))
      return $this->raiseError(DB_ERROR_UNSUPPORTED);
    $ret = @linter_blob_purge($cursor, $colnum);
    if($ret < 0)
      return $this->setError($cursor);
  } else {
    $ret = @linter_blob_clear($cursor);
    if($ret < 0)
      return $this->setError($cursor);
  }
  return(1);
}


// }}}
// {{{ nextId()

/**
 * Returns the next free id in a sequence
 *
 * @param string  $seq_name  name of the sequence
 * @param boolean $ondemand  when true, the seqence is automatically
 *                            created if it does not exist
 *
 * @return int  the next id number in the sequence.
 *               A DB_Error object on failure.
 *
 * @see DB_common::nextID(), DB_common::getSequenceName(),
 *      DB_odbc::createSequence(), DB_odbc::dropSequence()
 */
function nextId($seq_name, $ondemand = true)
{
  //printf("debug linter::nextId, $ondemand\n");
  $seqname = $this->getSequenceName($seq_name);

  $this->pushErrorHandling(PEAR_ERROR_RETURN);
  $result = $this->query("select ${seqname}.nextval");
  $this->popErrorHandling();

  if (DB::isError($result))
  {
    if (($ondemand == true)&& ($result->getCode() == DB_ERROR_NOSUCHTABLE))
    {
      $result = $this->createSequence($seq_name);
      if ($result != DB_OK)
        return $result;
      $result = $this->query("select ${seqname}.nextval");
      if (DB::isError($result))
        return $result;
    }
    else
    {
      $this->raiseError($result);
      return $result;
    }
  }

  $row = $result->fetchRow(DB_FETCHMODE_ORDERED);
  if (DB::isError($row || !$row))
    return $row;
  return $row[0] / 3; //strange behavior of linter sequences
}


/**
 * Creates a new sequence
 *
 * @param string $seq_name  name of the new sequence
 *
 * @return int  DB_OK on success.  A DB_Error object on failure.
 *
 * @see DB_common::createSequence(), DB_common::getSequenceName(),
 *      DB_odbc::nextID(), DB_odbc::dropSequence()
 */
function createSequence($seq_name)
{
  //printf("debug linter::createSequence\n");
  return $this->query('CREATE OR REPLACE SEQUENCE '
                      . $this->getSequenceName($seq_name)
                      . " START WITH 1 INCREMENT BY 1"
                     );
}


// }}}
// {{{ dropSequence()

/**
 * Deletes a sequence
 *
 * @param string $seq_name  name of the sequence to be deleted
 *
 * @return int  DB_OK on success.  A DB_Error object on failure.
 *
 * @see DB_common::dropSequence(), DB_common::getSequenceName(),
 *      DB_odbc::nextID(), DB_odbc::createSequence()
 */
function dropSequence($seq_name)
{
  //printf("debug linter::dropSequence\n");
  return $this->query('DROP SEQUENCE ' . $this->getSequenceName($seq_name));
}


// }}}
// {{{ affectedRows()

/**
 * Determines the number of rows affected by a data maniuplation query
 *
 * 0 is returned for queries that don't manipulate data.
 *
 * @return int  the number of rows.  A DB_Error object on failure.
 */
function affectedRows()
{
  //printf("debug linter::affectedRows\n");
  if (DB::isManip($this->last_query)) {
    $rows = linter_get_cursor_opt($this->last_cursor, CO_ROW_COUNT);
    if ($rows < 0)
      return $this->setError($this->last_cursor);
    else
      return $rows;
  } else
    return 0;
}


// }}}
// {{{ modifyLimitQuery()

/**
 * Adds LIMIT clauses to a query string according to current DBMS standards
 *
 * @param string $query   the query to modify
 * @param int    $from    the row to start to fetching (0 = the first row)
 * @param int    $count   the numbers of rows to fetch
 * @param mixed  $params  array, string or numeric data to be used in
 *                         execution of the statement.  Quantity of items
 *                         passed must match quantity of placeholders in
 *                         query:  meaning 1 placeholder for non-array
 *                         parameters or 1 placeholder per array element.
 *
 * @return string  the query string with LIMIT clauses added
 *
 * @access protected
 */
function modifyLimitQuery($query, $from, $count, $params = array())
{
  //printf("debug linter::modifyLimitQuery\n");
  if (DB::isManip($query)) {
    return $query . " LIMIT $count";
  } else {
    return $query . " LIMIT $from, $count";
  }
}


// }}}
// {{{ tableInfo()

/**
 * Returns information about a table or a result set
 *
 * @param object|string  $result  DB_result object from a query or a
 *                                 string containing the name of a table.
 *                                 While this also accepts a query result
 *                                 resource identifier, this behavior is
 *                                 deprecated.
 * @param int            $mode    a valid tableInfo mode
 *
 * @return array  an associative array with the information requested.
 *                 A DB_Error object on failure.
 *
 * @see DB_common::tableInfo()
 */


function tableInfo($result, $mode = null)
{
  //printf("debug linter::tableInfo\n");
  $con = 0;
  if (is_string($result)) {
    $result = $this->query("select * from $result");
    if (DB::isError($result))
      return $result;
    $con = $result->result;
  }
  elseif (isset($result->result)) {
    $con = $result->result;
  }
  else
    return $this->raiseError(DB_ERROR_NEED_MORE_DATA);

  if ($this->options['portability'] & DB_PORTABILITY_LOWERCASE) {
      $case_func = 'strtolower';
  } else {
      $case_func = 'strval';
  }
   
  $count = Linter_Get_Cursor_Opt($con, CO_COL_COUNT);
  $res   = array();

  if ($mode) {
      $res['num_fields'] = $count;
  }

  for ($i = 0; $i < $count; $i++) {
    $tmp = Linter_Get_Col_Prop($con, $i + 1);
    $res[$i] = array(
        'table' => $case_func($tmp->table),
        'name'  => $case_func($tmp->name),
        'type'  => $tmp->type,
        'len'   => $tmp->length,
        'flags' => $tmp->is_null?'not_null':'',
    );
    if ($mode & DB_TABLEINFO_ORDER) {
        $res['order'][$res[$i]['name']] = $i;
    }
    if ($mode & DB_TABLEINFO_ORDERTABLE) {
        $res['ordertable'][$res[$i]['table']][$res[$i]['name']] = $i;
    }
  }
  return $res;
}


// }}}
// {{{ getSpecialQuery()

/**
 * Obtains the query string needed for listing a given type of objects
 *
 * @param string $type  the kind of objects you want to retrieve
 *
 * @return string  the SQL query string or null if the driver doesn't
 *                  support the object type requested
 *
 * @access protected
 * @see DB_common::getListOf()
 */
function getSpecialQuery($type)
{
  //printf("debug linter::tableInfo\n");
  switch ($type) {
    case 'tables':
        return "SELECT TABLE_NAME from TABLES where TABLE_TYPE = 'TABLE'";
    case 'views':
        return "SELECT TABLE_NAME from TABLES where TABLE_TYPE = 'VIEW'";
    case 'users':
        return "SELECT USER_NAME from USER_USERS";
    default:
        return null;
  }
}

// }}}


}




// Local variables:
// tab-width: 4
// c-basic-offset: 4
// End:
?>
