 /***************************************************************************
 *                 RELEX, Inc. Linter Python driver source code             *
 *        This is unpublished proprietary source code of RELEX, Inc.        *
 *     The copyright notice does not evidence any actual or intended        *
 *                  publication of such source code.                        *
 *                                                                          *
 *                 Copyright (c) 2003-2013 RELEX Inc.                       *
 *                                                                          *
 *                   This software was developed by                         *
 *                                                                          *
 *                            RELEX, Inc.                                   *
 *                         Voronezh, Russia                                 *
 *                                                                          *
 *                        All rights reserved.                              *
 ****************************************************************************
 AUTHOR: Boris Fedorov
 ----------------------------------------------------------------------------
 PURPOSE: common type arrays manipulations
 ----------------------------------------------------------------------------
 REVISION HISTORY:
 31.10.03 -- Boris Fedorov (Removed C++ style comments, disable unicode 
 support for Python older whan 2.2)
 29.07.08 -- Ermakov (Corrected warnings, added Boolean type)
 04.09.08 -- Ermakov (ObjectDel instead of MemDel)
 28.11.08 -- Ermakov (Correct ver type handling. Added NCHAR support)
 25.12.13 -- Muratova (Python 3 support)
*/

#undef _DEBUG
#include <Python.h>
#include <linapi.h>
#include "LinPy.h"

static int Array_InternalBind(t_Array *);

PyTypeObject g_ArrayType = {                                 /* 25.12.13 ... */
    PyVarObject_HEAD_INIT(NULL, 0)
    "Array",           // tp_name
    sizeof(t_Array),                    // tp_basicsize
    0,                                  // tp_itemsize
    (destructor) Array_Free,            // tp_dealloc
    0,                                  // tp_print
    0,                                  // tp_getattr
    0,                                  // tp_setattr
    0,                                  // tp_compare
    0,                                  // tp_repr
    0,                                  // tp_as_number
    0,                                  // tp_as_sequence
    0,                                  // tp_as_mapping
    0,                                  // tp_hash
    0,                                  // tp_call
    0,                                  // tp_str
    0,                                  // tp_getattro
    0,                                  // tp_setattro
    0,                                  // tp_as_buffer
    Py_TPFLAGS_DEFAULT,                 // tp_flags
    0,                                  // tp_doc
#if PY_VERSION_HEX >= 0x02060000	
    0,                                  // tp_traverse
    0,                                  // tp_clear
    0,                                  // tp_richcompare
    0,                                  // tp_weaklistoffset
    0,                                  // tp_iter
    0,                                  // tp_iternext
    0,                                  // tp_methods
    0                                   // tp_members
#endif
};                                                           /* ... 25.12.13 */

/**
 * Allocate a new Array.
 */
t_Array *Array_New(
  t_Cursor *cursor,
  unsigned numElements,
  t_ArrayType *type,
  t_ParamDesc *desc)
{
  unsigned long dataLength;
  t_Array *array;
  unsigned i;
  
  array = PyObject_NEW(t_Array, type->pythonType);
  if (array == NULL)
    return NULL;

  memcpy(&array->desc, desc, sizeof(t_ParamDesc));
  array->index = 0;
  array->boundName = NULL;
  array->boundCursor = NULL;
  array->allocatedElements = numElements;
  array->actualElements = 0;
  array->type = type;
  array->indicator = NULL;
  array->data = NULL;
  array->actualLength = NULL;
  array->maxLength = desc->Length;
  array->blobPortion = -1;

#if (PY_VERSION_HEX >= 0x02020000)                           /* 28.11.08 ... */
  if (desc->Type == tNChar || desc->Type == tNVarChar)
    array->maxLength *= sizeof(L_WORD);
#endif                                                       /* ... 28.11.08 */

  if (type->isVariableLength)
    array->maxLength += sizeof(L_WORD);
  
  array->actualLength = PyMem_Malloc(numElements * sizeof(L_LONG));
  if (array->actualLength == NULL)
  {
    PyErr_NoMemory();
    Py_DECREF(array);
    return NULL;
  }

  dataLength = (unsigned long) numElements * (unsigned long) array->maxLength;
  if (dataLength > INT_MAX)
  {
    PyErr_SetString(PyExc_ValueError, "array size too large");
    Py_DECREF(array);
    return NULL;
  }

  array->indicator = PyMem_Malloc(numElements * sizeof(char));
  if (array->indicator == NULL)
  {
    PyErr_NoMemory();
    Py_DECREF(array);
    return NULL;
  }

  array->data = PyMem_Malloc(dataLength);
  if (array->data == NULL)
  {
    PyErr_NoMemory();
    Py_DECREF(array);
    return NULL;
  }

  for (i = 0; i < numElements; i++)
  {
    array->indicator[i] = 1;
  }

  if (array->type->initializeProc)
  {
    if ((*array->type->initializeProc)(array, cursor) < 0)
    {
      Py_DECREF(array);
      return NULL;
    }
  }

  return array;
}


/**
 * Create parameter Array for cursor statement
 */
extern t_Array *Array_NewParameter(
  t_Cursor *cursor,
  int index,
  int arraySize)
{
  t_Array *var;
  t_ArrayType *varType;
  t_ParamDesc desc;
  L_LONG length;
  int status;

  trace(("Array_NewParameter: cursor %08x index %i\n", cursor, index));

  length = sizeof(t_ParamDesc);
  status = LINTER_GetStatementOption(
    cursor->statementHandle,
    sParamDesc,
    (L_SWORD)index,
    &desc,
    &length);
  if (Cursor_CheckForError(cursor, status) < 0)
    return NULL;

  varType = Array_TypeByDataType(desc.Type);
  if (!varType)
    return NULL;

  var = Array_New(cursor, arraySize, varType, &desc);
  if (!var)
    return NULL;

  var->index = index;

  return var;
}


/**
 * Free an existing Array.
 */
void Array_Free(
  t_Array *var)
{
  trace(("Array_Free: var %08x\n", var));
  if (var->type->finalizeProc)
    (*var->type->finalizeProc)(var);
  if (var->indicator)
    PyMem_Free(var->indicator);
  if (var->data)
    PyMem_Free(var->data);
  if (var->actualLength)
    PyMem_Free(var->actualLength);
  Py_XDECREF(var->boundName);
  PyObject_DEL(var);                                             /* 04.09.08 */
}


/**
 * Resize the Array.
 */
int Array_Resize(
  t_Array *var,
  unsigned maxLength)
{
  char *newData;
  int i;

  newData = PyMem_Malloc(var->allocatedElements * maxLength);
  if (!newData)
  {
    PyErr_NoMemory();
    return -1;
  }

  for (i = 0; i < var->allocatedElements; i++)
    memcpy(newData + maxLength * i, var->data + var->maxLength * i, var->maxLength);

  PyMem_Free(var->data);
  var->data = newData;
  var->maxLength = maxLength;

  /* force rebinding */
  if ((var->boundCursor) && (Array_InternalBind(var) < 0))
    return -1;

  return 0;
}


/**
 * Returns a boolean indicating if the object is a Array.
 */
int Array_Check(
  PyObject *object)
{
  if (!object)
    return 0;
  return ((Py_TYPE(object) == &g_StringType) ||
          (Py_TYPE(object) == &g_BinaryType) ||
          (Py_TYPE(object) == &g_TimestampType) ||
          (Py_TYPE(object) == &g_NumberType));
}


/**
 * Return a Array type given a Python type object or NULL if the Python
 * type does not have a corresponding Array type.
 */
t_ArrayType *Array_TypeByPythonType(
  PyObject *type)
{
  if (type == (PyObject*) &g_StringType)
    return &vt_VarChar;
  if (type == (PyObject*) &g_NumberType)
    return &vt_Integer;
  PyErr_SetString(g_NotSupportedErrorException,
    "Array_TypeByPythonType: unhandled data type");
  return NULL;
}


/**
 * Return a Array type given a Python object or NULL if the Python
 * object does not have a corresponding Array type.
 */
static t_ArrayType *Array_TypeByValue(
  PyObject *value)
{
  PyObject *elementValue;
  int i;

  if (value == Py_None)
    return &vt_VarChar;

  if (PyString_Check(value))                               /* 26.11.13 */
    return &vt_VarChar;

  if (PyInt_Check(value))
    return &vt_Integer;
  
  if (PyFloat_Check(value))
    return &vt_Double;

  if (PyList_Check(value))
  {
    elementValue = Py_None;
    for (i = 0; i < PyList_GET_SIZE(value); i++)
    {
      elementValue = PyList_GET_ITEM(value, i);
      if (elementValue != Py_None)
        break;
    }
    return Array_TypeByValue(elementValue);
  }

  PyErr_Format(g_NotSupportedErrorException, 
    "Array_TypeByValue: unhandled data type %s",
    Py_TYPE(value)->tp_name);
  return NULL;
}


/**
 * Return a Array type given data type or NULL if
 * data type does not have a corresponding Array type.
 */
t_ArrayType *Array_TypeByDataType(
  int dataType)
{
  switch (dataType)
  {
    case tChar:
      return &vt_Char;
    case tVarChar:
      return &vt_VarChar;
    case tByte:
      return &vt_Byte;
    case tVarByte:
      return &vt_VarByte;
    case tSmallInt:
      return &vt_SmallInt;
    case tInteger:
      return &vt_Integer;
    case tBigInt:
      return &vt_BigInt;
    case tBoolean:                                          /* 29.07.08 */
      return &vt_Boolean;
    case tReal:
      return &vt_Real;
    case tDouble:
      return &vt_Double;
    case tTimeStamp:
      return &vt_Timestamp;
    case tDecimal:
      return &vt_Decimal;
#if (PY_VERSION_HEX >= 0x02020000)                         /* 31.10.03 */
    case tNChar:
      return &vt_NChar;
    case tNVarChar:
      return &vt_NVarChar;
#else                                                      /* 31.10.03 ... */
    case tNChar:
      return &vt_Char;
    case tNVarChar:
      return &vt_VarChar;
#endif                                                     /* ... 31.10.03 */
    case tBlob:
      return &vt_Blob;
    case tExtFile:
      return &vt_VarChar;
  }
  PyErr_Format(g_NotSupportedErrorException,
    "Array_TypeByDataType: unhandled data type %d",
    dataType);
  return NULL;
}


/**
 * Bind array data to cursor
 */
int Array_InternalBindAnswer(
  t_Array *array,
  t_Cursor *cursor)
{
  int status;
  
  status = LINTER_BindAnswer(
    (L_WORD)((cursor->handleFromSP) ? cursor->handleFromSP : cursor->handle),
    (cursor->handleFromSP) ? 0 : cursor->statementHandle,   /* 26.11.13 */
    (L_SWORD)array->index,
    array->data,
    array->indicator,
    array->maxLength,
    (L_SWORD)array->type->linterType,
    array->maxLength,
    array->desc.Prec,
    array->desc.Scale,
    array->actualLength);
  
  if (Cursor_CheckForError(cursor, status) < 0)
    return -1;

  return 0;
}


/**
 * Allocate a Array and define it for the given statement.
 */
t_Array *Array_Define(
  t_Cursor *cursor,
  unsigned numElements,
  unsigned column)
{
  t_ArrayType *varType;
  t_Array *array;
  t_ParamDesc desc;
  L_LONG length;
  int status;
  
  length = sizeof(t_ParamDesc);
  status = LINTER_GetCursorOption(
    (L_WORD)((cursor->handleFromSP) ? cursor->handleFromSP : cursor->handle),
    cAnswerDesc,
    (L_SWORD)column,
    &desc,
    &length);
  if (Cursor_CheckForError(cursor, status) < 0)
    return NULL;

  varType = Array_TypeByDataType(desc.Type);
  if (!varType)
    return NULL;

  array = Array_New(cursor, numElements, varType, &desc);
  if (array == NULL)
    return NULL;
    
  array->index = column;
  
  if (array->type->bindAnswerProc != NULL)
  {
    status = array->type->bindAnswerProc(array, cursor);
    if (status < 0)
    {
      Py_DECREF(array);
      return NULL;
    }
  }

  return array;
}


/**
 *
 */
int Array_GetData(
  t_Array *array,
  t_Cursor *cursor)
{
  int status;

  status = LINTER_GetData(
    cursor->handle,
    (L_SWORD)array->index,
    array->data,
    array->maxLength,
    (L_SWORD)array->type->linterType,
    0,
    0,
    array->actualLength);
  
  if (Cursor_CheckForError(cursor, status) < 0)
    return -1;

  if (array->actualLength[0] == -1)
  {
    array->indicator[0] = 1;
    array->actualLength[0] = 0;
  }
  else
  {
    array->indicator[0] = 0;
  }

  return 0;
}


/**
 * Allocate a Array and bind it to the given statement.
 */
static int Array_InternalBind(
  t_Array *var)
{
  int status;
  t_Cursor *cursor;
  PyObject * val;
  
  cursor = var->boundCursor;

  if (var->boundName)                                       /* 26.11.13 .. */ /* */
  { 
    val = (PyString_Check(var->boundName)) ? var->boundName : (PyUnicode_Check(var->boundName)) ?
	   PyUnicode_AsEncodedString(var->boundName, NULL, NULL) : NULL;
  }
  
  status = LINTER_BindParameter(
    (L_SWORD)((cursor->handleFromSP) ? cursor->handleFromSP : cursor->handle),
    cursor->statementHandle,
    (L_SWORD)var->index,

    (var->boundName) ? PyString_AS_STRING(val): NULL,        /* 26.11.13 */ /* */
    var->indicator,
    var->data,
    var->allocatedElements,
    var->maxLength,
    (L_SWORD)(var->type->linterType | pInput),
    &var->actualElements,
    var->actualLength);

  if (Cursor_CheckForError(var->boundCursor, status) < 0)
    return -1;

  return 0;
}

/**
 * Allocate a Array and bind it to the given statement.
 */
int Array_BindNamedParameter(
  t_Array *var,
  t_Cursor *cursor,
  PyObject *name)
{
  PyObject *val;
  
  if ((cursor == var->boundCursor) &&
      (name == var->boundName))
    return 0;

  val = (PyString_Check(name)) ? name : (PyUnicode_Check(name)) ? 
    PyUnicode_AsEncodedString(name, NULL, NULL) : NULL;
	  
  trace(("Array_BindNamedParameter: var %08x cursor %08x name \"%s\"\n",
    var, cursor, PyString_AS_STRING(val)));

  Py_XDECREF(var->boundName);
  var->boundName = name;
  Py_INCREF(var->boundName);
  var->boundCursor = cursor;
  var->index = 0;

  if (Array_InternalBind(var) < 0)
  {
    Py_DECREF(var->boundName);
    var->boundName = NULL;
    var->boundCursor = NULL;
    return -1;
  }

  return 0;
}


/**
 * Allocate a Array and bind it to the given statement.
 */
int Array_BindQmarkParameter(
  t_Array *var,
  t_Cursor *cursor,
  int index)
{
  if ((cursor == var->boundCursor) &&
      (index == var->index))
    return 0;

  trace(("Array_BindQmarkParameter: var %08x cursor %08x index %i\n",
    var, cursor, var->index));

  Py_XDECREF(var->boundName);
  var->boundName = NULL;
  var->index = index;
  var->boundCursor = cursor;

  if (Array_InternalBind(var) < 0)
  {
    var->boundCursor = NULL;
    var->index = 0;
    return -1;
  }

  return 0;
}


/**
 * Return the value of the Array.
 */
PyObject *Array_GetValue(t_Array *array, unsigned arrayPos)
{
  if (arrayPos >= (unsigned)array->allocatedElements)        /* 26.11.13 */
  {
    PyErr_SetString(PyExc_IndexError, "Array_GetValue: array size exceeded");
    return NULL;
  }
  return (*array->type->getValueProc)(array, arrayPos);
}


/**
 * Set the value of the Array.
 */
int Array_SetValue(t_Array *array, unsigned arrayPos, PyObject *value)
{
  if (arrayPos >= (unsigned)array->allocatedElements)
  {
    PyErr_SetString(PyExc_IndexError, "Array_SetValue: array size exceeded");
    return -1;
  }
  return (*array->type->setValueProc)(array, arrayPos, value);
}
