 /***************************************************************************
 *                 RELEX, Inc. Linter Python driver source code             *
 *        This is unpublished proprietary source code of RELEX, Inc.        *
 *     The copyright notice does not evidence any actual or intended        *
 *                  publication of such source code.                        *
 *                                                                          *
 *                 Copyright (c) 2003-2014 RELEX Inc.                       *
 *                                                                          *
 *                   This software was developed by                         *
 *                                                                          *
 *                            RELEX, Inc.                                   *
 *                         Voronezh, Russia                                 *
 *                                                                          *
 *                        All rights reserved.                              *
 ****************************************************************************
 AUTHOR: Boris Fedorov
 ----------------------------------------------------------------------------
 PURPOSE: BLOB type interface
 ----------------------------------------------------------------------------
 REVISION HISTORY:
 31.10.03 -- Boris Fedorov (Remove C++ style comments)
 29.07.08 -- Ermakov (Corrected warnings)
 04.09.08 -- Ermakov (ObjectDel instead of MemDel)
 25.12.13 -- Muratova (Python 3 support)
 03.10.14 -- Kononov (Fixed binding blob data as string for prepared query)
 14.10.14 -- Kononov (Implemented function setoutputsize() for setting up fetch size of BLOB columns or one of them)
*/

#undef _DEBUG
#include <Python.h>
#include <linapi.h>
#include "LinPy.h"
#include "Blob.h"

static int Blob_Initialize(t_Array*, t_Cursor*);
static void Blob_Free(t_Array *self);
#if PY_VERSION_HEX < 0x02060000
  static PyObject *Blob_GetAttr(t_Array*, PyObject*);
#else
  static int Blob_GetType(t_Array *array);
  static int Blob_GetLength(t_Array *array);
#endif
static PyObject *Blob_Read(t_Array*, PyObject*);
static PyObject *Blob_Write(t_Array*, PyObject*);
static PyObject *Blob_Clear(t_Array*, PyObject*);
static int Blob_SetValue(t_Array *array, unsigned pos, PyObject *value);
static PyObject *Blob_GetValue(t_Array *var, unsigned pos);

static PyObject *Blob_ReadEx(t_Array *array, int pos, int length); /* 14.10.14 */
static int Blob_AfterPosChanged(t_Array *array);

static PyMethodDef g_BlobMethods[] =                         /* 25.12.13 ... */
{
  {"read", (PyCFunction) Blob_Read, METH_VARARGS },
  {"write", (PyCFunction) Blob_Write, METH_VARARGS },
  {"clear", (PyCFunction) Blob_Clear, METH_VARARGS },
  {NULL, NULL}
};

#if PY_VERSION_HEX >= 0x02060000                             
static PyGetSetDef g_BlobCalcMembers[] = {
    { "type", (getter) Blob_GetType, 0, 0, 0 },
    { "length", (getter) Blob_GetLength, 0, 0, 0 },
    { NULL }
};                                                           /* ... 25.12.13 */
#endif

PyTypeObject g_BlobType =
{
	  PyVarObject_HEAD_INIT(NULL, 0)                         /* 25.12.13 */
	  "BLOB",			                     /*tp_name*/
	  sizeof(t_Array),				         /*tp_basicsize*/
	  0,									 /*tp_itemsize*/
	  (destructor)Blob_Free,                 /*tp_dealloc*/
	  0,									 /*tp_print*/
	  0,									 /*tp_getattr*/
	  0,									 /*tp_setattr*/
	  0,									 /*tp_compare*/
	  0,									 /*tp_repr*/
	  0,									 /*tp_as_number*/
	  0,									 /*tp_as_sequence*/
	  0,									 /*tp_as_mapping*/
	  0,									 /*tp_hash */
	  0,									 /*tp_call*/
	  0,									 /*tp_str*/
  #if PY_VERSION_HEX < 0x02060000
   (getattrofunc) Blob_GetAttr,    /* tp_getattro */
  #else
	  0,                                     /*tp_getattro*/ /* 25.12.13 ... */
  #endif
	  0,									 /*tp_setattro*/
	  0,									 /*tp_as_buffer*/
	  Py_TPFLAGS_DEFAULT,					 /*tp_flags*/
    0         /* tp_doc */
  #if PY_VERSION_HEX >= 0x02060000
	 , 0,									 /* tp_traverse */
	  0,									 /* tp_clear */
	  0,									 /* tp_richcompare */
	  0,									 /* tp_weaklistoffset */
	  0,									 /* tp_iter */
	  0,									 /* tp_iternext */
	  g_BlobMethods,						 /* tp_methods */
	  0,									 /* tp_members */
	  g_BlobCalcMembers,			         /* tp_getset */
	  0,									 /* tp_base */
	  0,									 /* tp_dict */
	  0,									 /* tp_descr_get */
	  0,									 /* tp_descr_set */
	  0,									 /* tp_dictoffset */
	  0,									 /* tp_init */   /* ... 25.12.13 */
   #endif 
};

t_ArrayType vt_Blob =
{
  (InitializeProc) Blob_Initialize,
  (FinalizeProc) NULL,
  (SetValueProc) Blob_SetValue,
  (GetValueProc) Blob_GetValue,
  NULL,
  NULL,
  Blob_AfterPosChanged,
  &g_BlobType,
  tBlob,
  4000,
  0
};


/**
 *
 */
static int Blob_Initialize(t_Array *array, t_Cursor *cursor)
{
  t_Blob *blob;

  blob = (t_Blob *) PyMem_Malloc(sizeof(t_Blob));
  if (blob == NULL)
  {
    PyErr_NoMemory();
    return -1;
  }

  blob->cursor = cursor;
  blob->length = 0;
  blob->type = 0;

  array->data = (void*) blob;

  return 0;
}


/**
 * Destructor
 */
static void Blob_Free(t_Array *array)
{
  t_Blob *data = (t_Blob *) array->data;
  if (data != NULL)
    PyMem_Free(data);
  PyObject_DEL(array);                                           /* 04.09.08 */
}


static int Blob_Add(t_Array *array, char *buf, int length)
{
  t_Blob *blob = (t_Blob *) array->data;
  int status;

  if (Cursor_IsOpen(blob->cursor) < 0)
    return -1;

  status = LINTER_AddBlob(blob->cursor->handle, (L_SWORD)array->index, (L_WORD)blob->type, buf, length, NULL, NULL);
  if (Cursor_CheckForError(blob->cursor, status) < 0)
    return -1;

  return 0;
}


static int Blob_Purge(t_Array *array)
{
  t_Blob *blob = (t_Blob *) array->data;
  int status;

  if (Cursor_IsOpen(blob->cursor) < 0)
    return -1;

  status = LINTER_PurgeBlob(blob->cursor->handle, (L_SWORD)array->index, NULL, NULL);
  if (Cursor_CheckForError(blob->cursor, status) < 0)
    return -1;

  return 0;
}


static int Blob_Fetch(t_Array *array, char *buf, int pos, int length, int *outLength)
{
  t_Blob *blob = (t_Blob *) array->data;
  int status;
  L_LONG len;

  if (Cursor_IsOpen(blob->cursor) < 0)
    return -1;

  len = length;

  status = LINTER_FetchBlob(
    blob->cursor->handle,
    (L_SWORD)array->index,
    pos,
    &len,
    buf,
    NULL,
    NULL);

  if (Cursor_CheckForError(blob->cursor, status) < 0)
  {
    PyMem_Free(buf);
    return -1;
  }

  *outLength = length;

  return 0;
}


static int Blob_GetInfo(t_Array *array)
{
  t_Blob *blob = (t_Blob *) array->data;
  int status;
  L_LONG len;

  if (Cursor_IsOpen(blob->cursor) < 0)
    return -1;

  len = sizeof(int);

  status = LINTER_GetCursorOption(
    blob->cursor->handle, 
    cBlobLength, 
    (L_SWORD)array->index, 
    &blob->length, 
    &len);

  if (Cursor_CheckForError(blob->cursor, status) < 0)
    return -1;

  len = sizeof(int);

  status = LINTER_GetCursorOption(
    blob->cursor->handle,
    cBlobType,
    (L_SWORD)array->index,
    &blob->type,
    &len);

  if (Cursor_CheckForError(blob->cursor, status) < 0)
    return -1;

  return 0;
}


/**
 * Called after cursor change position
 */
static int Blob_AfterPosChanged(t_Array *array)
{
  return Blob_GetInfo(array);
}


/**
 * Set blob value
 */
static int Blob_SetValue(t_Array *array, unsigned pos, PyObject *value)
{
  long length;
  char *data;
  L_BOOL isUnicode;                                                       /* 03.10.14 ... */
  PyObject *val;

  isUnicode = PyUnicode_Check(value);
  if (!PyString_Check(value) && !isUnicode)
  {
    PyErr_SetString(PyExc_TypeError, "expecting string or unicode data"); /* ... 03.10.14 */
    return -1;
  }

  if (isUnicode)                                                          /* 03.10.14 ... */
  {
#ifdef Py_UNICODE_WIDE
    val = PyUnicode_AsUTF8String(value);
#else
    val = PyUnicode_AsEncodedString(value, NULL, NULL);
#endif
  }
  
  length = PyString_GET_SIZE(isUnicode ? val : value);                   /* ... 03.10.14 */

  if (length == 0)
  {
    array->indicator[pos] = 1;
  }
  else
  {
    array->indicator[pos] = 0;
    if (length > array->maxLength)
    {
      if (Array_Resize(array, length) < 0)
        return -1;
    }
    data = Array_ELEMENT(array, pos);
    memcpy(data, PyString_AS_STRING(isUnicode ? val : value), length);   /* 03.10.14 */
    array->actualLength[pos] = length;
  }

  return 0;
}


/**
 * Get blob value
 */
static PyObject *Blob_GetValue(t_Array *array, unsigned pos)
{
  if (array->blobPortion != -1)                        /* 14.10.14 ... */
  {
    return Blob_ReadEx(array, 1, array->blobPortion);
  }                                                    /* ... 14.10.14 */
  Py_INCREF(array);
  return (PyObject*) array;
}

#if PY_VERSION_HEX < 0x02060000
static PyObject *Blob_GetAttr(t_Array *array, PyObject *nameObject)
{
  char *name;
  t_Blob *blob = (t_Blob*) array->data;

  name = PyString_AS_STRING(nameObject);
  if ((name[0] == 'l') && (strcmp(name, "length") == 0))
  {
    return PyInt_FromLong(blob->length);
  }
  else if ((name[0] == 't') && (strcmp(name, "type") == 0))
  {
    return PyInt_FromLong(blob->type);
  }
  return Py_FindMethod(g_BlobMethods, (PyObject*) array, name);
}
#else
static int Blob_GetType(t_Array *array)                      /* 25.12.13 ... */
{
  t_Blob *blob = (t_Blob*) array->data;
  
  return (int)PyInt_FromLong(blob->type);
}

static int Blob_GetLength(t_Array *array)
{
  t_Blob *blob = (t_Blob*) array->data;
  
  return (int)PyInt_FromLong(blob->length);
}                                                            /* ... 25.12.13 */
#endif


/**
 * Read blob portion
 *
 * Syntax:
 * buffer = blob.read([pos[, length]])
 */
static PyObject *Blob_Read(t_Array *array, PyObject *args)        /* 14.10.14 */
{
  int pos;
  int length;
  t_Blob *blob;

  pos = 1;
  blob = (t_Blob *) array->data;
  length = ((t_Blob*)array->data)->length;

  if (!args || !PyArg_ParseTuple(args, "|ii", &pos, &length))
    return NULL;

  return Blob_ReadEx(array, pos, length);
}


static PyObject *Blob_ReadEx(t_Array *array, int pos, int length) /* 14.10.14 */
{
#if PY_VERSION_HEX < 0x02060000
  PyObject *bufferObject;
#endif
  t_Blob *blob = (t_Blob *) array->data;
  int outLength = 0;
  char *buf;

  if (pos < 1)
  {
    PyErr_SetString(g_InterfaceErrorException, "Blob_Read: read position must be 1 or greater");
    return NULL;
  }

  if (length < 0)
  {
    PyErr_SetString(g_InterfaceErrorException, "Blob_Read: read size must be positive");
    return NULL;
  }

  if (pos > blob->length)
    length = 0;

  if (length > (blob->length - pos + 1))
    length = blob->length - pos + 1;

  if (length <= 0)
  {
    Py_INCREF(Py_None);
    return Py_None;
  }

  buf = PyMem_Malloc(length);
  if (buf == NULL)
  {
    PyErr_NoMemory();
    return NULL;
  }

  if (Blob_Fetch(array, buf, pos, length, &outLength) < 0)
  {
    PyMem_Free(buf);
    return NULL;
  }

#if PY_VERSION_HEX < 0x02060000
  bufferObject = PyBuffer_FromMemory(buf, outLength);
  
  if (bufferObject == NULL)
  {
    PyMem_Free(buf);
    return NULL;
  }

  return bufferObject;
#else
  return PyString_FromStringAndSize(buf, outLength);
#endif

#if PY_MAJOR_VERSION > 3
     return PyUnicode_Decode(buf, outLength, NULL, NULL);   /* ... 25.12.13 */
#endif  
}


/**
 *
 */
static PyObject *Blob_Write(t_Array *array, PyObject *args)
{
  PyObject *object;
#if PY_VERSION_HEX < 0x02060000
  PyBufferProcs *pb;
#endif
  char *buf;
  int length;

  object = NULL;
  if (!args || !PyArg_ParseTuple(args, "O", &object))
    return NULL;
  
  if (object == Py_None)
  {
    if (Blob_Purge(array) < 0)
      return NULL;
    Py_INCREF(Py_None);
    return Py_None;
  }

#if PY_VERSION_HEX >= 0x02060000
  if (PyUnicode_Check(object))                               /* 25.12.13 ... */
  {
    object = PyUnicode_AsEncodedString(object, NULL, NULL);
  }
  
  buf  = PyString_AS_STRING(object);
  length = PyString_GET_SIZE(object);                         /* ... 25.12.13 */
#else
  pb = Py_TYPE(object)->tp_as_buffer;
  if ((pb == NULL) || (pb->bf_getreadbuffer == NULL))
  {
    PyErr_SetString(
      g_InterfaceErrorException,
      "Blob_Write: object to write must have buffer interface");
    return NULL;
  }
    
  length = pb->bf_getreadbuffer(object, 0, (void**) &buf);
  if (length == -1)
    return NULL;
  if (Blob_Purge(array) < 0)
    return NULL;
#endif
  if (Blob_Add(array, buf, length) < 0)
    return NULL;  
  
  if (Blob_GetInfo(array) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


/**
 * Clear blob.
 */
static PyObject *Blob_Clear(t_Array *array, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ""))
    return NULL;

  if (Blob_Purge(array) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}
