 /***************************************************************************
 *                 RELEX, Inc. Linter Python driver source code             *
 *        This is unpublished proprietary source code of RELEX, Inc.        *
 *     The copyright notice does not evidence any actual or intended        *
 *                  publication of such source code.                        *
 *                                                                          *
 *                 Copyright (c) 2003-2014 RELEX Inc.                       *
 *                                                                          *
 *                   This software was developed by                         *
 *                                                                          *
 *                            RELEX, Inc.                                   *
 *                         Voronezh, Russia                                 *
 *                                                                          *
 *                        All rights reserved.                              *
 ****************************************************************************
 AUTHOR: Boris Fedorov
 ----------------------------------------------------------------------------
 PURPOSE: Connection interface (DBAPI)
 ----------------------------------------------------------------------------
 REVISION HISTORY:
 31.10.03 -- Boris Fedorov (Remove C++ style comments)
 04.09.08 -- Ermakov (ObjectDel instead of MemDel)
 28.11.08 -- Ermakov (Added channel pool)
 25.12.13 -- Muratova (Python 3 support)
 30.09.14 -- Kononov (Removed an unreferenced local variables, fixed compiler warnings)
 20.10.14 -- Kononov (Fixed the processing of unsuccessful connection)
*/

#undef _DEBUG
#include <Python.h>
#include <linapi.h>
#include "LinPy.h"

#undef INITERROR 

#if PY_MAJOR_VERSION >= 3
//#define PyInt_AS_LONG(op) (((PyLongObject *)(op))->ob_ival)
#endif

#if PY_VERSION_HEX >= 0x02060000
#define INITERROR return 0
#else
#define INITERROR return NULL
#endif

static int Connection_InternalRollback(t_Connection *connection);
static void Connection_Free(t_Connection*);

#if PY_VERSION_HEX < 0x02060000
  static PyObject *Connection_GetAttr(t_Connection*, PyObject*);
  static t_Connection *Connection_New(void);
#else
  static PyObject *Connection_New(PyTypeObject*, PyObject*, PyObject*);
  static int Connection_Init(t_Connection*, PyObject*, PyObject*);
#endif
static PyObject *Connection_Close(t_Connection*, PyObject*);
static PyObject *Connection_Commit(t_Connection*, PyObject*);
static PyObject *Connection_Rollback(t_Connection*, PyObject*);
static PyObject *Connection_NewCursor(t_Connection*, PyObject*, PyObject*);
static PyObject *Connection_Version(t_Connection*);

static PyMethodDef g_ConnectionMethods[] =
{
  {"cursor", (PyCFunction) Connection_NewCursor, METH_VARARGS | METH_KEYWORDS},
  {"commit", (PyCFunction) Connection_Commit, METH_VARARGS},
  {"rollback", (PyCFunction) Connection_Rollback, METH_VARARGS},
  {"close", (PyCFunction) Connection_Close, METH_VARARGS},
  {NULL, NULL}
};

#if PY_VERSION_HEX >= 0x02060000
static PyMemberDef g_ConnectionMembers[] = {
  { "username", T_OBJECT, offsetof(t_Connection, username), READONLY },
  { "password", T_OBJECT, offsetof(t_Connection, password), 0 },
  { "database", T_OBJECT, offsetof(t_Connection, database), READONLY },
  { "mode", T_INT, offsetof(t_Connection, mode), READONLY },
  { NULL }
};

static PyGetSetDef g_ConnectionCalcMembers[] = 
{
  { "version", (getter) Connection_Version, 0, 0, NULL },    
  { NULL }                                                 
};
#endif

PyTypeObject g_ConnectionType =
{
  PyVarObject_HEAD_INIT(NULL, 0)                         /* 25.12.13 */
  "Connection",             /* tp_name */
  sizeof(t_Connection),           /* tp_basicsize */
  0,          /* tp_itemsize */
  (destructor) Connection_Free,         /* tp_dealloc */
  0,          /* tp_print */
  0,          /* tp_getattr */
  0,          /* tp_setattr */
  0,          /* tp_compare */
  0,          /* tp_repr */
  0,          /* tp_as_number */
  0,          /* tp_as_sequence */
  0,          /* tp_as_mapping */
  0,          /* tp_hash */
  0,          /* tp_call */
  0,          /* tp_str */
  #if PY_VERSION_HEX < 0x02060000
    (getattrofunc) Connection_GetAttr,  /* tp_getattro */
  #else
   0,                                   /* tp_getattro */
  #endif  
  0,          /* tp_setattro */
  0,          /* tp_as_buffer */
  Py_TPFLAGS_DEFAULT,     /* tp_flags */ 
    0         /* tp_doc */
  #if PY_VERSION_HEX >= 0x02060000
    ,0,                   /* tp_traverse */
    0,                   /* tp_clear */
    0,                   /* tp_richcompare */
    0,                   /* tp_weaklistoffset */
    0,                   /* tp_iter */
    0,                   /* tp_iternext */
    g_ConnectionMethods,           /* tp_methods */
    g_ConnectionMembers,           /* tp_members */
    g_ConnectionCalcMembers,       /* tp_getset */
    0,                   /* tp_base */
    0,                   /* tp_dict */
    0,                   /* tp_descr_get */
    0,                   /* tp_descr_set */
    0,                   /* tp_dictoffset */
    (initproc) Connection_Init,            /* tp_init */
    0,                                    /* tp_alloc */
    (newfunc) Connection_New,             /* tp_new */   
	0,                                     /* tp_free */
    0,                                     /* tp_is_gc */  /* ... 25.12.13 */
    0                                      /* tp_bases */
   #endif 
};


/**
 * Check for error on connection
 */
int Connection_CheckForError(
  t_Connection *connection,
  int status)
{
  t_Error *error;
  short messageLength;
  char errorText[ERROR_MSG_SIZE];                            /* 25.12.13 */

  
  if (status == LINAPI_SUCCESS)
    return 0;

#if PY_VERSION_HEX >= 0x02060000                                           /* 20.10.14 ... */
  error = (t_Error *)PyObject_Call( (PyObject*) &g_ErrorType, NULL, NULL); /* 25.12.13 */
#else
  error = Error_New();
#endif                                                                     /* ... 20.10.14 */
  
  if (error == NULL)
    return -1;

  messageLength = ERROR_MSG_SIZE;
  LINTER_Error(connection->handle, 0, 0, 
    &error->apiCode, &error->linCode, &error->sysCode,
		errorText, &messageLength);

#if PY_VERSION_HEX >= 0x02060000                                                   /* 20.10.14 ... */
  error->message = PyUnicode_Decode(errorText, strlen(errorText), "cp866", NULL);
#else
  error->message = PyString_FromString(errorText);
#endif                                                                             /* ... 20.10.14 */
  
  PyErr_SetObject(g_DatabaseErrorException, (PyObject*) error);
  Py_DECREF(error);

  return -1;
}

/**
 * If connection not opened throw exception
 */
int Connection_IsConnected(t_Connection *connection)
{
  if (connection->handle == -1)
  {
    PyErr_SetString(g_InterfaceErrorException, "not connected");
    return -1;
  }
  return 0;
}

/**
 * Allocate new connection object
 */
#if PY_VERSION_HEX >= 0x02060000
static PyObject *Connection_New(                            /* 26.11.13 .. */
    PyTypeObject *type,                 // type object
    PyObject *args,                     // arguments
    PyObject *keywordArgs)              // keyword arguments
#else
static t_Connection *Connection_New(void)
#endif
{
  t_Connection *connection;
#if PY_VERSION_HEX >= 0x02060000
  connection = (t_Connection*) type->tp_alloc(type, 0);
#else
  connection = PyObject_NEW(t_Connection, &g_ConnectionType);
#endif
  if (!connection)
    return NULL;

  connection->handle = -1;
  connection->username = NULL;
  connection->password = NULL;
  connection->database = NULL;
  connection->version = NULL;
#if PY_VERSION_HEX >= 0x02060000 /* 30.09.14 ... */
  return (PyObject*)connection;
#else
  return connection;
#endif                           /* ... 30.09.14 */
}

/**
 * Get database version from LINAPI
 */
static PyObject *Connection_Version(t_Connection *connection)
{
  t_DBDesc desc;
  char version[128];
  int status;
  L_LONG length;
  
  if (connection->version)
  {
    Py_INCREF(connection->version);
    return connection->version;
  }
 
  length = sizeof(t_DBDesc);
  status = LINTER_GetConnectOption(connection->handle, cDBDesc, (void *) &desc, &length);
  if (Connection_CheckForError(connection, status) < 0)
    return NULL;

  sprintf(version, "%d.%d.%d",
    desc.VerMajor,
    desc.VerMinor,
    desc.VerBuild);
#if PY_MAJOR_VERSION < 3
  connection->version = PyString_FromString(version);         /* 25.12.13 */
#else
  connection->version = PyUnicode_Decode(version, strlen(version), NULL, NULL);
#endif
  if (!connection->version)
    return NULL;

  Py_INCREF(connection->version);
  return connection->version;
}


/**
 * Connect to database.
 */
#if PY_VERSION_HEX >= 0x02060000
static int Connection_Init(                                 /* 26.11.13 */
    t_Connection *connection,               // connection
    PyObject *args,                         // arguments
    PyObject *kwargs)
#else
PyObject *Connection_Connect(PyObject *self, PyObject *args, PyObject *kwargs)
#endif
{
  #if PY_VERSION_HEX < 0x02060000
    t_Connection *connection;
  #endif
  char *user = NULL;
  char *passwd = NULL;
  char *node = "";
  int mode = mExclusive;
  static char *kwlist[] =
  {
    "user",
    "password",
    "database",
    "mode",
    NULL
  };
  int status;

  trace(("Connection_Init: connection %08x\n", connection));
  status = PyArg_ParseTupleAndKeywords(args, kwargs, "ss|si", kwlist,
                                       &user, &passwd, &node, &mode);
  if (status == 0)
    INITERROR;
  
  #if PY_VERSION_HEX < 0x02060000
    connection = Connection_New();
    if (!connection)
      INITERROR;
  #endif  
	
  #if PY_MAJOR_VERSION >= 3
    connection->username = PyUnicode_Decode(user, strlen(user), NULL, NULL); 
    connection->password = PyUnicode_Decode(passwd, strlen(passwd), NULL, NULL);  
    connection->database = PyUnicode_Decode(node, strlen(node), NULL, NULL);
  #else
  connection->username = PyString_FromString(user);
	if (!connection->username)
    {
      Py_DECREF(connection);
      INITERROR;
    }
	connection->password = PyString_FromString(passwd);
	if (!connection->password)
	{
	  Py_DECREF(connection);
	  INITERROR;
	}
	connection->database = PyString_FromString(node);
	if (!connection->database)
	{
      Py_DECREF(connection);
      INITERROR;
	}
  #endif
  
  connection->mode = mode;
  if (strlen(node) == 0)
    node = NULL;
  
  #if PY_VERSION_HEX >= 0x02060000
   status = LINTER_ConnectCS(user, 0, passwd, 0, node, "UTF-8", mode, &connection->handle);
  #else
   status = LINTER_Connect(user, 0, passwd, 0, (node && (strlen(node) == 0) ? NULL : node), mode, &connection->handle);
  #endif
  if (Connection_CheckForError(connection, status))
  {
    connection->handle = -1;
  #if PY_VERSION_HEX >= 0x02060000
      return -1;
  #else          
      INITERROR;
  #endif
  }
  connection->ChanLink = NULL;                                   /* 28.11.08 */
  #if PY_VERSION_HEX >= 0x02060000
    INITERROR;
  #else
   return (PyObject*) connection;
  #endif
}


/**
 * Destructor.
 */
static void Connection_Free(t_Connection *connection)
{
  trace(("Connection_Free: connection %08x\n", connection));
  if (connection->handle != -1)
  {
    Connection_InternalRollback(connection);
    if (connection->ChanLink)                                /* 28.11.08 ... */
    {
      t_ChanLink *ChanDsc = connection->ChanLink;
      while (ChanDsc)
      {
        LINTER_RollBackCursor(ChanDsc->handle, NULL, NULL);
        LINTER_CloseCursor(ChanDsc->handle);
        connection->ChanLink = ChanDsc->Next;
        PyMem_Free(ChanDsc);
        ChanDsc = connection->ChanLink;
      }
    }                                                        /* ... 28.11.08 */
    LINTER_CloseConnect(connection->handle);
    connection->handle = -1;
  }
  Py_XDECREF(connection->username);
  Py_XDECREF(connection->password);
  Py_XDECREF(connection->database);
  Py_XDECREF(connection->version);
  PyObject_DEL(connection);                                      /* 04.09.08 */
}

#if PY_VERSION_HEX < 0x02060000
/**
 * Attributes
 */
static PyObject *Connection_GetAttr(t_Connection *connection, PyObject *nameObject)
{
  char *name;

  name = PyString_AS_STRING(nameObject);
  if ((name[0] == 'v') && (strcmp(name, "version") == 0))
  {
    return Connection_Version(connection);
  }
  else if ((name[0] == 'u') && (strcmp(name, "username") == 0))
  {
    Py_INCREF(connection->username);
    return connection->username;
  }
  else if ((name[0] == 'p') && (strcmp(name, "password") == 0))
  {
    Py_INCREF(connection->password);
    return connection->password;
  }
  else if ((name[0] == 'd') && (strcmp(name, "database") == 0))
  {
    Py_INCREF(connection->database);
    return connection->database;
  }
  else if ((name[0] == 'm') && (strcmp(name, "mode") == 0))
  {
    return PyInt_FromLong(connection->mode);
  }
    return Py_FindMethod(g_ConnectionMethods, (PyObject*) connection, name);
}
#endif

/**
 * Force closing connection.
 */
static PyObject *Connection_Close(t_Connection *connection, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ""))
    return NULL;

  if (Connection_IsConnected(connection) < 0)
    return NULL;

/*      ??? !!! */
/*
  if (Connection_InternalRollback(connection) < 0)
    return NULL;
*/
  LINTER_Commit(connection->handle, NULL, NULL);             /* 28.11.08 ... */

  if (connection->ChanLink)
  {
    t_ChanLink *ChanDsc = connection->ChanLink;
    while (ChanDsc)
    {
      LINTER_CommitCursor(ChanDsc->handle, NULL, NULL);
      LINTER_CloseCursor(ChanDsc->handle);
      connection->ChanLink = ChanDsc->Next;
      PyMem_Free(ChanDsc);
      ChanDsc = connection->ChanLink;
    }
  }                                                          /* ... 28.11.08 */
  LINTER_CloseConnect(connection->handle);

  connection->handle = -1;

  Py_INCREF(Py_None);
  return Py_None;
}

/**
 * Commit all cursors.
 */
static PyObject *Connection_Commit(t_Connection *connection, PyObject *args)
{
  int status;

  if (!PyArg_ParseTuple(args, ""))
    return NULL;

  if (Connection_IsConnected(connection) < 0)
    return NULL;

  status = LINTER_Commit(connection->handle, NULL, NULL);
  if (Connection_CheckForError(connection, status) < 0)
    return NULL;
  if (connection->ChanLink)                                  /* 28.11.08 ... */
  {
    t_ChanLink *ChanDsc = connection->ChanLink;
    while (ChanDsc)
    {
      LINTER_CommitCursor(ChanDsc->handle, NULL, NULL);
      LINTER_CloseCursor(ChanDsc->handle);
      connection->ChanLink = ChanDsc->Next;
      PyMem_Free(ChanDsc);
      ChanDsc = connection->ChanLink;
    }
  }                                                          /* ... 28.11.08 */
  Py_INCREF(Py_None);                                        
  return Py_None;
}


/**
 * Call rollback on connection
 */
static int Connection_InternalRollback(t_Connection *connection)
{
  int status;

  status = LINTER_RollBack(connection->handle, NULL, NULL);
  if (Connection_CheckForError(connection, status) < 0)
    return -1;

  return 0;
}


/**
 * Rollback all cursor.
 */
static PyObject *Connection_Rollback(t_Connection *connection, 
                                     PyObject *args)
{
  if (!PyArg_ParseTuple(args, ""))
    return NULL;

  if (Connection_IsConnected(connection) < 0)
    return NULL;

  if (Connection_InternalRollback(connection) < 0)
    return NULL;

  if (connection->ChanLink)                                  /* 28.11.08 ... */
  {
    t_ChanLink *ChanDsc = connection->ChanLink;
    while (ChanDsc)
    {
      LINTER_RollBackCursor(ChanDsc->handle, NULL, NULL);
      LINTER_CloseCursor(ChanDsc->handle);
      connection->ChanLink = ChanDsc->Next;
      PyMem_Free(ChanDsc);
      ChanDsc = connection->ChanLink;
    }
  }                                                          /* ... 28.11.08 */

  Py_INCREF(Py_None);
  return Py_None;
}

/**
 * Create new cursor object.
 */
  static PyObject *Connection_NewCursor(
  t_Connection *con,
  PyObject *args,
  PyObject *kwargs)
{
  #if PY_VERSION_HEX < 0x02060000
    int mode = con->mode;
    static char *kwlist[] = {"mode", NULL};

    if (!PyArg_ParseTuple(args, ""))
      return NULL;

    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|i", kwlist,
                                   &mode))
      return NULL;

    if (Connection_IsConnected(con) < 0)
      return NULL;
    
    return Cursor_New(con, mode);
  #else
    PyObject *createArgs, *result;
    int mode;
    if (Connection_IsConnected(con) < 0)
      return NULL;

    if ( PyTuple_GET_SIZE(args) == 0 )
      createArgs = PyTuple_New(1);
    else if ( PyTuple_GET_SIZE(args) == 1 )
      createArgs = PyTuple_New(2);
    if (!createArgs)
      return NULL;

    Py_INCREF(con);
    PyTuple_SET_ITEM(createArgs, 0, (PyObject*) con);
    if ( PyTuple_GET_SIZE(args) == 1 )
      if (PyArg_ParseTuple(args, "|i", &mode))
        PyTuple_SET_ITEM(createArgs, 1, PyInt_FromLong(mode));

    result = PyObject_Call( (PyObject*) &g_CursorType, createArgs, kwargs);
    Py_DECREF(createArgs);
    return result; 
  #endif
}
