 /***************************************************************************
 *                 RELEX, Inc. Linter Python driver source code             *
 *        This is unpublished proprietary source code of RELEX, Inc.        *
 *     The copyright notice does not evidence any actual or intended        *
 *                  publication of such source code.                        *
 *                                                                          *
 *                 Copyright (c) 2003-2014 RELEX Inc.                       *
 *                                                                          *
 *                   This software was developed by                         *
 *                                                                          *
 *                            RELEX, Inc.                                   *
 *                         Voronezh, Russia                                 *
 *                                                                          *
 *                        All rights reserved.                              *
 ****************************************************************************
 AUTHOR: Boris Fedorov
 ----------------------------------------------------------------------------
 PURPOSE: Cursor interface (DBAPI)
 ----------------------------------------------------------------------------
 REVISION HISTORY:
 06.09.03 -- Boris Fedorov (Supress compilation warnings)
 31.10.03 -- Boris Fedorov (Remove C++ style comments, PySequence_Size
 replaced by PySequence_Length for compatibility)
 29.07.08 -- Ermakov (Corrected warnings, added Boolean type)
 08.08.08 -- Ermakov (Added list to execute arguments)
 18.08.08 -- Ermakov (Corrected list handle)
 04.09.08 -- Ermakov (ObjectDel instead of MemDel)
 08.02.10 -- KOV     (Added constructions 'TABLE <table_name>' and 'VALUES (<value_list>) [,(<value_list>)] ...')
 25.12.13 -- Muratova (Python 3 support)
 06.10.14 -- Kononov (Fixed support for Python < 2.6)
 07.10.14 -- Kononov (Fixed binding named parameters to a prepared query when keys are not in uppercase)
 13.10.14 -- Kononov (Fixed behavior - if BLOB column exists then cursor will return only one row)
 14.10.14 -- Kononov (Implemented function setoutputsize() for setting up fetch size of BLOB columns or one of them)
*/

#undef _DEBUG
#include <Python.h>
#include <linapi.h>
#include "LinPy.h"

#define DEFAULT_ARRAY_SIZE 1

static void Cursor_Free(t_Cursor*);
static int Cursor_SetAttr(t_Cursor*, PyObject*, PyObject*);
#if (PY_VERSION_HEX >= 0x02020000)
static PyObject *Cursor_GetIter(PyObject*);
static PyObject *Cursor_GetNext(t_Cursor*);
#endif
static PyObject *Cursor_Close(t_Cursor*, PyObject*);
static PyObject *Cursor_CallProc(t_Cursor*, PyObject*);
static PyObject *Cursor_Execute(t_Cursor*, PyObject*, PyObject*);
static PyObject *Cursor_ExecuteMany(t_Cursor*, PyObject*);
static PyObject *Cursor_ExecuteManyPrepared(t_Cursor*, PyObject*);
static PyObject *Cursor_FetchOne(t_Cursor*, PyObject*);
static PyObject *Cursor_FetchMany(t_Cursor*, PyObject*);
static PyObject *Cursor_FetchAll(t_Cursor*, PyObject*);
static PyObject *Cursor_Prepare(t_Cursor*, PyObject*);
static PyObject *Cursor_SetInputSizes(t_Cursor*, PyObject*, PyObject*);
static PyObject *Cursor_SetOutputSize(t_Cursor*, PyObject*);
static PyObject *Cursor_Var(t_Cursor*, PyObject*);
static PyObject *Cursor_ExternalSetValue(t_Cursor *cursor, PyObject *args); /* 06.10.14 */

static PyObject *Cursor_Rollback(t_Cursor *cursor, PyObject *args);
static PyObject *Cursor_Commit(t_Cursor *cursor, PyObject *args);

#if PY_VERSION_HEX < 0x02060000
static PyObject *Cursor_GetAttr(t_Cursor*, PyObject*);             /* 25.12.13 */
#else
static PyObject *Cursor_New(PyTypeObject*, PyObject*, PyObject*);  /* 25.12.13 ... */
static int Cursor_Init(t_Cursor*, PyObject*, PyObject*);
static int Cursor_SetName(t_Cursor *, PyObject *);
static PyObject *Cursor_GetName(t_Cursor *);
static PyObject *Cursor_Description(t_Cursor *);
static int Cursor_SetArraysize(t_Cursor *, PyObject *);
static int Cursor_GetArraysize(t_Cursor *);
static PyObject *Cursor_ArrayVar(t_Cursor*, PyObject*);
#endif

static PyMethodDef g_CursorMethods[] =
{
  {"execute", (PyCFunction) Cursor_Execute, METH_VARARGS | METH_KEYWORDS },
  {"fetchall", (PyCFunction) Cursor_FetchAll, METH_VARARGS },
  {"fetchone", (PyCFunction) Cursor_FetchOne, METH_VARARGS },
  {"fetchmany", (PyCFunction) Cursor_FetchMany, METH_VARARGS },
  {"prepare", (PyCFunction) Cursor_Prepare, METH_VARARGS },
  {"setinputsizes", (PyCFunction) Cursor_SetInputSizes, METH_VARARGS | METH_KEYWORDS },
  {"executemany", (PyCFunction) Cursor_ExecuteMany, METH_VARARGS },
  {"callproc", (PyCFunction) Cursor_CallProc, METH_VARARGS },
  {"executemanyprepared", (PyCFunction) Cursor_ExecuteManyPrepared, METH_VARARGS},
  {"setoutputsize", (PyCFunction) Cursor_SetOutputSize, METH_VARARGS },
  {"setvalue", (PyCFunction) Cursor_ExternalSetValue, METH_VARARGS },
  #if PY_VERSION_HEX >= 0x02060000
  { "arrayvar", (PyCFunction) Cursor_ArrayVar, METH_VARARGS },
  #endif
  {"commit", (PyCFunction) Cursor_Commit, METH_VARARGS},
  {"rollback", (PyCFunction) Cursor_Rollback, METH_VARARGS},
  {"close", (PyCFunction) Cursor_Close, METH_VARARGS },
  {NULL, NULL}
};

#if PY_VERSION_HEX >= 0x02060000
static PyMemberDef g_CursorMembers[] = {
    { "mode", T_INT, offsetof(t_Cursor, mode), 0 },
    { "rowcount", T_INT, offsetof(t_Cursor, rowCount), READONLY },
    { "statement", T_OBJECT, offsetof(t_Cursor, statement), READONLY },
    { "connection", T_OBJECT_EX, offsetof(t_Cursor, connection), READONLY },
    { NULL }
};

static PyGetSetDef g_CursorCalcMembers[] = {
    { "name", (getter) Cursor_GetName, (setter) Cursor_SetName, 0, 0 },
    { "description", (getter) Cursor_Description, 0, 0, 0 },
    { "arraysize", (getter) Cursor_GetArraysize, (setter) Cursor_SetArraysize, 0, 0 },
    { NULL }
};                                                           /* ... 25.12.13 */
#endif

PyTypeObject g_CursorType =
{
  PyVarObject_HEAD_INIT(NULL, 0)
  "Cursor",                           /* tp_name */
  sizeof(t_Cursor),                   /* tp_basicsize */
  0,                                  /* tp_itemsize */
  (destructor)Cursor_Free,            /* tp_dealloc */
  0,                                  /* tp_print */
  0,                                  /* tp_getattr */
  0,                                  /* tp_setattr */
  0,                                  /* tp_compare */
  0,                                  /* tp_repr */
  0,                                  /* tp_as_number */
  0,                                  /* tp_as_sequence */
  0,                                  /* tp_as_mapping */
  0,                                  /* tp_hash */
  0,                                  /* tp_call */
  0,                                  /* tp_str */
  #if PY_VERSION_HEX < 0x02060000
    (getattrofunc) Cursor_GetAttr,    /* tp_getattro */
    (setattrofunc) Cursor_SetAttr,    /* tp_setattro */
  #else
    0,                                /* tp_getattro */
    0,                                /* tp_setattro */
  #endif
  0,                                  /* tp_as_buffer */
  Py_TPFLAGS_DEFAULT,                 /* tp_flags */
  0                                   /* tp_doc */
  #if (PY_VERSION_HEX >= 0x02020000)
    , 0,                              /* tp_traverse */
    0,                                /* tp_clear */
    0,                                /* tp_richcompare */
    0,                                /* tp_weaklistoffset */
    (getiterfunc) Cursor_GetIter,     /* tp_iter */
    (iternextfunc) Cursor_GetNext     /* tp_iternext */
  #endif
  #if PY_VERSION_HEX >= 0x02060000
    , g_CursorMethods,                /* tp_methods */
    g_CursorMembers,                  /* tp_members */
    g_CursorCalcMembers,              /* tp_getset */
    0,                                /* tp_base */
    0,                                /* tp_dict */
    0,                                /* tp_descr_get */
    0,                                /* tp_descr_set */
    0,                                /* tp_dictoffset */
    (initproc) Cursor_Init,           /* tp_init */
    0,                                /* tp_alloc */
    (newfunc) Cursor_New,             /* tp_new */
    0,                                /* tp_free */
    0,                                /* tp_is_gc */
    0                                 /* tp_bases */
  #endif
};                                                           /* ... 25.12.13 */


int Cursor_CheckForError(
  t_Cursor *cursor,
  int status)
{
  t_Error *error;
  short messageLength;
  char errorText[ERROR_MSG_SIZE];                            /* 25.12.13 */

  if (status == LINAPI_SUCCESS)
    return 0;
#if PY_VERSION_HEX >= 0x02060000
  error = (t_Error *)PyObject_Call( (PyObject*) &g_ErrorType, NULL, NULL); /* 25.12.13 */
#else
  error = Error_New();                                                     /* 06.10.14 */
#endif
  if (error == NULL)
    return -1;

  messageLength = ERROR_MSG_SIZE;
  LINTER_Error(0, (cursor->handleFromSP) ? cursor->handleFromSP : cursor->handle, 0,
    &error->apiCode, &error->linCode, &error->sysCode,
    errorText, &messageLength);                             /* 25.12.13 */
    
  error->code = error->apiCode;                             /* 25.12.13 */
  
  if (error->linCode == 0)
  {
    messageLength = ERROR_MSG_SIZE;
    LINTER_Error((cursor->handleFromSP) ? cursor->handleFromSP : cursor->connection->handle, 0, 0,
      &error->apiCode, &error->linCode, &error->sysCode,
      errorText, &messageLength);                           /* 25.12.13 */
    error->code  = error->apiCode;

    if ((error->linCode == 0) && (cursor->statement))
    {
      messageLength = ERROR_MSG_SIZE;
      LINTER_Error(0, 0, (cursor->handleFromSP) ? cursor->handleFromSP : cursor->statementHandle,
        &error->apiCode, &error->linCode, &error->sysCode,
        errorText, &messageLength);                         /* 25.12.13 */
      error->code  = error->apiCode;
    }
  }

#if PY_VERSION_HEX >= 0x02060000
  error->message = PyUnicode_Decode(errorText, strlen(errorText), "cp866", NULL); /* 25.12.13 */
#else
  error->message = PyString_FromString(errorText);
#endif

  PyErr_SetObject(g_DatabaseErrorException, (PyObject*) error);
  Py_DECREF(error);
  return -1;
}

int Cursor_IsOpen(t_Cursor *cursor)
{
  if (cursor->handle == -1)
  {
    PyErr_SetString(g_InterfaceErrorException, "cursor not open");
    return -1;
  }
  return Connection_IsConnected(cursor->connection);
}


#if PY_VERSION_HEX >= 0x02060000
static PyObject *Cursor_New(                                 /* 25.12.13 ... */
    PyTypeObject *type,                 // type object
    PyObject *args,                     // arguments
    PyObject *keywordArgs)              // keyword arguments
{
  trace(("Cursor_New\n"));
  return type->tp_alloc(type, 0);
}


static int Cursor_Init(
    t_Cursor *cursor,                   // cursor object
    PyObject *args,                     // arguments
    PyObject *keywordArgs)              // keyword arguments
{
  t_Connection *con;
#else
PyObject *Cursor_New(t_Connection *con, int mode) /* 06.10.14 */
{
  t_Cursor *cursor;
#endif
  int status = 0;                               /* 28.11.08 */
#if PY_VERSION_HEX >= 0x02060000
  int mode = -1;

  static char *kwlist[] =
  {
    "con",
    "mode",
    NULL
  };

  trace(("Cursor_Init: cursor %08x\n", cursor));
  if (!PyArg_ParseTupleAndKeywords(args, keywordArgs, "O!|i", kwlist,
                                   &g_ConnectionType, &con, &mode))
    return -1;
  else
    {
      if (mode == -1)
        mode = con->mode;
        //cursor->mode = mode;
    }

  Py_INCREF(con);                                            /* ... 25.12.13 */
#else
  cursor = PyObject_NEW(t_Cursor, &g_CursorType);
  if (!cursor)
    return NULL;
#endif
  cursor->connection = NULL;
  cursor->mode = mode;
  cursor->handle = -1;
  cursor->handleFromSP = 0;
  cursor->statementHandle = -1;
  cursor->statement = NULL;
  cursor->bindArrays = NULL;
  cursor->fetchArrays = NULL;
  cursor->fetchArraySize = 0;
  cursor->blobOutputSizes = NULL;                            /* 14.10.14 ... */
  cursor->blobOutputSize = -1;                               /* ... 14.10.14 */
  cursor->arraySize = DEFAULT_ARRAY_SIZE;
  cursor->rowCount = -1;
  cursor->fetchOffset = 1;
  cursor->actualRows = -1;
  cursor->rowNum = 0;
  cursor->statementType = 0;
  cursor->isDML = 0;
  cursor->blobColumns = 0;                                   /* 13.10.14 */

#if PY_VERSION_HEX >= 0x02060000
  if (con->mode != mAutocommit)                                   /* 28.11.08 ... */
#else  
  if (mode != mAutocommit)                                   /* 28.11.08 ... */
#endif
  {
    t_ChanLink *ChanDsc = con->ChanLink;
	#if PY_VERSION_HEX >= 0x02060000
	  while ((ChanDsc) && (ChanDsc->mode != con->mode))
	#else
      while ((ChanDsc) && (ChanDsc->mode != mode))
	#endif
      ChanDsc = ChanDsc->Next;
    if (ChanDsc)
    {
      cursor->handle = ChanDsc->handle;
      con->ChanLink = ChanDsc->Next;
      PyMem_Free(ChanDsc);
    }
  }
  if (cursor->handle == -1)                                  /* ... 28.11.08 */
  {
    status = LINTER_OpenCursor(con->handle,
                               &cursor->handle,
                               NULL,
                               0,
                               cursor->mode);
  }
  if (Connection_CheckForError(con, status) < 0)
  {
    Py_DECREF(cursor);
#if PY_VERSION_HEX >= 0x02060000
    return -1;
#else
    return NULL;
#endif
  }

  Py_INCREF(con);
  cursor->connection = con;

#if PY_VERSION_HEX >= 0x02060000
  return 0;
#else
    return (PyObject*) cursor;
#endif
}


static void Cursor_Free(t_Cursor *cursor)
{
  if (cursor->handle != -1)
  {
    if (cursor->mode != mAutocommit)                         /* 28.11.08 ... */
    {
      t_ChanLink *ChanDsc = PyMem_Malloc(sizeof(t_ChanLink));
      if (ChanDsc)
      {
        ChanDsc->handle = cursor->handle;
        ChanDsc->mode = cursor->mode;
        ChanDsc->Next = cursor->connection->ChanLink;
        cursor->connection->ChanLink = ChanDsc;
      }
      else
      {
        LINTER_CloseCursor(cursor->handle);
      }
    }
    else
    {
      LINTER_RollBackCursor(cursor->handle, NULL, NULL);
      LINTER_CloseCursor(cursor->handle);
    }
  }                                                          /* ... 28.11.08 */
  Py_DECREF(cursor->connection);
  PyObject_DEL(cursor);                                          /* 04.09.08 */
}


/**
 * Perform the defines for the cursor. At this point it is assumed that the
 * statement being executed is in fact a query.
 */
static int Cursor_PerformDefine(
  t_Cursor *cursor)
{
  L_LONG columnCount;
  L_LONG len;
  t_Array *array;
  int status;
  int column;
  int pos = 0;
  PyObject *key;
  PyObject *value;

  Py_XDECREF(cursor->fetchArrays);
  cursor->fetchArrays = NULL;
  cursor->fetchArraySize = 0;
  cursor->actualRows = 0;
  cursor->blobColumns = 0;           /* 13.10.14 */

  if ((cursor->statementType == syExecute) &&
      (!cursor->handleFromSP))
  {
    /* no cursor from sp returned */
    return 0;
  }

  len = sizeof(columnCount);
  status = LINTER_GetCursorOption(
    (L_WORD)((cursor->handleFromSP) ? cursor->handleFromSP : cursor->handle),
    cColumnCount,
    1,
    &columnCount,
    &len);
  if (Cursor_CheckForError(cursor, status))
    return -1;

  if (columnCount <= 0)
    return 0;

  cursor->fetchArrays = PyList_New(columnCount);
  if (cursor->fetchArrays == NULL)
    return -1;

  for (column = 1; column <= columnCount; column++)
  {
    pos = 0;
    array = Array_Define(cursor, cursor->arraySize, column);
    if (array == NULL)
    {
      Py_DECREF(cursor->fetchArrays);
      cursor->fetchArrays = NULL;
      return -1;
    }
    if (LinPyBlob_Check(array))
    {
      cursor->blobColumns++;                                                 /* 13.10.14 */
      if (cursor->blobOutputSizes)                                           /* 14.10.14 ... */
      {
        if (cursor->blobOutputSize < 0)
        {
          while (PyDict_Next(cursor->blobOutputSizes, &pos, &key, &value))
          {
            if (array->index == PyLong_AsLong(key))
            {
              array->blobPortion = PyLong_AsLong(value);
              break;
            }
          }
        }
        else
        {
          array->blobPortion = cursor->blobOutputSize;
        }
      }
    }
    PyList_SET_ITEM(cursor->fetchArrays, column - 1, (PyObject *) array);    /* ... 14.10.14 */
  }

  if (cursor->blobColumns > 0)
  {
    cursor->fetchArraySize = 1;
  }
  else
  {
    cursor->fetchArraySize = cursor->arraySize;
  }

  return 0;
}


/**
 * Create cursor parameters Array as dictionary.
 */
static int Cursor_DefineNamedParameters(
  t_Cursor *cursor,
  int arraySize)
{
  L_LONG length;
  int status;
  int paramCount;
  int i;
  PyObject *params;
  t_Array *var;

  Py_XDECREF(cursor->bindArrays);
  cursor->bindArrays = NULL;

  params = PyDict_New();
  if (!params)
    return -1;

  length = sizeof(paramCount);
  status = LINTER_GetStatementOption(
    cursor->statementHandle,
    sParamCount,
    1,
    (void*) &paramCount,
    &length);
  if (Cursor_CheckForError(cursor, status) < 0)
  {
    Py_DECREF(params);
    return -1;
  }

  for (i = 1; i <= paramCount; i++)
  {
    var = Array_NewParameter(cursor, i, arraySize);
    if (!var)
    {
      Py_DECREF(params);
      return -1;
    }                                                
	
#if PY_MAJOR_VERSION < 3                                     /* 25.12.13 ... */
  if (PyDict_SetItem(params, PyString_FromString(var->desc.Column), (PyObject *) var) < 0)
#else
  if (PyDict_SetItem(params, PyUnicode_FromString(var->desc.Column), (PyObject *)var) < 0)
#endif	                                                     /* ... 25.12.13 */
    {
      Py_DECREF(params);
      return -1;
    }
    Py_DECREF(var);
  }

  cursor->bindArrays = params;

  return 0;
}


/**
 * Create cursor parameters Array as tuple.
 */
static int Cursor_DefineQmarkParameters(
  t_Cursor *cursor,
  int arraySize)
{
  L_LONG length;
  int status;
  int paramCount;
  int i;
  PyObject *params;
  t_Array *var;

  trace(("Cursor_DefineQmarkParameters\n"));
  Py_XDECREF(cursor->bindArrays);
  cursor->bindArrays = NULL;

  length = sizeof(paramCount);
  status = LINTER_GetStatementOption(
    cursor->statementHandle,
    sParamCount,
    1,
    (void*) &paramCount,
    &length);
  if (Cursor_CheckForError(cursor, status) < 0)
    return -1;

  params = PyTuple_New(paramCount);
  if (!params)
    return -1;

  for (i = 0; i < paramCount; i++)
  {
    var = Array_NewParameter(cursor, i + 1, arraySize);
    if (!var)
    {
      Py_DECREF(params);
      return -1;
    }
    if (PyTuple_SetItem(params, i, (PyObject *) var) < 0)
    {
      Py_DECREF(var);
      Py_DECREF(params);
      return -1;
    }
  }

  cursor->bindArrays = params;

  return 0;
}


/**
 * Get cursor column description
 * (name, type, displaySize, internalSize, precision, scale, nullable)
 */
static PyObject *Cursor_ColumnDescription(
  t_Cursor *cursor,
  unsigned column)
{
  PyObject *tuple;
  PyObject *columnName;
  t_Array *array;
  int displaySize;
  int columnNameLength;

  array = (t_Array *) PyList_GET_ITEM(cursor->fetchArrays, column - 1);
  if (array == NULL)
    return NULL;

  switch (array->type->linterType)
  {
    case tExtFile:
      displaySize = 512;
      break;
    case tVarChar:
      displaySize = array->maxLength - sizeof(L_WORD);             /* 28.11.08 */
      break;
    case tChar:
      displaySize = array->maxLength;
      break;
    case tByte:
      displaySize = array->maxLength << 1;
      break;
    case tVarByte:                                               /* 28.11.08 */
      displaySize = (array->maxLength-sizeof(L_WORD)) << 1;
      break;
    case tSmallInt:
      displaySize = 6;
      break;
    case tInteger:
      displaySize = 11;
      break;
    case tReal:
      displaySize = 13;
      break;
    case tDouble:
      displaySize = 22;
      break;
    case tBigInt:
      displaySize = 20;
      break;
    case tDecimal:
      displaySize = array->desc.Prec;
      break;
    case tTimeStamp:
      displaySize = 23;
      break;
    case tBlob:
      displaySize = 0;
      break;
    case tBoolean:
      displaySize = 1;
      break;
    case tNVarChar:
      displaySize = ((array->maxLength-sizeof(L_WORD)) >> 1);      /* 28.11.08 */
      break;
    case tNChar:
      displaySize = array->maxLength >> 1;
      break;
    default:
      PyErr_Format(g_NotSupportedErrorException,
        "Cursor_ColumnDescription: unhandled data type %d",
        array->type->linterType);
      return NULL;
  }

  tuple = PyTuple_New(7);
  if (tuple == NULL)
    return NULL;

  columnNameLength = strlen(array->desc.Column);
  if (columnNameLength >= linNameLen)
    columnNameLength = linNameLen;

  columnName = PyString_FromStringAndSize(array->desc.Column, columnNameLength);

  PyTuple_SET_ITEM(tuple, 0, columnName);
  Py_INCREF(Py_TYPE(array));                                /* 26.11.13 .. */
  PyTuple_SET_ITEM(tuple, 1, (PyObject *) Py_TYPE(array));  /* .. 26.11.13 */
  PyTuple_SET_ITEM(tuple, 2, PyInt_FromLong(displaySize));
  PyTuple_SET_ITEM(tuple, 3, PyInt_FromLong(array->desc.Length));
  PyTuple_SET_ITEM(tuple, 4, PyInt_FromLong(array->desc.Prec));
  PyTuple_SET_ITEM(tuple, 5, PyInt_FromLong(array->desc.Scale));
  PyTuple_SET_ITEM(tuple, 6, PyInt_FromLong(array->desc.NullIndicator));

  return tuple;
}


/**
 * Get cursor description
 */
static PyObject *Cursor_Description(t_Cursor *cursor)
{
  PyObject *results;
  PyObject *tuple;
  int numItems;
  int index;

  if (Cursor_IsOpen(cursor) < 0)
    return NULL;

  if (!cursor->fetchArrays)
  {
    Py_INCREF(Py_None);
    return Py_None;
  }

  numItems = PyList_GET_SIZE(cursor->fetchArrays);
  results = PyList_New(numItems);
  if (!results)
    return NULL;

  for (index = 0; index < numItems; index++)
  {
    tuple = Cursor_ColumnDescription(cursor, index + 1);
    if (!tuple)
    {
      Py_DECREF(results);
      return NULL;
    }
    PyList_SET_ITEM(results, index, tuple);
  }

  return results;
}

#if PY_VERSION_HEX >= 0x02060000
/*number rows for fetch */
static int Cursor_SetArraysize(t_Cursor *cursor, PyObject *value) /* 25.12.13 ... */
{
  if (!PyInt_Check(value))
  {
    PyErr_SetString(PyExc_TypeError, "array size must be an integer");
    return -1;
  }
  cursor->arraySize = PyInt_AS_LONG(value);
  return 0;
}

static int Cursor_GetArraysize(t_Cursor *cursor)
{
  return (int)PyInt_FromLong(cursor->arraySize);
}                                                            /* ... 25.12.13 */
#endif

static PyObject *Cursor_GetName(t_Cursor *cursor)
{
  char buf[linNameLen+1];
  L_LONG len;
  int status;

  if (Cursor_IsOpen(cursor) < 0)
    return NULL;

  len = sizeof(buf);

  status = LINTER_GetCursorOption(
    cursor->handle,
    cCursorName,
    0,
    buf,
    &len);

  if (Cursor_CheckForError(cursor, status) < 0)
    return NULL;                                             /* 06.09.03 */

  return PyString_FromStringAndSize(buf, len);
}

#if PY_VERSION_HEX >= 0x02060000
static int Cursor_SetName(t_Cursor *cursor, PyObject *value) /* 25.12.13 ...*/
#else
static int Cursor_SetName(t_Cursor *cursor, char *name)
#endif
{
  int status;
  L_LONG len;
#if PY_VERSION_HEX >= 0x02060000
  char * name;

  //name = (PyString_Check(value)) ? PyString_AS_STRING(value) : (PyUnicode_Check(name)) ? 
   //   PyString_AS_STRING(PyUnicode_AsEncodedString(value, NULL, NULL)) : NULL;
	  
  if (PyUnicode_Check(value))
  {    
	name  = PyBytes_AS_STRING(PyUnicode_AsEncodedString(value, NULL, NULL));
  }
  else
  {
    name  = PyBytes_AS_STRING(value);
  }                                                          /* ... 25.12.13 */
#endif  

  if (Cursor_IsOpen(cursor) < 0)
    return -1;                                             /* 06.09.03 */

  len = strlen(name);
    
  status = LINTER_SetCursorOption(
    cursor->handle,
    cCursorName,
    name,
    &len);

  if (Cursor_CheckForError(cursor, status) < 0)
    return -1;
  return 0;
}

#if PY_VERSION_HEX < 0x02060000
static PyObject *Cursor_GetAttr(t_Cursor *cursor, PyObject *nameObject)
{
  char *name;
  name = PyString_AS_STRING(nameObject);
  if ((name[0] == 'r') && (strcmp(name, "rowcount") == 0))
    return PyInt_FromLong(cursor->rowCount);
  if ((name[0] == 'a') && (strcmp(name, "arraysize") == 0))
    return PyInt_FromLong(cursor->arraySize);
  if ((name[0] == 'd') && (strcmp(name, "description") == 0))
    return Cursor_Description(cursor);
  if ((name[0] == 's') && (strcmp(name, "statement") == 0))
  {
    if (!cursor->statement)
    {
      Py_INCREF(Py_None);
      return Py_None;
    }
    Py_INCREF(cursor->statement);
    return (PyObject*) cursor->statement;
  }
  if ((name[0] == 'c') && (strcmp(name, "connection") == 0))
  {
    Py_INCREF(cursor->connection);
    return (PyObject*) cursor->connection;
  }
  if ((name[0] == 'm') && (strcmp(name, "mode") == 0))
    return PyInt_FromLong(cursor->mode);
  if ((name[0] == 'n') && (strcmp(name, "name") == 0))
    return Cursor_GetName(cursor);
  return Py_FindMethod(g_CursorMethods, (PyObject*) cursor, name);
}
static int Cursor_SetAttr(
  t_Cursor *cursor,
  PyObject *nameObject,
  PyObject *value)
{
  const char *name;
  name = PyString_AS_STRING(nameObject);
  if ((name[0] == 'a') && (strcmp(name, "arraysize") == 0))
  {
    if (!PyInt_Check(value))
    {
      PyErr_SetString(PyExc_TypeError, "array size must be an integer");
      return -1;
    }
    cursor->arraySize = PyInt_AS_LONG(value);
  return 0;
  }
  if ((name[0] == 'n') && (strcmp(name, "name") == 0))
  {
    if (!PyString_Check(value))
    {
      PyErr_SetString(PyExc_TypeError, "expecting string data");
      return -1;
    }
    return Cursor_SetName(cursor, PyString_AS_STRING(value));
  }
  PyErr_SetObject(PyExc_AttributeError, nameObject);
  return -1;
}
#endif

static PyObject *Cursor_Close(t_Cursor *cursor, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ""))
    return NULL;

  if (Cursor_IsOpen(cursor) < 0)
    return NULL;

  if (cursor->mode != mAutocommit)                           /* 28.11.08 ... */
  {
    t_ChanLink *ChanDsc = PyMem_Malloc(sizeof(t_ChanLink));
    if (ChanDsc)
    {
      ChanDsc->handle = cursor->handle;
      ChanDsc->mode = cursor->mode;
      ChanDsc->Next = cursor->connection->ChanLink;
      cursor->connection->ChanLink = ChanDsc;
    }
    else
    {
      LINTER_CloseCursor(cursor->handle);
    }
    if (cursor->handleFromSP)
    {
      ChanDsc = PyMem_Malloc(sizeof(t_ChanLink));
      if (ChanDsc)
      {
        ChanDsc->handle = cursor->handleFromSP;
        ChanDsc->Next = cursor->connection->ChanLink;
        ChanDsc->mode = cursor->mode;
        cursor->connection->ChanLink = ChanDsc;
      }
      else
      {
        LINTER_CloseCursor(cursor->handleFromSP);
      }
    }
  }
  else
  {
    if (cursor->handleFromSP)
      LINTER_CloseCursor(cursor->handleFromSP);
    LINTER_CloseCursor(cursor->handle);
  }
  cursor->handleFromSP = 0;                                  /* ... 28.11.08 */
  cursor->handle = -1;

  Py_INCREF(Py_None);
  return Py_None;
}


/**
 * Clear all bind Arrays for the cursor.
 */
/*
static int Cursor_ClearBindArrays(
  t_Cursor *cursor)
{
  Py_XDECREF(cursor->bindArrays);
  cursor->bindArrays = PyDict_New();
  if (!cursor->bindArrays)
    return -1;
  return 0;
}
*/


/**
 * Create or set bind Arrays.
 */
static int Cursor_SetParametersFromDict(
  t_Cursor *cursor,
  PyObject *parameters,
  int arraySize,
  unsigned arrayPos)
{
  PyObject *key;
  PyObject *value;
  t_Array *var;
  char *keyName;
  int i, pos, size;

  if (!cursor->bindArrays)
  {
    if (Cursor_DefineNamedParameters(cursor, arraySize) < 0)
      return -1;
  }

  pos = 0;
  while (PyDict_Next(parameters, &pos, &key, &value))
  {
    keyName = (PyUnicode_Check(key) ?
      PyString_AS_STRING(PyUnicode_AsEncodedString(key, NULL, NULL)) : PyString_AS_STRING(key)); /* 07.10.14 ... */
    size = (keyName) ? strlen(keyName) : 0;
    for (i = 0; i < size; i++)
    {
      keyName[i] = toupper(keyName[i]);
    }                                                                                            /* ... 07.10.14 */
    var = (t_Array*) PyDict_GetItem(cursor->bindArrays, PyString_FromString(keyName));
    if (!var)
    {
      PyErr_SetString(g_InterfaceErrorException,
        "Cursor_SetParametersFromDict: expecting a list of dictionaries");
      return -1;
    }
    if (!Array_Check(value))
    {
      if (Array_SetValue(var, arrayPos, value) < 0)
        return -1;
    }
  }

  return 0;
}


/**
 * Create or set bind Arrays from tuple.
 */
static int Cursor_SetParametersFromTuple(
  t_Cursor *cursor,
  PyObject *parameters,
  int arraySize,
  unsigned arrayPos)
{
  PyObject *value;
  t_Array *var;
  int pos;
  int paramCount;

  trace(("Cursor_SetParametersFromTuple: cursor %08x row %i\n",
    cursor, arrayPos));

  if (!cursor->bindArrays)
  {
    if (Cursor_DefineQmarkParameters(cursor, arraySize) < 0)
      return -1;
  }
  
  paramCount = PyTuple_Size(cursor->bindArrays);
  
  for (pos = 0; pos < paramCount; pos++)
  {
    var = (t_Array *) PyTuple_GetItem(cursor->bindArrays, pos);
    value = PyTuple_GetItem(parameters, pos);
    if (!value || !var)
    {
      PyErr_Format(g_InterfaceErrorException,
        "Cursor_SetParametersFromTuple: expecting tuple of %d items",
	PyTuple_Size(cursor->bindArrays));
      return -1;
    }
    if (!Array_Check(value))
    {
      if (Array_SetValue(var, arrayPos, value) < 0)
        return -1;
    }
  }

  return 0;
}


/**
 * Create or set bind Arrays from list.
 */
static int Cursor_SetParametersFromList(                     /* 08.08.08 ... */
  t_Cursor *cursor,
  PyObject *parameters,
  int arraySize,
  unsigned arrayPos)
{
  PyObject *value;
  t_Array *var;
  int pos;
  int paramCount;

  trace(("Cursor_SetParametersFromList: cursor %08x row %i\n",
        cursor, arrayPos));

  if (!cursor->bindArrays)
  {
    if (Cursor_DefineQmarkParameters(cursor, arraySize) < 0)
      return -1;
  }

  paramCount = PyTuple_Size(cursor->bindArrays);             /* 18.08.08 */

  for (pos = 0; pos < paramCount; pos++)
  {
    var = (t_Array *) PyTuple_GetItem(cursor->bindArrays, pos);
    value = PyList_GetItem(parameters, pos);
    if (!value || !var)
    {
      PyErr_Format(g_InterfaceErrorException,
        "Cursor_SetParametersFromList: expecting list of %d items",
	PyList_Size(cursor->bindArrays));
      return -1;
    }
    if (!Array_Check(value))
    {
      if (Array_SetValue(var, arrayPos, value) < 0)
        return -1;
    }
  }

  return 0;
}                                                            /* ... 08.08.08 */

/*
Setting parameters from function setvalue
*/
int Cursor_SetVarsFromTuple(                                 /* 25.12.13 ... */
  t_Cursor *cursor,
  PyObject *parameters,
  unsigned colNum)
{
  PyObject *value;
  t_Array *var;
  int pos;
  int rowCount;

  trace(("Cursor_SetParametersFromTuple: cursor %08x row %i\n",
    cursor, colNum));

  if (!cursor->bindArrays)
  {
    if (Cursor_DefineQmarkParameters(cursor, cursor->arraySize) < 0)
      return -1;
  }

  rowCount = PyTuple_Size(parameters);

  var = (t_Array *) PyTuple_GetItem(cursor->bindArrays, colNum);

  for (pos = 0; pos < rowCount; pos++)
  {
    value = PyTuple_GetItem(parameters, pos);
    if (!value || !var)
    {
      PyErr_Format(g_InterfaceErrorException,
        "Cursor_SetParametersFromTuple: expecting tuple of %d items", PyTuple_Size(cursor->bindArrays));
      return -1;
    }
    if (!Array_Check(value))
    {
      if (Array_SetValue(var, pos, value) < 0)
        return -1;
    }
  }


  return 0;
}                                                            /* ... 25.12.13 */
//#endif

/**
 * Perform the binds on the cursor.
 */
static int Cursor_PerformBind(
  t_Cursor *cursor)
{
  PyObject *key;
  PyObject *var;
  int pos;
  int paramCount;

  if (cursor->bindArrays)
  {
    if (PyDict_Check(cursor->bindArrays))
    {
      pos = 0;
      while (PyDict_Next(cursor->bindArrays, &pos, &key, &var))
      {
        if (Array_BindNamedParameter((t_Array *) var, cursor, key) < 0)
          return -1;
      }
    }
    else if (PyTuple_Check(cursor->bindArrays))
    {
      paramCount = PyTuple_Size(cursor->bindArrays);
      for (pos = 0; pos < paramCount; pos++)
      {
        var = PyTuple_GetItem(cursor->bindArrays, pos);
	      if (!var)
	      {
          PyErr_Format(g_InterfaceErrorException,
            "Cursor_PerformBind: expecting tuple of %d items",
	          PyTuple_Size(cursor->bindArrays));
          return -1;
	      }
	      if (Array_BindQmarkParameter((t_Array *) var, cursor, pos + 1) < 0)
	        return -1;
      }
    }
  }

  return 0;
}


/**
 * Determine if the cursor is executing a select statement.
 */
static int Cursor_IsQuery(t_Cursor *cursor)
{
  L_LONG length;
  L_LONG statementType;
  int status;

  length = sizeof(statementType);
  status = LINTER_GetCursorOption(
    cursor->handle,
    cStmtType,
    1,
    &statementType,
    &length);

  if (Cursor_CheckForError(cursor, status) < 0)
    return -1;

  cursor->statementType = statementType;

  return 0;
}


/**
 * Determine if the cursor is executing a select statement.
 */
static int Cursor_GetStatementInfo(t_Cursor *cursor)
{
  L_LONG length;
  L_LONG statementType;
  int status;

  length = sizeof(statementType);
  status = LINTER_GetStatementOption(
    cursor->statementHandle,
    sStmtType,
    1,
    &statementType,
    &length);

  if (Cursor_CheckForError(cursor, status) < 0)
    return -1;

  cursor->statementType = statementType;

  return 0;
}


/**
 * Create a tuple consisting of each of the items in the select-list.
 */
static PyObject *Cursor_CreateTuple(t_Cursor *cursor)
{
  PyObject *tuple;
  PyObject *item;
  int numItems;
  int pos;
  t_Array *var;

  trace(("Cursor_CreateTuple: cursor %08x row %i\n", cursor, cursor->rowNum));

  numItems = PyList_GET_SIZE(cursor->fetchArrays);
  tuple = PyTuple_New(numItems);
  if (!tuple)
    return NULL;

  for (pos = 0; pos < numItems; pos++)
  {
    var = (t_Array*) PyList_GET_ITEM(cursor->fetchArrays, pos);
    item = Array_GetValue(var, cursor->rowNum);
    if (!item)
    {
      Py_DECREF(tuple);
      return NULL;
    }
    PyTuple_SET_ITEM(tuple, pos, item);
  }

  cursor->rowNum++;
  return tuple;
}


/**
 * Internal method for preparing a statement for execution.
 */
static int Cursor_InternalPrepare(
  t_Cursor *cursor,
  PyObject *statement)
{
  int status;

  if (cursor->handleFromSP)
  {
    LINTER_CloseCursor(cursor->handleFromSP);
    cursor->handleFromSP = 0;
  }

  if (cursor->statement)
  {
    Py_DECREF(cursor->statement);
    cursor->statement = NULL;
  }

  if (cursor->fetchArrays)
  {
    Py_DECREF(cursor->fetchArrays);
    cursor->fetchArrays = NULL;
  }

  if (cursor->bindArrays)
  {
    Py_DECREF(cursor->bindArrays);
    cursor->bindArrays = NULL;
  }

  cursor->statement = statement;
  if (PyUnicode_Check(statement)) {                          /* 25.12.13 ... */
    statement = PyUnicode_AsEncodedString(statement, NULL, NULL); } 
  status = LINTER_CreateStatement(
    cursor->connection->handle,
    PyString_AS_STRING(statement),
    PyString_GET_SIZE(statement),
    &cursor->statementHandle);

  trace(("Cursor_InternalPrepare: cursor %08x sql \"%s\"\n",
    cursor, PyString_AS_STRING(statement)));                  /* ... 25.12.13 */
  if (Cursor_CheckForError(cursor, status) < 0)
    return -1;

  Py_INCREF(cursor->statement);

  if (Cursor_GetStatementInfo(cursor) < 0)
    return -1;

  return 0;
}


/**
 * Prepare the statement for execution.
 */
static PyObject *Cursor_Prepare(t_Cursor *cursor, PyObject *args)
{
  PyObject *statement;

  if (!PyArg_ParseTuple(args, "O", &statement))              /* 25.12.13 */
    return NULL;

  if (Cursor_IsOpen(cursor) < 0)
    return NULL;

  if (Cursor_InternalPrepare(cursor, statement) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


/**
 * Set the rowcount Array.
 */
static int Cursor_GetCursorInfo(
  t_Cursor *cursor)
{
  L_LONG rowCount;
  L_LONG length;
  L_LONG isCursor = 0;
  int status;

  trace(("Cursor_GetCursorInfo: cursor %08x", cursor));

  if (cursor->statementType == syExecute)
  {
    length = sizeof(isCursor);
    status = LINTER_GetCursorOption(
      cursor->handle,
      cProcCursorArg,
      (L_SWORD) 0,
      &isCursor,
      &length);
    if (Cursor_CheckForError(cursor, status))
      return -1;

    if (isCursor)
    {
      length = sizeof(L_SWORD);
      status = LINTER_GetData(
        cursor->handle,
        (L_SWORD) 0,
        &cursor->handleFromSP,
        sizeof(L_SWORD),
        tSmallInt,
        0,
        0,
        &length);
      if (Cursor_CheckForError(cursor, status))
        return -1;
    }
  }

  trace((" isCursor %i", isCursor));

  length = sizeof(rowCount);
  status = LINTER_GetCursorOption(
    (L_WORD)((cursor->handleFromSP) ? cursor->handleFromSP : cursor->handle),
    cRowCount,
    0,
    &rowCount,
    &length);
  trace((" rowCount %i\n", rowCount));
  if (Cursor_CheckForError(cursor, status))
    return -1;

  cursor->rowCount = rowCount;
  cursor->fetchOffset = 1;

  return 0;
}


/**
 *
 */
static PyObject *Cursor_GetOutputParameters(
  t_Cursor *cursor)                                          /* 25.12.13 */
{
  L_LONG length;
  L_LONG paramCount;
  L_LONG paramMode;
  int status;
  PyObject *param;
  PyObject *outParams;
  PyObject *var;
  int i, out_cnt;                                            /* 25.12.13 ... */
  out_cnt = 0;                                               /* ... 25.12.13 */

  trace(("Cursor_GetOutputParameters: cursor %08x\n", cursor));

  length = sizeof(paramCount);
  status = LINTER_GetStatementOption(
    cursor->statementHandle,
    sParamCount,
    0,
    &paramCount,
    &length);
  trace((" paramCount %i", paramCount));
  if (Cursor_CheckForError(cursor, status))
    return NULL;

  if (paramCount <= 0)
  {
    trace((" none\n"));
    Py_INCREF(Py_None);
    return Py_None;
  }

  outParams = PyTuple_New(paramCount);
  if (!outParams)
    return NULL;

  for (i = 0; i < paramCount; i++)
  {
    length = sizeof(paramMode);
    status = LINTER_GetStatementOption(
      cursor->statementHandle,
      sParamType,
      (L_SWORD) (i+1),
      &paramMode,
      &length);
    trace((" param %i\n", (i+1)));
    if (Cursor_CheckForError(cursor, status))
    {
      Py_DECREF(outParams);
      return NULL;
    }

    if (paramMode == pInput)
    {
      trace((" mode input\n"));
      var = PyTuple_GET_ITEM(cursor->bindArrays, i);         /* 25.12.13 */
      if (!var)
      {
        Py_DECREF(var);
        return NULL;
      }
      param = Array_GetValue((t_Array*) var, 0);
      if (!param)
      {
        Py_DECREF(outParams);
        return NULL;
      }
      /*Py_INCREF(param);*/
    }
    else
    {
      out_cnt++;
      trace((" mode output\n)"));
      var = (PyObject *) Array_NewParameter(cursor, out_cnt, 1); /* 25.12.13 */     
      if (!var)
      {
        Py_DECREF(outParams);
        return NULL;
      }

      if (Array_GetData((t_Array*) var, cursor) < 0)
      {
        Py_DECREF(outParams);
        return NULL;
      }

      param = Array_GetValue((t_Array*) var, 0);
      if (!param)
      {
        Py_DECREF(outParams);
        return NULL;
      }
      Py_DECREF(var);
    }
    PyTuple_SET_ITEM(outParams, i, param);
  }

  return outParams;
}


/**
 * Call a stored procedure and return the (possibly modified) arguments.
 */
static PyObject *Cursor_CallProc(
  t_Cursor *cursor,
  PyObject *args)
{
  PyObject *listOfArguments;
  PyObject *temp;
  PyObject *results;
  int i, procedureNameLength, numArguments;
  char *procedureName;
  char *statement;
  char *ptr;

  trace(("Cursor_CallProc: cursor %08x", cursor));

  numArguments = 0;
  listOfArguments = NULL;
  if (!PyArg_ParseTuple(args, "s#|O", &procedureName, &procedureNameLength,
                        &listOfArguments))
  {
    return NULL;
  }

  trace((" proc \"%s\"\n", procedureName));

  if (listOfArguments)
  {
    if (!PySequence_Check(listOfArguments))
    {
      PyErr_SetString(PyExc_TypeError, "arguments must be a sequence");
      return NULL;
    }
    numArguments = PySequence_Length(listOfArguments);     /* 31.10.03 */
    if (numArguments < 0)
      return NULL;
  }

  if (Cursor_IsOpen(cursor) < 0)
    return NULL;

  statement = PyMem_Malloc(procedureNameLength + numArguments * 2 + 12);
  if (!statement)
  {
    return PyErr_NoMemory();
  }

  strcpy(statement, "execute ");
  strcat(statement, procedureName);
  ptr = statement + strlen(statement);
  *ptr++ = '(';
  for (i = 0; i < numArguments; i++)
  {
    if (i > 0)
      *ptr++ = ',';
    *ptr++ = '?';
  }
  *ptr = '\0';
  strcat(statement, ")");

  if (listOfArguments)
    temp = Py_BuildValue("(sO)", statement, listOfArguments);
  else
    temp = Py_BuildValue("(s)", statement);
  PyMem_Free(statement);
  if (!temp)
    return NULL;

  results = Cursor_Execute(cursor, temp, NULL);
  Py_DECREF(temp);
  if (!results)
    return NULL;
  Py_DECREF(results);

  results = Cursor_GetOutputParameters(cursor);              /* 25.12.13 */
  return results;
}


/**
 * Prepare statement, bind parameters, execute statement and bind answers.
 */
static PyObject *Cursor_Execute(
  t_Cursor *cursor,
  PyObject *args,
  PyObject *keywordArgs)
{
  PyObject *statement;
  PyObject *params;
  int numArgs;
  int status;
  L_LONG executed;

  params = NULL;

  numArgs = PyTuple_GET_SIZE(args);
  if (numArgs < 1)
  {
    PyErr_SetString(PyExc_TypeError, "expected at least 1 argument, got 0");
    return NULL;
  }
  if (numArgs > 2)
  {
    PyErr_Format(PyExc_TypeError, "expected at most 2 arguments, got %d", numArgs);
    return NULL;
  }

   if (!PyArg_ParseTuple(args, "O|O", &statement, &params))  /* 25.12.13 ... */  
   {  
     return NULL; 
   }                                                         /* ... 25.12.13 */

  if ((statement != Py_None) && !PyString_Check(statement) && /* 25.12.13 ... */ /* 06.10.14 */
      !PyUnicode_Check(statement))                           /* ... 25.12.13 */
  {
    PyErr_SetString(PyExc_TypeError, "expecting None or a string");
    return NULL;
  }


  if (params && !PyDict_Check(params) && !PyTuple_Check(params) && !PyList_Check(params)) /* 08.08.08 */
  {
    PyErr_SetString(PyExc_TypeError, "expecting a dictionary, tuple or list");
    return NULL;
  }
  if (params)
  {
    if (keywordArgs && PyDict_Size(keywordArgs))
    {
      PyErr_SetString(g_InterfaceErrorException,
        "expecting argument or keyword arguments, not both");
      return NULL;
    }
  }
  else
  {
    if (keywordArgs && PyDict_Size(keywordArgs))
      params = keywordArgs;
  }

  if (Cursor_IsOpen(cursor) < 0)
    return NULL;

  if ((statement != Py_None) && (statement != cursor->statement))
  {
    status = Cursor_InternalPrepare(cursor, statement);
    if (status < 0)
      return NULL;
  }

  if (params)
  {
    if (PyDict_Check(params))
    {
      if (Cursor_SetParametersFromDict(cursor, params, 1, 0) < 0)
        return NULL;
    }
    else if (PyTuple_Check(params))
    {
      if (Cursor_SetParametersFromTuple(cursor, params, 1, 0) < 0)
        return NULL;
    }
    else if (PyList_Check(params))                               /* 08.08.08 */
    {
      if (Cursor_SetParametersFromList(cursor, params, 1, 0) < 0)
        return NULL;
    }
  }

  if (Cursor_PerformBind(cursor) < 0)
    return NULL;

  executed = 1;

  status = LINTER_ExecuteStatement(
    cursor->handle,
    cursor->statementHandle,
    &executed,
    NULL,
    NULL);

  if (Cursor_CheckForError(cursor, status) < 0)
    return NULL;

  if (Cursor_GetCursorInfo(cursor) < 0)
    return NULL;

  if ((cursor->statementType == sySelect) ||
      (cursor->statementType == syTable) ||                /* 08.02.10 */
      (cursor->statementType == syValues) ||               /* 08.02.10 */
      (cursor->statementType == syTestTable) ||
      (cursor->statementType == syExecute))
  {
    if ((!cursor->fetchArrays) ||
        (cursor->fetchArraySize != cursor->arraySize))
    {
      if (Cursor_PerformDefine(cursor) < 0)
        return NULL;
    }
  }
  Py_INCREF(Py_None);
  return Py_None;
}


/**
 * Execute the statement many times. The number of times is equivalent to the
 * number of elements in the array of dictionaries.
 */
static PyObject *Cursor_ExecuteMany(
  t_Cursor *cursor,
  PyObject *args)
{
  PyObject *arguments;
  PyObject *listOfArguments;
  PyObject *statement;
  int i;
  int numRows;
  int status;
  L_LONG executed;

  if (!PyArg_ParseTuple(args, "OO!", &statement, &PyList_Type, &listOfArguments)) /* 25.12.13 */
    return NULL;

  if (Cursor_IsOpen(cursor) < 0)
    return NULL;

  if (Cursor_InternalPrepare(cursor, statement) < 0)
    return NULL;

  if ((cursor->statementType == sySelect) ||
      (cursor->statementType == syTable) ||                /* 08.02.10 */
      (cursor->statementType == syValues) ||               /* 08.02.10 */
      (cursor->statementType == syTestTable) ||
      (cursor->statementType == syExecute))
  {
    PyErr_SetString(g_NotSupportedErrorException,
      "queries not supported: results undefined");
    return NULL;
  }

  numRows = PyList_GET_SIZE(listOfArguments);
  for (i = 0; i < numRows; i++)
  {
    arguments = PyList_GET_ITEM(listOfArguments, i);
    if (PyDict_Check(arguments))
    {
      if (Cursor_SetParametersFromDict(cursor, arguments, numRows, i) < 0)
        return NULL;
    }
    else if (PyTuple_Check(arguments))
    {
      if (Cursor_SetParametersFromTuple(cursor, arguments, numRows, i) < 0)
        return NULL;
    }
    else if (PyList_Check(arguments))                            /* 08.08.08 */
    {
      if (Cursor_SetParametersFromList(cursor, arguments, numRows, i) < 0)
        return NULL;
    }
    else
    {
      PyErr_SetString(g_InterfaceErrorException,
        "expecting a list of dictionaries, tuples or lists");  /* 25.12.13 */
      return NULL;
    }
  }

  if (Cursor_PerformBind(cursor) < 0)
      return NULL;

  executed = PyList_GET_SIZE(listOfArguments);
  status = LINTER_ExecuteStatement(
    cursor->handle,
    cursor->statementHandle,
    &executed,
    NULL,
    NULL);
  if (Cursor_CheckForError(cursor, status) < 0)
    return NULL;

#if 0
  if (Cursor_GetCursorInfo(cursor) < 0)
      return NULL;
#else
  cursor->rowCount = -1;
  cursor->fetchOffset = 1;
#endif

  Py_INCREF(Py_None);
  return Py_None;
}


/**
 * Execute the prepared statement the number of times requested. At this
 * point, the statement must have been already prepared and the bind Arrays
 * must have their values set.
 */
static PyObject *Cursor_ExecuteManyPrepared(
  t_Cursor *cursor,
  PyObject *args)
{
  L_LONG numIters;
  int status;

  /* expect number of times to execute the statement */
  if (!PyArg_ParseTuple(args, "i", &numIters))
    return NULL;

  if (numIters > cursor->arraySize)
  {
    PyErr_SetString(g_InterfaceErrorException,
      "iterations exceed array size");
    return NULL;
  }

  if (Cursor_IsOpen(cursor) < 0)
    return NULL;

  if ((cursor->statementType == sySelect) ||
      (cursor->statementType == syTable) ||                /* 08.02.10 */
      (cursor->statementType == syValues) ||               /* 08.02.10 */
      (cursor->statementType == syTestTable) ||
      (cursor->statementType == syExecute))
  {
    PyErr_SetString(g_NotSupportedErrorException,
      "queries not supported: results undefined");
    return NULL;
  }

  if (Cursor_PerformBind(cursor) < 0)
    return NULL;

  status = LINTER_ExecuteStatement(
    cursor->handle,
    cursor->statementHandle,
    &numIters,
    NULL,
    NULL);
  if (Cursor_CheckForError(cursor, status) < 0)
    return NULL;

  if (Cursor_GetCursorInfo(cursor) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


/**
 * Verify that fetching may happen from this cursor.
 */
static int Cursor_VerifyFetch(
  t_Cursor *cursor)
{
  if (Cursor_IsOpen(cursor) < 0)
      return -1;

  if ((cursor->statementType != sySelect) &&
      (cursor->statementType != syTable) &&                /* 08.02.10 */
      (cursor->statementType != syValues) &&               /* 08.02.10 */
      (cursor->statementType != syTestTable) &&
      (cursor->statementType != syExecute))
  {
    PyErr_SetString(g_InterfaceErrorException, "not a query");
    return -1;
  }

  return 0;
}


/**
 *
 */
static int Cursor_NotifyPosChanged(t_Cursor *cursor)
{
  t_Array *array;
  int numItems;
  int i;

  if (cursor->fetchArrays != NULL)
  {
    numItems = PyList_GET_SIZE(cursor->fetchArrays);
    for (i = 0; i < numItems; i++)
    {
      array = (t_Array *) PyList_GET_ITEM(cursor->fetchArrays, i);
      if ((array != NULL) && (array->type != NULL) &&
          (array->type->afterPosChanged != NULL))
      {
        if (array->type->afterPosChanged(array) < 0)
          return -1;
      }
    }
  }
  return 0;
}


/**
 * Performs the actual fetch.
 */
static int Cursor_InternalFetch(t_Cursor *cursor, int numRows)
{
  int status;
  L_LONG actualRows;
  L_LONG apiCode, linCode, sysCode;

  status = LINTER_Fetch((cursor->handleFromSP) ? cursor->handleFromSP : cursor->handle, 
      toAbsNumber, cursor->fetchOffset, numRows, &actualRows, NULL, NULL); /* 25.12.13 */
  if (status != LINAPI_SUCCESS)
  {
    LINTER_Error(0, (cursor->handleFromSP) ? cursor->handleFromSP : cursor->handle, 
	   0, &apiCode, &linCode, &sysCode, NULL, NULL);  /* 25.12.13 */
    if (linCode == 2)
      return 0;
  }
  if (Cursor_CheckForError(cursor, status) < 0)
    return -1;

  cursor->fetchOffset += actualRows;
  cursor->actualRows = actualRows;
  cursor->rowNum = 0;

  if (Cursor_NotifyPosChanged(cursor) < 0)
    return -1;

  return 0;
}


/**
 * Returns an integer indicating if more rows can be retrieved from the
 * cursor.
 */
static int Cursor_MoreRows(t_Cursor *cursor)
{
  if (cursor->rowCount > 0 && cursor->rowNum >= cursor->actualRows) /* 25.12.13 */
  {
    
    if ((cursor->actualRows <= 0) ||
        (cursor->actualRows >= cursor->fetchArraySize))
    {
      if (Cursor_InternalFetch(cursor, cursor->fetchArraySize) < 0)
        return -1;
    }
    if (cursor->rowNum >= cursor->actualRows)
      return 0;
  }
  else if (cursor->rowCount == 0)                            /* 25.12.13 */
    return 0;

  return 1;
}


/**
 * Return a list consisting of the remaining rows up to the given row limit
 */
static PyObject *Cursor_MultiFetch(t_Cursor *cursor, int rowLimit)
{
  PyObject *results;
  PyObject *tuple;
  int row;
  int status;

  results = PyList_New(0);
  if (!results)
    return NULL;

  for (row = 0; (rowLimit == 0) || (row < rowLimit); row++)
  {
    status = Cursor_MoreRows(cursor);
    if (status < 0)
    {
      Py_DECREF(results);
      return NULL;
    }
    else if (status == 0)
    {
      break;
    }
    else
    {
      tuple = Cursor_CreateTuple(cursor);
      if (!tuple)
      {
        Py_DECREF(results);
        return NULL;
      }
      if (PyList_Append(results, tuple) < 0)
      {
        Py_DECREF(tuple);
        Py_DECREF(results);
        return NULL;
      }
      Py_DECREF(tuple);
    }
  }
  return results;
}


/**
 * Fetch a single row from the cursor.
 */
static PyObject *Cursor_FetchOne(t_Cursor *cursor, PyObject *args)
{
  int status;

  trace(("Cursor_FetchOne\n"));

  if (args && !PyArg_ParseTuple(args, ""))
    return NULL;

  if (Cursor_VerifyFetch(cursor) < 0)
    return NULL;

  status = Cursor_MoreRows(cursor);
  if (status < 0)
  {
    return NULL;
  }
  else if (status > 0)
  {
    return Cursor_CreateTuple(cursor);
  }

  Py_INCREF(Py_None);
  return Py_None;
}


/**
 * Fetch multiple rows from the cursor based on the arraysize.
 */
static PyObject *Cursor_FetchMany(t_Cursor *cursor, PyObject *args)
{
  int fetchSize;

  /* optional fetchSize expected */
  fetchSize = cursor->arraySize;
  if (!PyArg_ParseTuple(args, "|i", &fetchSize))
    return NULL;

  if (Cursor_VerifyFetch(cursor) < 0)
    return NULL;

  return Cursor_MultiFetch(cursor, (cursor->blobColumns > 0) ? 1 : fetchSize); /* 13.10.14 */
}


/**
 * Fetch all remaining rows from the cursor.
 */
static PyObject *Cursor_FetchAll(
  t_Cursor *cursor,
  PyObject *args)
{
  if (!PyArg_ParseTuple(args, ""))
    return NULL;

  if (Cursor_VerifyFetch(cursor) < 0)
    return NULL;

  return Cursor_MultiFetch(cursor, (cursor->blobColumns > 0) ? 1 : 0); /* 13.10.14 */
}


/**
 * Set the sizes of the bind Arrays.
 * NOTE: this is dummy
 */
static PyObject *Cursor_SetInputSizes(
  t_Cursor *cursor,
  PyObject *args,
  PyObject *keywordArgs)
{
  if (Cursor_IsOpen(cursor) < 0)
    return NULL;
  Py_INCREF(Py_None);
  return Py_None;
}


/**
 * Set the size of all of the long columns (blob) or just one of them.
 */
static PyObject *Cursor_SetOutputSize(                                           /* 14.10.14 ... */
  t_Cursor *cursor,
  PyObject *args)
{
  int i;
  int size = -1;
  int column = -1;
  int pos = 0;
  t_Array *array;
  PyObject *key;
  PyObject *value;
  PyObject *object;

  if (!PyArg_ParseTuple(args, "O|i", &object, &column))
    return NULL;

  if (object != Py_None)
    size = PyLong_AsLong(object);

  if (object != Py_None && size < 0)
  {
    PyErr_SetString(g_InterfaceErrorException, "size must be positive");
    return NULL;
  }

  if (!cursor->blobOutputSizes)
  {
    cursor->blobOutputSizes = PyDict_New();
    if (!cursor->blobOutputSizes)
    {
      Py_DECREF(cursor->blobOutputSizes);
      return NULL;
    }
  }

  if (column < 0)
  {
    PyDict_Clear(cursor->blobOutputSizes);
    cursor->blobOutputSize = (object == Py_None) ? -1 : size;
  }
  else
  {
    if (PyDict_GetItem(cursor->blobOutputSizes, PyLong_FromLong(column)))
    {
      if (PyDict_DelItem(cursor->blobOutputSizes, PyLong_FromLong(column)) < 0)
      {
        Py_DECREF(cursor->blobOutputSizes);
        return NULL;
      }
    }
    if (object != Py_None && PyDict_SetItem(cursor->blobOutputSizes, PyLong_FromLong(column), PyLong_FromLong(size)) < 0)
    {
      Py_DECREF(cursor->blobOutputSizes);
      return NULL;
    }
    if (PyDict_Size(cursor->blobOutputSizes) > 0)
      cursor->blobOutputSize = -1;
  }

  /* if query has already been executed and there are BLOB columns */
  if (cursor->fetchArrays && cursor->blobColumns > 0)
  {
    for (i = 0; i < PyList_GET_SIZE(cursor->fetchArrays); i++)
    {
      pos = 0;
      array = (t_Array*)PyList_GetItem(cursor->fetchArrays, i);
      if (array->desc.Type == tBlob)
      {
        if (PyDict_Size(cursor->blobOutputSizes) > 0)
        {
          array->blobPortion = -1;
          while (PyDict_Next(cursor->blobOutputSizes, &pos, &key, &value))
          {
            if (array->index == PyLong_AsLong(key))
            {
              array->blobPortion = PyLong_AsLong(value);
              break;
            }
          }
        }
        else
        {
          array->blobPortion = cursor->blobOutputSize;
        }
      }
    }
  }

  Py_INCREF(Py_None);
  return Py_None;
}                                                                                /* ... 14.10.14 */


#if (PY_VERSION_HEX >= 0x02020000)
/**
 * Return a reference to the cursor which supports the iterator protocol.
 */
static PyObject *Cursor_GetIter(
  PyObject *cursor)
{
  Py_INCREF(cursor);
  return cursor;
}

/**
 * Return a reference to the cursor which supports the iterator protocol.
 */
static PyObject *Cursor_GetNext(t_Cursor *cursor)
{
  int status;

  if (Cursor_VerifyFetch(cursor) < 0)
    return NULL;

  status = Cursor_MoreRows(cursor);
  if (status < 0)
    return NULL;
  else if (status > 0)
    return Cursor_CreateTuple(cursor);

  /* no more rows, raise StopIteration as per the protocol */
  Py_INCREF(Py_None);
  PyErr_SetObject(PyExc_StopIteration, Py_None);
  return NULL;
}
#endif


/**
 * Commit cursor.
 */
static PyObject *Cursor_Commit(
  t_Cursor *cursor,
  PyObject *args)
{
  int status;

  if (!PyArg_ParseTuple(args, ""))
    return NULL;

  if (Cursor_IsOpen(cursor) < 0)
    return NULL;

  status = LINTER_CommitCursor(cursor->handle, NULL, NULL);
  if (Cursor_CheckForError(cursor, status) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


/**
 * Rollback cursor.
 */
static PyObject *Cursor_Rollback(
  t_Cursor *cursor,
  PyObject *args)
{
  int status;

  if (!PyArg_ParseTuple(args, ""))
    return NULL;

  if (Cursor_IsOpen(cursor) < 0)
    return NULL;

  status = LINTER_RollBackCursor(cursor->handle, NULL, NULL);
  if (Cursor_CheckForError(cursor, status) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *Cursor_ExternalSetValue(
    t_Cursor *cursor,                  // variable to set
    PyObject *args)                    // arguments
{
  PyObject *value;
  unsigned pos;

  if (!PyArg_ParseTuple(args, "iO", &pos, &value))
    return NULL;

  cursor->arraySize = PyTuple_Size(value);

  if (PyTuple_Check(value))
     if (Cursor_SetVarsFromTuple(cursor, value, pos) < 0)
        return NULL;


  Py_INCREF(Py_None);
  return Py_None;
 }

#if PY_VERSION_HEX >= 0x02060000
/**
 * Create an array bind variable and return it.
 */
static PyObject *Cursor_ArrayVar(                            /* 25.12.13 ... */
    t_Cursor *cursor,                   // cursor to fetch from
    PyObject *args)                     // arguments
{
  t_ArrayType *varType;
  PyObject *type;
  t_Array *array;


  // parse arguments
  if (!PyArg_ParseTuple(args, "O", &type))
      return NULL;

  // determine the type of variable
  varType = Array_TypeByPythonType(type);

  if (!varType)
      return NULL;

  array = (t_Array*) varType->pythonType->tp_alloc(varType->pythonType, 0);
  if (array == NULL)
    return NULL;


  array->type = varType;

  array->boundCursor = cursor;

  return (PyObject*) array;
}
#endif                                                           /* 25.12.13 ... */
