 /***************************************************************************
 *                 RELEX, Inc. Linter Python driver source code             *
 *        This is unpublished proprietary source code of RELEX, Inc.        *
 *     The copyright notice does not evidence any actual or intended        *
 *                  publication of such source code.                        *
 *                                                                          *
 *               Copyright (c) 2003-2014 RELEX Inc.                         *
 *                                                                          *
 *                   This software was developed by                         *
 *                                                                          *
 *                            RELEX, Inc.                                   *
 *                         Voronezh, Russia                                 *
 *                                                                          *
 *                        All rights reserved.                              *
 ****************************************************************************
 AUTHOR: Boris Fedorov
 ----------------------------------------------------------------------------
 PURPOSE: driver entry point (DBAPI)
 ----------------------------------------------------------------------------
 REVISION HISTORY:
 25.07.07 -- Galizin ( time.h has been included for localtime call )
 11.08.08 -- Ermakov ( initialization g_TimestampArrayType added)
 04.09.08 -- Ermakov ( added native DATETIME support (2.4 and higher) )
 22.09.08 -- Ermakov (Added date and unicode configuration)
 25.12.13 -- Muratova (Python 3 support)
 06.10.14 -- Kononov (Fixed support for Python < 2.6)
*/

#include <time.h>
#undef _DEBUG
#include <Python.h>
#include <linapi.h>
#include "LinPy.h"

// define whether or not we are building with the native datetime module
PyObject *g_WarningException = NULL;
PyObject *g_ErrorException = NULL;
PyObject *g_InterfaceErrorException = NULL;
PyObject *g_DatabaseErrorException = NULL;
PyObject *g_DataErrorException = NULL;
PyObject *g_OperationalErrorException = NULL;
PyObject *g_IntegrityErrorException = NULL;
PyObject *g_InternalErrorException = NULL;
PyObject *g_ProgrammingErrorException = NULL;
PyObject *g_NotSupportedErrorException = NULL;
#if PY_VERSION_HEX < 0x02060000
  PyObject *g_ModuleDictionary = NULL;
#endif
PyObject *g_DateEmulation = NULL;                            /* 22.09.08 ... */
PyObject *g_UnicodeData = NULL;

/**
 * Set UnicodeData variable
 */
static PyObject* SetUnicodeData(PyObject* self, PyObject* args)
{
  int val;

  if (!PyArg_ParseTuple(args, "i", &val))
    return NULL;
  Py_DECREF(g_UnicodeData);
  if (val)
    g_UnicodeData = PyInt_FromLong(1);
  else
    g_UnicodeData = PyInt_FromLong(0);
  return g_UnicodeData;
}

/**
 * Set DateEmulation variable
 */
static PyObject* SetDateEmulation(PyObject* self, PyObject* args)
{
  int val;

  if (!PyArg_ParseTuple(args, "i", &val))
    return NULL;
  Py_DECREF(g_DateEmulation);
  if (val)
    g_DateEmulation = PyInt_FromLong(1);
  else
    g_DateEmulation = PyInt_FromLong(0);
  return g_DateEmulation;
}                                                            /* ... 22.09.08 */

/**
 * DBAPI 2.0 specification require Binary constructor
 */
static PyObject *Binary(PyObject* self, PyObject* args)
{
  char *str;
  int strLength;
  PyObject *obj;

  if (!PyArg_ParseTuple(args, "s#", &str, &strLength))
    return NULL;

  obj = PyString_FromStringAndSize(str, strLength);

  return obj;
}


/**
 * Returns a date value suitable for binding.
 */
static PyObject* Date(PyObject* self, PyObject* args)
{
  int year, month, day;

  if (!PyArg_ParseTuple(args, "iii", &year, &month, &day))
    return NULL;

  return Timestamp_New(1, year, month, day, 0, 0, 0, 0);       /* 22.09.08 */
}

/**
 * Returns a time value suitable for binding.
 */
static PyObject* Time(PyObject* self, PyObject* args)
{
  int hour, minute, second, tick = 0;

  if (!PyArg_ParseTuple(args, "iii|i", &hour, &minute, &second, &tick))
    return NULL;

  return Timestamp_New(2, 0, 0, 0, hour, minute, second, tick); /* 22.09.08 */
}


/**
 * Returns a time value suitable for binding.
 */
static PyObject* TimeFromTicks(PyObject* self, PyObject* args)
{
  struct tm *time = NULL;                                    /* 25.12.13 ... */
  time_t ticks;
  double t;       

  if (!PyArg_ParseTuple(args, "d", &t))
    return NULL;
    
  ticks = (time_t)t;                                         /* ... 25.12.13 */

  time = localtime(&ticks);

  if (time == NULL)
    return PyErr_SetFromErrno(g_DataErrorException);

  return Timestamp_New(2, 0, 0, 0,                       
                       time->tm_hour,
                       time->tm_min,
                       time->tm_sec, 0);                     /* 22.09.08 */
}

/**
 * Returns a date value suitable for binding.
 */
static PyObject* Timestamp(PyObject* self, PyObject* args)
{
  int year, month, day, hour, minute, second, tick = 0;

  if (!PyArg_ParseTuple(args, "iiiiii|i", &year, &month, &day, &hour,
                        &minute, &second, &tick))
    return NULL;

  return Timestamp_New(0, year, month, day, hour, minute, second, tick); /* 22.09.08 */
}

/**
 * Returns a date value suitable for binding.
 */
static PyObject* DateFromTicks(PyObject* self, PyObject* args)
{
  struct tm *time = NULL;                                    /* 25.12.13 ... */
  time_t ticks;
  double t;

  if (!PyArg_ParseTuple(args, "d", &t))
    return NULL;

  ticks = (time_t)t;                                         /* ... 25.12.13 */
  time = localtime(&ticks);
  if (time == NULL)
    return PyErr_SetFromErrno(g_DataErrorException);

  return Timestamp_New(1, time->tm_year + 1900,
                       time->tm_mon + 1,
                       time->tm_mday,
                       0, 0, 0, 0);                        /* 22.09.08 */
}



/**
 * Returns a date value suitable for binding.
 */
static PyObject* TimestampFromTicks(PyObject* self, PyObject* args)
{
  struct tm *time = NULL;                                    /* 25.12.13 ... */
  time_t ticks;
  double t;

  if (!PyArg_ParseTuple(args, "d", &t))
    return NULL;

  ticks = (time_t)t;                                         /* ... 25.12.13 */
  time = localtime(&ticks);
  if (time == NULL)
    return PyErr_SetFromErrno(g_DataErrorException);

  return Timestamp_New(0,
    time->tm_year + 1900,
    time->tm_mon + 1,
    time->tm_mday,
    time->tm_hour,
    time->tm_min,
    time->tm_sec,
    0);                                                /* 22.09.08 */
}

/*
 * Module definition
 */

static PyMethodDef g_ModuleMethods[] =
{
#if PY_VERSION_HEX < 0x02060000
  {"connect", (PyCFunction) Connection_Connect, METH_VARARGS | METH_KEYWORDS, "Create connection"},
#endif
  {"Binary", Binary, METH_VARARGS},
  {"Date", Date, METH_VARARGS },
  {"Time", Time, METH_VARARGS},
  {"TimeFromTicks", TimeFromTicks, METH_VARARGS},
  {"Timestamp", Timestamp, METH_VARARGS},
  {"DateFromTicks", DateFromTicks, METH_VARARGS},
  {"TimestampFromTicks", TimestampFromTicks, METH_VARARGS},
  {"SetUnicodeData", SetUnicodeData, METH_VARARGS},              /* 22.08.09 */
  {"SetDateEmulation", SetDateEmulation, METH_VARARGS},          /* 22.09.08 */
  {NULL, NULL, 0, NULL}
};

#if PY_MAJOR_VERSION >= 3                                    /* 25.12.13 ... */
    struct PyModuleDef moduledef = {
        PyModuleDef_HEAD_INIT,
        "LinPy",                                             /* m_name */ 
        "LinPy is module for RDBMS Linter",                  /* m_doc */
        -1,                                                  /* m_size */
        g_ModuleMethods,                                     /* m_methods */
        NULL,                                                /* m_reload */
        NULL,                                                /* m_traverse */
        NULL,                                                /* m_clear */      
        NULL,                                                /* m_free */
    };
#endif                                                       /* ... 25.12.13 */

static int SetException(
  PyObject *module, 
  PyObject **exception,
  char *name,
  PyObject *baseException)
{
  char buffer[100];
  sprintf(buffer, "LinPy.%s", name);
  *exception = PyErr_NewException(buffer, baseException, NULL);
  if (!*exception)
    return -1;
#if PY_VERSION_HEX < 0x02060000                             /* 06.10.14 ... */
  return PyDict_SetItemString(module, name, *exception);
#else
  return PyModule_AddObject(module, name, *exception);       /* 25.12.13 */
#endif                                                       /* ... 06.10.14 */
}

/*
#if PY_MAJOR_VERSION >= 3
PyObject * PyInit_LinPy(void)
#else
void initLinPy(void)
#endif
*/
/* declarations for DLL import/export */                     /* 25.12.13 ...*/
#ifndef PyMODINIT_FUNC	
#define PyMODINIT_FUNC void
#endif

PyMODINIT_FUNC
INIT_LINPY(void)                                             /* ... 25.12.13 */
{
  PyObject *module;

  TimeStampInit();                                               /* 04.09.08 */
  if (PyErr_Occurred())                                          /* 04.09.08 */
    INITERROR;

  #if PY_VERSION_HEX < 0x02060000
    g_ErrorType.ob_type = &PyType_Type;
    g_ConnectionType.ob_type = &PyType_Type;
    g_CursorType.ob_type = &PyType_Type;
    g_BinaryType.ob_type = &PyType_Type;
    g_StringType.ob_type = &PyType_Type;
    g_BlobType.ob_type = &PyType_Type;
    g_TimestampType.ob_type = &PyType_Type;
    g_TimestampArrayType.ob_type = &PyType_Type;                   /* 11.08.08 */
    g_NumberType.ob_type = &PyType_Type;
  #else                                                            /* ... 25.12.13 */
    g_ErrorType.tp_new = PyType_GenericNew;
    (g_ErrorType).tp_base = &g_ArrayType;
	if (PyType_Ready(&g_ErrorType) < 0)
		INITERROR;

    g_ConnectionType.tp_new = PyType_GenericNew;
	if (PyType_Ready(&g_ConnectionType) < 0)
		INITERROR;
    
    g_CursorType.tp_new = PyType_GenericNew;
	if (PyType_Ready(&g_CursorType) < 0)
		INITERROR;
    
    g_BinaryType.tp_new = PyType_GenericNew;
	if (PyType_Ready(&g_BinaryType) < 0)
		INITERROR;
    
    g_TimestampType.tp_new = PyType_GenericNew;
	if (PyType_Ready(&g_TimestampType) < 0)
		INITERROR;
    
    g_TimestampArrayType.tp_new = PyType_GenericNew;
	if (PyType_Ready(&g_TimestampArrayType) < 0)
		INITERROR;    
    
    g_NumberType.tp_new = PyType_GenericNew;
    g_NumberType.tp_base = &g_ArrayType;
	if (PyType_Ready(&g_NumberType) < 0)
		INITERROR;   

    g_StringType.tp_new = PyType_GenericNew;
	if (PyType_Ready(&g_StringType) < 0)
		INITERROR;   

    g_BlobType.tp_new = PyType_GenericNew;
	if (PyType_Ready(&g_BlobType) < 0)
		INITERROR;                                                     /* ... 25.12.13 */
  #endif
  
  g_DateEmulation = PyInt_FromLong(0);                           /* 22.09.08 */
  g_UnicodeData = PyInt_FromLong(0);                             /* 22.09.08 */

  #if PY_MAJOR_VERSION >= 3
    module = PyModule_Create(&moduledef);
  #else
    module = Py_InitModule("LinPy", g_ModuleMethods);
  #endif  

  if (module == NULL)
    INITERROR;

  #if PY_VERSION_HEX < 0x02060000                                                                  /* 06.10.14 ... */
    g_ModuleDictionary = PyModule_GetDict(module);
    if (g_ModuleDictionary == NULL)
      INITERROR;
    /* setup exceptions */
    if (SetException(g_ModuleDictionary, &g_WarningException, "Warning", PyExc_StandardError) < 0)
      INITERROR;
    if (SetException(g_ModuleDictionary, &g_ErrorException, "Error", PyExc_StandardError) < 0)
      INITERROR;
    if (SetException(g_ModuleDictionary, &g_InterfaceErrorException, "InterfaceError", g_ErrorException) < 0)
      INITERROR;
    if (SetException(g_ModuleDictionary, &g_DatabaseErrorException, "DatabaseError", g_ErrorException) < 0)
      INITERROR;
    if (SetException(g_ModuleDictionary, &g_DataErrorException, "DataError", g_DatabaseErrorException) < 0)
      INITERROR;
    if (SetException(g_ModuleDictionary, &g_OperationalErrorException, "OperationalError", g_DatabaseErrorException) < 0)
      INITERROR;
    if (SetException(g_ModuleDictionary, &g_IntegrityErrorException, "IntegrityError", g_DatabaseErrorException) < 0)
      INITERROR;
    if (SetException(g_ModuleDictionary, &g_InternalErrorException, "InternalError", g_DatabaseErrorException) < 0)
      INITERROR;
    if (SetException(g_ModuleDictionary, &g_ProgrammingErrorException, "ProgrammingError", g_DatabaseErrorException) < 0)
      INITERROR;
    if (SetException(g_ModuleDictionary, &g_NotSupportedErrorException, "NotSupportedError", g_DatabaseErrorException) < 0)
      INITERROR;

    /* constants required by Python DB API 2.0 */
    if (PyDict_SetItemString(g_ModuleDictionary, "apilevel", PyString_FromString("2.0")) < 0)
      INITERROR;
    if (PyDict_SetItemString(g_ModuleDictionary, "threadsafety", PyInt_FromLong(1)) < 0)
      INITERROR;
    if (PyDict_SetItemString(g_ModuleDictionary, "paramstyle", PyString_FromString("qmark")) < 0)
      INITERROR;
    /* version for support */
    if (PyDict_SetItemString(g_ModuleDictionary, "version", PyString_FromString("1.0")) < 0)
      INITERROR;
    /* threads may share the module, but not connections */
    if (PyDict_SetItemString(g_ModuleDictionary, "threadsafety", PyInt_FromLong(1)) < 0)
      INITERROR;

    /* setup types */
    if (PyDict_SetItemString(g_ModuleDictionary, "BINARY", (PyObject*)&g_BinaryType) < 0)
      INITERROR;
    if (PyDict_SetItemString(g_ModuleDictionary, "TIMESTAMP", (PyObject*)&g_TimestampType) < 0)
      INITERROR;
    if (PyDict_SetItemString(g_ModuleDictionary, "DATETIME", (PyObject*)&g_TimestampArrayType) < 0)
      INITERROR;
    if (PyDict_SetItemString(g_ModuleDictionary, "NUMBER", (PyObject*)&g_NumberType) < 0)
      INITERROR;
    if (PyDict_SetItemString(g_ModuleDictionary, "STRING", (PyObject*)&g_StringType) < 0)
      INITERROR;
    if (PyDict_SetItemString(g_ModuleDictionary, "BLOB", (PyObject*)&g_BlobType) < 0)
      INITERROR;

    InitTimeStampObject();

    /* modes for connection and cursors */
    if (PyDict_SetItemString(g_ModuleDictionary, "M_EXCLUSIVE", PyInt_FromLong(mExclusive)) < 0)
      INITERROR;
    if (PyDict_SetItemString(g_ModuleDictionary, "M_OPTIMISTIC", PyInt_FromLong(mOptimistic)) < 0)
      INITERROR;
    if (PyDict_SetItemString(g_ModuleDictionary, "M_AUTOCOMMIT", PyInt_FromLong(mAutocommit)) < 0)
      INITERROR;
  #else                                                                                            /* ... 06.10.14 */
    /* setup exceptions */
    if (SetException(module, &g_WarningException,                           /* 25.12.13 ... */
              "Warning", PyExc_StandardError) < 0)
      INITERROR;

    if (SetException(module, &g_ErrorException,
              "Error", PyExc_StandardError) < 0)
      INITERROR;

    if (SetException(module, &g_InterfaceErrorException,
              "InterfaceError", g_ErrorException) < 0)
      INITERROR;

    if (SetException(module, &g_DatabaseErrorException,
              "DatabaseError", g_DatabaseErrorException) < 0)
      INITERROR;

    if (SetException(module, &g_DataErrorException,
              "DataError", g_DatabaseErrorException) < 0)
      INITERROR;

    if (SetException(module, &g_OperationalErrorException,
             "OperationalError", g_DatabaseErrorException) < 0)
      INITERROR;

    if (SetException(module, &g_IntegrityErrorException,
             "IntegrityError", g_DatabaseErrorException) < 0)
      INITERROR;

    if (SetException(module, &g_InternalErrorException,
             "InternalError", g_DatabaseErrorException) < 0)
      INITERROR;

    if (SetException(module, &g_ProgrammingErrorException,
             "ProgrammingError", g_DatabaseErrorException) < 0)
      INITERROR;

    if (SetException(module, &g_NotSupportedErrorException,
             "NotSupportedError", g_DatabaseErrorException) < 0)
      INITERROR;
    ADD_OBJECT("Error", &g_ErrorType);
    ADD_OBJECT("connect", &g_ConnectionType);
    ADD_OBJECT("cursor", &g_CursorType);
    ADD_OBJECT("BINARY", &g_BinaryType); 
    ADD_OBJECT("TIMESTAMP", &g_TimestampType);
    ADD_OBJECT("DATETIME", &g_TimestampArrayType);
    ADD_OBJECT("NUMBER", &g_NumberType);
    ADD_OBJECT("STRING", &g_StringType);
    ADD_OBJECT("Blob", &g_BlobType);

    /* constants required by Python DB API 2.0 */
    if (PyModule_AddStringConstant(module, "apilevel", "2.0") < 0)
      INITERROR;

    /* threads may share the module, but not connections */
    if (PyModule_AddIntConstant(module, "threadsafety", 1) < 0)
      INITERROR;

    if (PyModule_AddStringConstant(module, "paramstyle", "qmark") < 0)
      INITERROR;

    /* version for easier support */
    if (PyModule_AddStringConstant(module, "version", "1.0") < 0)
      INITERROR;

    /* modes for connection and cursors */
    if (PyModule_AddIntConstant(module, "M_EXCLUSIVE", mExclusive) < 0)
      INITERROR;

    if (PyModule_AddIntConstant(module, "M_OPTIMISTIC", mOptimistic) < 0)
      INITERROR;

    if (PyModule_AddIntConstant(module, "M_AUTOCOMMIT", mAutocommit) < 0)
      INITERROR;                                                            /* ... 25.12.13 */

  #endif

  #if PY_VERSION_HEX >= 0x02060000
    return module;
  #endif
}

#ifdef DEBUG
void linter_trace(char *fmt, ...)
{
  static FILE *file = NULL;
  va_list arglist;
  va_start(arglist, fmt);
  if (file == NULL)
  {
    file = fopen("linter.log", "w");
    if (file == NULL)
    {
      file = stderr;
    }
  }
  vfprintf(file, fmt, arglist);
  fflush(file);
  va_end(arglist);
}
#endif
