 /***************************************************************************
 *                 RELEX, Inc. Linter Python driver source code             *
 *        This is unpublished proprietary source code of RELEX, Inc.        *
 *     The copyright notice does not evidence any actual or intended        *
 *                  publication of such source code.                        *
 *                                                                          *
 *                 Copyright (c) 2003-2014 RELEX Inc.                       *
 *                                                                          *
 *                   This software was developed by                         *
 *                                                                          *
 *                            RELEX, Inc.                                   *
 *                         Voronezh, Russia                                 *
 *                                                                          *
 *                        All rights reserved.                              *
 ****************************************************************************
 AUTHOR: Boris Fedorov
 ----------------------------------------------------------------------------
 PURPOSE: Number interface (DBAPI)
 ----------------------------------------------------------------------------
 REVISION HISTORY:
 31.10.03 -- Boris Fedorov (Remove C++ style comments)
 29.07.08 -- Ermakov (Corrected warnings, added Boolean type)
 28.11.08 -- Ermakov (Added unicode for some cases)
 08.12.08 -- Ermakov (Corrected NULL for number)
 25.12.13 -- Muratova (Python 3 support)
 01.10.14 -- Kononov (Fixed return of Py_None value if DECIMAL is NULL)
 02.10.14 -- Kononov (Fixed binding boolean value as string for prepared query)
*/

#undef _DEBUG
#include <Python.h>
#include <linapi.h>
#include "LinPy.h"
#include <decimals.h>
#include <int64.h>                                               /* 28.11.08 */

static int Number_SetValue(t_Array*, unsigned, PyObject*);
static PyObject *Number_GetValue(t_Array*, unsigned);
static int DECIMAL_SetValue(t_Array*, unsigned, PyObject*);
static PyObject *DECIMAL_GetValue(t_Array*, unsigned);


PyTypeObject g_NumberType = {
	  PyVarObject_HEAD_INIT(NULL, 0)                         /* 25.12.13 */
	  "NUMBER",			 /*tp_name*/
	  sizeof(t_Array),					/*tp_basicsize*/
	  0,									 /*tp_itemsize*/
	  (destructor) Array_Free, /*tp_dealloc*/
	  0,									 /*tp_print*/
	  0,									 /*tp_getattr*/
	  0,									 /*tp_setattr*/
	  0,									 /*tp_compare*/
	  0,									 /*tp_repr*/
	  0,									 /*tp_as_number*/
	  0,									 /*tp_as_sequence*/
	  0,									 /*tp_as_mapping*/
	  0,									 /*tp_hash */
	  0,									 /*tp_call*/
	  0,									 /*tp_str*/
	  0,                   /*tp_getattro*/
	  0,									 /*tp_setattro*/
	  0,									 /*tp_as_buffer*/
	  Py_TPFLAGS_DEFAULT,					/*tp_flags*/
	  0	                                    /* tp_doc */
#if PY_VERSION_HEX >= 0x02060000	  												/* 25.12.13 ... */
	 , 0,									 /* tp_traverse */
	  0,									 /* tp_clear */
	  0,									 /* tp_richcompare */
	  0,									 /* tp_weaklistoffset */
	  0,									 /* tp_iter */
	  0,									 /* tp_iternext */
	  0,									 /* tp_methods */
	  0,									 /* tp_members */
	  0,									 /* tp_getset */
	  0,									 /* tp_base */
	  0,									 /* tp_dict */
	  0,									 /* tp_descr_get */
	  0,									 /* tp_descr_set */
	  0,									 /* tp_dictoffset */
	  0,									 /* tp_init */   /* ...25.12.13 */
#endif	  
};

t_ArrayType vt_SmallInt =
{
  (InitializeProc) NULL,
  (FinalizeProc) NULL,
  (SetValueProc) Number_SetValue,
  (GetValueProc) Number_GetValue,
  Array_InternalBindAnswer,
  NULL,
  NULL,
  &g_NumberType,
  tSmallInt,
  sizeof(L_SWORD),
  0
};

t_ArrayType vt_Integer =
{
  (InitializeProc) NULL,
  (FinalizeProc) NULL,
  (SetValueProc) Number_SetValue,
  (GetValueProc) Number_GetValue,
  Array_InternalBindAnswer,
  NULL,
  NULL,
  &g_NumberType,
  tInteger,
  sizeof(L_LONG),
  0
};

t_ArrayType vt_BigInt =
{
  (InitializeProc) NULL,
  (FinalizeProc) NULL,
  (SetValueProc) Number_SetValue,
  (GetValueProc) Number_GetValue,
  Array_InternalBindAnswer,
  NULL,
  NULL,
  &g_NumberType,
  tBigInt,
  sizeof(L_DLONG),
  0
};

t_ArrayType vt_Boolean =                                        /* 29.07.08 */
{
  (InitializeProc) NULL,
  (FinalizeProc) NULL,
  (SetValueProc) Number_SetValue,
  (GetValueProc) Number_GetValue,
  Array_InternalBindAnswer,
  NULL,
  NULL,
  &g_NumberType,
  tBoolean,
  1,
  0
};

t_ArrayType vt_Real =
{
  (InitializeProc) NULL,
  (FinalizeProc) NULL,
  (SetValueProc) Number_SetValue,
  (GetValueProc) Number_GetValue,
  Array_InternalBindAnswer,
  NULL,
  NULL,
  &g_NumberType,
  tReal,
  sizeof(L_REAL),
  0
};

t_ArrayType vt_Double =
{
  (InitializeProc) NULL,
  (FinalizeProc) NULL,
  (SetValueProc) Number_SetValue,
  (GetValueProc) Number_GetValue,
  Array_InternalBindAnswer,
  NULL,
  NULL,
  &g_NumberType,
  tDouble,
  sizeof(L_DOUBLE),
  0
};

t_ArrayType vt_Decimal =
{
  (InitializeProc) NULL,
  (FinalizeProc) NULL,
  (SetValueProc) DECIMAL_SetValue,
  (GetValueProc) DECIMAL_GetValue,
  Array_InternalBindAnswer,
  NULL,
  NULL,
  &g_NumberType,
  tDecimal,
  sizeof(L_DECIMAL),
  0
};

/**
 * Set the value of the variable.
 */
static int Number_SetValue(t_Array *array, unsigned pos, PyObject *value)
{
  double doubleValue;
  int integerValue;

  if (value == Py_None)
  {
    array->indicator[pos] = 1;
    return 0;
  }
  
  if (!PyInt_Check(value) && !PyFloat_Check(value) && !PyUnicode_Check(value) && 
      !PyString_Check(value)) /* 28.11.08 */ /* 25.12.13 */ /* */
  {
    PyErr_SetString(PyExc_TypeError, "expecting numeric data");
    return -1;
  }

  array->indicator[pos] = 0;

  if (PyInt_Check(value))
  {
    integerValue = PyInt_AS_LONG(value);
    trace((" int %i\n", integerValue));
    switch (array->type->linterType)
    {
      case tSmallInt:
        if (integerValue > L_MAXSWORD)
        {
          PyErr_SetString((PyObject *) &g_DataErrorException, "integer truncation error");
          return -1;
        }
        ((L_SWORD *) array->data)[pos] = (L_SWORD) integerValue;
        array->actualLength[pos] = sizeof(L_SWORD);
        break;
      case tInteger:
        ((L_LONG *) array->data)[pos] = (L_LONG) integerValue;
        array->actualLength[pos] = sizeof(L_LONG);
        break;
      case tBigInt:
        ((L_DLONG *) array->data)[pos] = (L_DLONG) integerValue;
        array->actualLength[pos] = sizeof(L_DLONG);
        break;
      case tBoolean:                                          /* 29.07.08 */
        ((L_BYTE *) array->data)[pos] = (L_BYTE) integerValue;
        array->actualLength[pos] = 1;
        break;
      case tReal:
        ((L_REAL *) array->data)[pos] = (L_REAL) integerValue;
        array->actualLength[pos] = sizeof(L_REAL);
        break;
      case tDouble:
        ((L_DOUBLE *) array->data)[pos] = (L_DOUBLE) integerValue;
        array->actualLength[pos] = sizeof(L_DOUBLE);
        break;
      default:
        PyErr_SetString((PyObject *) &g_InterfaceErrorException, "unsupported linter type assignment");
        return -1;
    }
  }
  else if (PyFloat_Check(value))
  {
    doubleValue = PyFloat_AS_DOUBLE(value);
    trace((" double %f\n", doubleValue));
    switch (array->type->linterType)
    {
      case tSmallInt:
        ((L_SWORD *) array->data)[pos] = (L_SWORD) doubleValue;
        array->actualLength[pos] = sizeof(L_SWORD);
        break;
      case tInteger:
        ((L_LONG *) array->data)[pos] = (L_LONG) doubleValue;
        array->actualLength[pos] = sizeof(L_LONG);
        break;
      case tBigInt:
        ((L_DLONG *) array->data)[pos] = (L_DLONG) doubleValue;
        array->actualLength[pos] = sizeof(L_DLONG);
        break;
      case tReal:
        ((L_REAL *) array->data)[pos] = (L_REAL) doubleValue;
        array->actualLength[pos] = sizeof(L_REAL);
        break;
      case tDouble:
        ((L_DOUBLE *) array->data)[pos] = (L_DOUBLE) doubleValue;
        array->actualLength[pos] = sizeof(L_DOUBLE);
        break;
      default:
        PyErr_SetString((PyObject *) &g_InterfaceErrorException, "unsupported linter type assignment");
        return -1;
    }
  } else
  {                                                          /* 28.11.08 ... */
    Py_UNICODE  uniValue[100];
    char        stringValue[100];
    int         Len;

    if (PyString_Check(value))                                /* 25.12.13 ... */ /* */
    {
      strncpy(stringValue,PyString_AS_STRING(value),100);
    }
    else
    {
      Len = PyUnicode_AsWideChar((PyObject *)value, (wchar_t *)uniValue,99); /* ... 25.12.13 */ /* */
      uniValue[Len] = 0;
      wcstombs(stringValue,(wchar_t *)uniValue,100);
    }
    stringValue[99] = 0;
    trace((" str %s\n", stringValue));
    switch (array->type->linterType)
    {
      case tSmallInt:
        ((L_SWORD *) array->data)[pos] = (L_SWORD) atoi(stringValue);
        array->actualLength[pos] = sizeof(L_SWORD);
        break;
      case tInteger:
        ((L_LONG *) array->data)[pos] = (L_LONG) atoi(stringValue);
        array->actualLength[pos] = sizeof(L_LONG);
        break;
      case tBigInt:
        ((L_DLONG *) array->data)[pos] = i64_atol(stringValue);
        array->actualLength[pos] = sizeof(L_DLONG);
        break;
      case tReal:
        ((L_REAL *) array->data)[pos] = (L_REAL) atof(stringValue);
        array->actualLength[pos] = sizeof(L_REAL);
        break;
      case tDouble:
        ((L_DOUBLE *) array->data)[pos] = (L_DOUBLE) atof(stringValue);
        array->actualLength[pos] = sizeof(L_DOUBLE);
        break;
      case tBoolean:                                                   /* 02.10.14 */
        ((L_BYTE *) array->data)[pos] = (L_BYTE) atoi(stringValue);
        array->actualLength[pos] = sizeof(L_BYTE);
        break;
      default:
        PyErr_SetString((PyObject *) &g_InterfaceErrorException, "unsupported linter type assignment");
        return -1;
    }
  }                                                          /* ... 28.11.08 */

  return 0;
}


/**
 * Returns the value stored at the given array position.
 */
static PyObject *Number_GetValue(t_Array *array, unsigned pos)
{
  double doubleValue;
  int integerValue;

  if (array->indicator[pos])
  {
    Py_INCREF(Py_None);
    return Py_None;
  }

  switch (array->type->linterType)
  {
    case tSmallInt:
      integerValue = (int) ((L_SWORD *) array->data)[pos];
      return PyInt_FromLong(integerValue);

    case tInteger:
      integerValue = (int) ((L_LONG *) array->data)[pos];
      return PyInt_FromLong(integerValue);

    case tBigInt:
      if (((L_DLONG *) array->data)[pos] > L_MAXLONG)
      {
        PyErr_SetString((PyObject *) &g_DataErrorException, "bigint truncation error");
        return NULL;
      }
      integerValue = (int) ((L_DLONG *) array->data)[pos];
      return PyInt_FromLong(integerValue);

    case tBoolean:                                           /* 29.07.08 */
      integerValue = (int) ((L_BYTE *) array->data)[pos];
      return PyInt_FromLong(integerValue);

    case tReal:
      doubleValue = (double) ((L_REAL *) array->data)[pos];
      return PyFloat_FromDouble(doubleValue);

    case tDouble:
      doubleValue = (double) ((L_DOUBLE *) array->data)[pos];
      return PyFloat_FromDouble(doubleValue);

    default:
      Py_INCREF(Py_None);
      return Py_None;
  }
}


static int DECIMAL_SetValue(
  t_Array *array,
  unsigned pos,
  PyObject *value)
{
  double doubleValue;
  int integerValue;
  char *stringValue;
  
  trace(("DECIMAL_SetValue(): array %08x pos %i", array, pos));

  array->indicator[pos] = 0;                                     /* 08.12.08 */
  if (PyInt_Check(value))
  {
    integerValue = PyInt_AS_LONG(value);
    trace((" int %i\n", integerValue));
    LongToDec(integerValue, ((L_DECIMAL *) array->data)[pos]);
    array->actualLength[pos] = sizeof(L_DECIMAL);
    return 0;
  }
  else if (PyFloat_Check(value))
  {
    doubleValue = PyFloat_AS_DOUBLE(value);
    trace((" double %f\n", doubleValue));
    DblToDec(doubleValue, ((L_DECIMAL *) array->data)[pos]);
    array->actualLength[pos] = sizeof(L_DECIMAL);
    return 0;
  }
  else if (PyString_Check(value))
  {
    stringValue = PyString_AS_STRING(value);
    trace((" str %s\n", stringValue));
    STRTODEC(stringValue, ((L_DECIMAL *) array->data)[pos]);
    array->actualLength[pos] = sizeof(L_DECIMAL);
    return 0;
  }
  else if (PyUnicode_Check(value))                           /* 28.11.08 ... */
  {
    Py_UNICODE  uniValue[100];
    char        stringValue[100];
    int         Len;

    Len = PyUnicode_AsWideChar((PyObject *)value, (wchar_t *)uniValue,99);
    uniValue[Len] = 0;
    wcstombs(stringValue,(wchar_t *)uniValue,100);
    trace((" str %s\n", stringValue));
    STRTODEC(stringValue, ((L_DECIMAL *) array->data)[pos]);
    array->actualLength[pos] = sizeof(L_DECIMAL);
    return 0;
  }                                                          /* ... 28.11.08 */

  PyErr_SetString(PyExc_TypeError, "expecting numeric data");
  array->indicator[pos] = 1;                                     /* 08.12.08 */
  return -1;
}


static PyObject *DECIMAL_GetValue(
  t_Array *array,
  unsigned pos)
{
  char str[48];
  if (array->indicator[pos]) /* 01.10.14 ... */
  {
    Py_INCREF(Py_None);
    return Py_None;
  }                          /* ... 01.10.14 */
  DECTOSTR(((L_DECIMAL *) array->data)[pos], str, array->desc.Prec, array->desc.Scale, 0);
  return PyString_FromString(str);
}
